/*
***                           File generat.c                      ***
***                        GeoFEST version 4.5.1
*** Copyright (c) 2003, California Institute of Technology        ***
*** U.S.Sponsorship under NASA Contract NAS7-1260 is acknowledged ***
***
*** This software is designated for public release under JPL Task ***
*** Plan Number 83-6791 and may be publicly released after        ***
*** approval for release has been granted by GSFC Software        ***
*** Release  Authority
***
*** This file contains modules for generating and reading 
*** node and element data and related isoparametric tasks 
***
***          == gen_element ==
***          == next_element ==
***          == read_surf ==
***          == read_buoy ==
***          == read_slip ==
***          == splitfind ==
***          == load_element ==
***          == gen_map ==
***          == gen_real ==
***          == gen_isopar_grid ==
***          == gen_shape_grid ==
***          == dot_sh ==
***
***/

/*
***GeoFEST version 4.5.1***
G. A. Lyzenga -- unified parallel/serial version and bug fixes 11/17/03
G. A. Lyzenga -- split node enhancements 8/16/02
G. A. Lyzenga -- modification to correctly relax hydrostatic stresses 6/15/00
J. W. Parker -- 3D support and deleted hypercube branches 9/16/98
G. A. Lyzenga  -- last modified on 3/03/95
Modification history:
2.1: fixes bug in boundary velocities and changes
     the way stresses are output.
2.2: implements split node faults
2.3: adds pass-thru comments
3.0: 3-D tetrahedra supported, hypercube branches deleted
3.1: hydrostatic relaxation changes
4.0: name change from visco to GeoFEST; correct implementation of 2-D out-of-
     plane stresses as well as hydrostatic relaxation; culling of remaining
     vestiges of hypercube stuff; one code without recompilation for both
     2-D and 3-D domains
4.1: Introduction of trilinear hexahedral bricks and PCG iterative solver option
4.2: Bug fixes; standard source prologues added (June 02)
4.3: Corrects inefficiency in split node algorithm and introduces new
     split node input format
4.4: Adds dummy calls to MPI and Pyramid to provide a single unified
     source for parallel and serial codes.  Also introduces a major
     algorithmic enhancement; stress equilibrium maintenance during visco
     elastic stepping for long-term stability.  Also more minor bug fixes.
4.5: Supports surface tractions and roller/spring duo options for tet elements.
     Provides enhanced conjugate gradient convergence monitoring.
4.5.1: Bug fix for split node faults in 2-d grids with degenerate quad triangles
 $Log: generat.c,v $
 Revision 1.4  2003/02/09 06:10:54  lyzenga
 *** empty log message ***

*/

#define EXTERN extern
#include "generat.h"
#include "utility.h"
void completion(void);

/* GC: global coordinate value of co (0,1,2 => x,y,z) for node nd */
#define GC(co,nd) (global.coords[(co) + (nd)*nsd])

/* XV:  tet basis vector, component co (0,1,2) for tet node en from node 0*/
#define XV(co,en) (GC(co,el_pt->ien[(en)] - 1) - GC(co,el_pt->ien[0] - 1))

static
     int      el_inc[3] ,
              nod_inc[3] ,
              el_end[3] ;
static
     int         current_element  ,
                 node             ,
                 gen_order        ,
                 nodes_gen[3]     ,
                 node_inc[3]      ;
static int    last_node ;
static int      *temp_list , *temp_where , *temp_grp ;
static real     *temp_trac , *temp_amount ;
static
     real  temp[3][8]   ;

static
     real shape[8] = { ZERO , ZERO , ZERO , ZERO
              , ZERO , ZERO , ZERO , ZERO } ;


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   gen_element   ***************
*/

/*  INTERFACE:   */

    void
    gen_element(
                pamr_Mesh mesh[],      /* parallel mesh */
                GROUP   *grp_ptr    /* pointer to current element group */
               )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine gen_element
** gen_element generates element storage and ien array
** It reads in element group information, individual elements,
** and surface traction records.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

   {
    char CVS_Id[] = "@(#) CVS id $Id: generat.c,v 1.4 2003/02/09 06:10:54 lyzenga Exp $, Revision: " REVNO;
    COUNTER  i , j , k , m , nel ;
    int       type , numsuf , numbuoy , ndof , nsd , nsf , plastic ,
              numface , tally , nfnode , count , this_node ;
    int       nel_local , mat , numel , numat , nen , nee,
              g_numel , num_local ;
    int       stress_size  , dbar_size , nint ; 
    int       x_index , y_index , x_offset , y_offset , next_el ;
    int       *fnode , *fgrp ;
    real      *fvec ;
    real      coeff , exponent , slip , ivec[3] , svec[3] , cross[3] ;
    real      xctr , yctr , zctr , dot , mult , norm , rho_g , upvec[3] ;
    ELEMENT_DATA   *el_pt ;
    BUOY_DATA      *buoy_ptr ;
    int       ien_current[8]  ;
    int       ien_temp[8] ;

/*  begin with element constants . . .  */

     fscanf(in_file,"%d",&g_numel) ;

     fscanf(in_file,"%d%d%d%d%d",&type,&numat,&numsuf,&numbuoy,&nfnode) ;
     
     if(pamr_num_procs() != 1) numel = pamr_local_element_count(mesh) ;
     else numel = g_numel ;
     
     (grp_ptr->el_info)->numel = numel ;
     (grp_ptr->el_info)->g_numel = g_numel ;
     (grp_ptr->el_info)->type = type ;
     (grp_ptr->el_info)->numat = numat ;
     (grp_ptr->el_info)->numsuf = numsuf ;
     (grp_ptr->el_info)->numbuoy = numbuoy ;
     (grp_ptr->el_info)->nfnode = nfnode ;
     (grp_ptr->el_info)->failed = 0 ;

     printf ("IAM: %d\t local numel = %d\n", pamr_my_pid() , numel) ;
     fflush(stdout) ;

     grp_ptr->el_data =
        (ELEMENT_DATA *) calloc(numel,sizeof(ELEMENT_DATA)) ;
     if (NULL == grp_ptr->el_data) attempt = ELD_MEM ;

     completion() ;

     switch( type )
        {
         case BILIN:
         case TET:
            nen = (grp_ptr->el_info)->nen = 4 ;
            break ;

         case SERENDIP:
         case TRILIN:
            nen = (grp_ptr->el_info)->nen = 8 ;
            break ;
        }
        
     ndof = fe_sys.ndof ;
     nsd = fe_sys.nsd ;
     nee = (grp_ptr->el_info)->nee = nen*ndof ;
     (grp_ptr->el_info)->el_size = nee*nee ;
     
     if( type == BILIN || type == SERENDIP || type == TET || type == TRILIN )
        {
         switch( ndof )
            {
             case 1:
                stress_size = (grp_ptr->el_info)->stress_size = 2 ;
                dbar_size   = (grp_ptr->el_info)->dbar_size   = 3 ;
                break ;
             case 2:
                stress_size = (grp_ptr->el_info)->stress_size = 4 ;
                dbar_size   = (grp_ptr->el_info)->dbar_size   = 10 ;
                break ;
             case 3:
                stress_size = (grp_ptr->el_info)->stress_size = 6 ;
                dbar_size   = (grp_ptr->el_info)->dbar_size   = 21 ;
                break ;
            }
         nsf = ndof ;
        }
     
/*  next get element material properties . . .  */

     (grp_ptr->el_mat)->mu     =
              (real *) calloc(numat,sizeof(real)) ;
     (grp_ptr->el_mat)->lambda =
              (real *) calloc(numat,sizeof(real)) ;
     (grp_ptr->el_mat)->visc   =
            (real *) calloc(2*numat,sizeof(real)) ;
     (grp_ptr->el_mat)->bforce =
         (real *) calloc(ndof*numat,sizeof(real)) ;
     (grp_ptr->el_mat)->plastic=
              (int *) calloc(numat,sizeof(int)) ;
     
     for(i=0 ; i<numat ; i++)
        {
         fscanf(in_file,"%lf",(grp_ptr->el_mat)->mu+i) ;
         fscanf(in_file,"%lf",(grp_ptr->el_mat)->lambda+i) ;
         fscanf(in_file,"%lf%lf",&coeff,&exponent) ;
         *((grp_ptr->el_mat)->visc+2*i) = coeff ;
         *((grp_ptr->el_mat)->visc+2*i+1) = exponent ;
         if(coeff == ZERO || exponent == ZERO)
            *((grp_ptr->el_mat)->plastic+i) = 0 ;
         else if(exponent == ONE)
            *((grp_ptr->el_mat)->plastic+i) = 1 ;
         else
            *((grp_ptr->el_mat)->plastic+i) = 2 ;
            
/* printf("mat=%d  mu=%g  lambda=%g  coeff=%g  exp=%g\n",
           i+1,*((grp_ptr->el_mat)->mu+i),*((grp_ptr->el_mat)->lambda+i),
           coeff,exponent) ;

   printf("material #%d plasticity code = %d\n",
             i+1,*((grp_ptr->el_mat)->plastic+i) ) ;
*/
         for(j=0;j<ndof;j++)
            fscanf(in_file,"%lf",(grp_ptr->el_mat)->bforce+ndof*i+j) ;
        }
    
     for( nel=0 ; nel<numel ; nel++ )
        {
         el_pt = grp_ptr->el_data + nel ;
          
   /*  now allocate memory for the ien and lm arrays... */

         el_pt->ien = (int *) calloc(nen,sizeof(int)) ;
         if (NULL == el_pt->ien) attempt = IEN_MEM ;

         el_pt->lm = (int *) calloc(nee,sizeof(int)) ;
         if (NULL == el_pt->lm) attempt = LM_MEM ;

        }

        completion() ;

/*  now generate elements, filling up the arrays just created... */

      nel_local = 0 ;  /* keep a tally of the elements local to this proc */
      
      while (next_element( &current_element , &gen_order ) !=  0 )
           {
	    bool mine = pamr_my_element(mesh, current_element) ;
             fscanf(in_file,"%d",&mat);  
             for ( i = 0 ; i < nen ; ++i) {
                 fscanf(in_file,"%d",&ien_current[i]);  
                 ien_current[i] = pamr_my_node_indx(mesh, ien_current[i]);
             }
             if ( gen_order != 0 )
                {
                 for ( i = 0 ; i < gen_order ; i++ )
                     fscanf(in_file,"%d",&el_inc[i]) ;  
                 for ( i = 0 ; i < gen_order ; i++ )
                     fscanf(in_file,"%d",&nod_inc[i]) ;  
                 for ( i = 0 ; i < gen_order ; i++ )
                     fscanf(in_file,"%d",&el_end[i]) ;  
                }
            if(mine)
               {
                if(0 == gen_order)
                   {
                   
   /* need to convert current_element into local element number... */
   
                    num_local = pamr_my_element_indx(mesh, current_element) ;
                    el_pt = grp_ptr->el_data + (num_local - 1) ;
                    load_element(el_pt,grp_ptr->el_info,
                                 num_local,ien_current,mat) ;
                    nel_local++ ;
                   }

                if(1 == gen_order)
                   {
                    x_index = (-1) ;
                    next_el = current_element - el_inc[0] ;
                    while( next_el < el_end[0] )
                       {
                        x_index++ ;
                        next_el += el_inc[0] ;
                        for(i=0 ; i<nen ; i++)
                        ien_temp[i] =
                           ien_current[i] + nod_inc[0]*x_index ;
                        el_pt = grp_ptr->el_data + next_el - 1 ;
                        load_element(el_pt,grp_ptr->el_info,next_el,
                                     ien_temp,mat) ;
                        nel_local++ ;
    
                       }
                   }
    
                if(2 == gen_order)
                   {
                    y_offset = -el_inc[1] ;
                    y_index = (-1) ;
                    while( y_offset < el_end[1]-current_element )
                       {
                        y_offset += el_inc[1] ;
                        y_index++ ;
    
                        x_offset = -el_inc[0] ;
                        x_index = (-1) ;
                        while( x_offset < el_end[0]-current_element )
                           {
                            x_offset += el_inc[0] ;
                            x_index++ ;
                            next_el =
                               current_element + y_offset + x_offset ;
                            for(i=0 ; i<nen ; i++)
                              ien_temp[i] = ien_current[i] +
                                       nod_inc[0]*x_index +
                                       nod_inc[1]*y_index ;
                            el_pt = grp_ptr->el_data + next_el - 1 ;
                            load_element(el_pt,grp_ptr->el_info,next_el,
                                         ien_temp,mat) ;
                            nel_local++ ;
                           }
                       }
                   }
               }
	    /*            completion() ; */
           }

      if(numel != nel_local) 
         {
          attempt = NEL_ERR ;
         }
/*         
      else
         printf ("IAM: %d\t element totals OK\n", pamr_my_pid()) ;
*/
         
      fflush(stdout) ;
     
      completion() ;

     for( nel=0 ; nel<numel ; nel++ )
        {
         el_pt = grp_ptr->el_data + nel ;
         m = el_pt->mat ;
         plastic = *( (grp_ptr->el_mat) -> plastic + m ) ;
         nint = el_pt->nint ;
          
   /*  next allocate memory for the stresses and visco arrays... */

         
         if( type == BILIN || type == SERENDIP || type == TET || type == TRILIN)
            {
             el_pt->stress =
                ( real * ) calloc( nint*stress_size , sizeof(real)) ;
             if (NULL == el_pt->stress) attempt = STRESS_MEM  ;
             
             el_pt->dilat =
                ( real * ) calloc( nint , sizeof(real)) ;
             if (NULL == el_pt->dilat) attempt = STRESS_MEM  ;

             el_pt->is_split =
                ( int * ) calloc( 8 , sizeof(int)) ;
  /* there should never be more than 7 split nodes in an element, I hope */              
             if (NULL == el_pt->is_split) attempt = STRESS_MEM  ;
             *(el_pt->is_split) = 0 ; /* initialize just to be sure */

             if( plastic )
                {
                 el_pt->bta =
                 ( real * ) calloc( nint*stress_size , sizeof(real)) ;
                 if (NULL == el_pt->bta) attempt = BTA_MEM  ;
            
                 el_pt->dbar =
                 ( real * ) calloc( nint*dbar_size , sizeof(real)) ;
                 if (NULL == el_pt->dbar) attempt = DBAR_MEM  ;

                }
            }
   /*  if PCG is being used, allocate memory for element stiffness storage */
         if(PCG == fe_sys.solver)
            {
             el_pt->stiff =
                ( real * ) calloc( (grp_ptr->el_info)->el_size , sizeof(real)) ;
             if (NULL == el_pt->stiff) attempt = ELSTIFF_MEM  ;
            }
           
        }
      completion() ;

   /*  now allocate memory and read in surface tractions... */
/*
      printf ("IAM: %d\t starting surface tractions...\n", pamr_my_pid()) ;
      fflush(stdout) ;
*/
      if( numsuf != 0 )
         {
          squawk("Processing surface tractions. . .\n") ;
          temp_list = (int *) calloc(numsuf,2*sizeof(int)) ;
          temp_trac = (real *) calloc(numsuf,nsf*sizeof(real)) ;
          
          tally = 0 ;
          
          read_surf( mesh , &tally , temp_list , temp_trac ,
                     type , numel , ndof ) ;

           if(tally != 0)
              {
               temp_list = (int *) realloc(temp_list , 2*sizeof(int)*tally) ;
               temp_trac = (real *) realloc(temp_trac , nsf*sizeof(real)*tally) ;
              }
           else  /* just a length-1 place holder */
              {
               temp_list = (int *) realloc(temp_list , 2*sizeof(int)) ;
               temp_trac = (real *) realloc(temp_trac , nsf*sizeof(real)) ;
              }

          (grp_ptr->el_info)->numsuf = tally ; /* numsuf now contains the proc LOCAL number! */
           printf ("IAM: %d\t local numsuf = %d\n", pamr_my_pid() , tally) ;
           fflush(stdout) ;
         
          (grp_ptr->el_info)->surf_list = temp_list ;
          (grp_ptr->el_info)->surf_trac = temp_trac ;
          squawk("Surface tractions all read in . . .\n") ;
         }
      else
        squawk("No surface tractions to process . . .\n") ;
        
      fflush(stdout) ;

   /*  now allocate memory and read in buoyancy surfaces... */

      if( numbuoy != 0 )
         {
          squawk("Processing buoyancy elements. . .\n") ;
          
          (grp_ptr->el_info)->buoy = (BUOY_DATA *) calloc(numbuoy,sizeof(BUOY_DATA)) ;

          for(i=0 ; i<numbuoy ; i++)
             {
              buoy_ptr = (grp_ptr->el_info)->buoy + i ;
              fscanf(in_file,"%d%lf%lf%lf%lf",&numface,&upvec[0],&upvec[1],&upvec[2],&rho_g);
              buoy_ptr->upvec[0] = upvec[0] ;
              buoy_ptr->upvec[1] = upvec[1] ;
              buoy_ptr->upvec[2] = upvec[2] ;
              buoy_ptr->numface = numface ;
              buoy_ptr->rho_g = rho_g ;
      /*
              printf("rho_g=%g\nupvec=(%g,%g,%g)\nnumface=%d\n\n",
                      rho_g,upvec[0],upvec[1],upvec[2],numface) ;
       */
              buoy_ptr->buoy_list = (int *) calloc(numface,2*sizeof(int)) ;
              tally = 0 ;
              read_buoy( mesh , &tally , buoy_ptr , 
                     type , numel , ndof ) ;
              if(tally != 0)
				  {
				   buoy_ptr->buoy_list = 
				     (int *) realloc(buoy_ptr->buoy_list , 2*sizeof(int)*tally) ;
				  }
              else  /* just a length-1 place holder */
				  {
				   buoy_ptr->buoy_list =
				      (int *) realloc(buoy_ptr->buoy_list , 2*sizeof(int)) ;
				  }
              buoy_ptr->numface = tally ; /* numface now contains the proc LOCAL number! */
              printf ("IAM: %d\t local numface = %d\n", pamr_my_pid() , tally) ;
              fflush(stdout) ;

             }
          
          squawk("Buoyancy elements all read in . . .\n") ;
         }
      else
        squawk("No buoyancy elements to process . . .\n") ;
        
      fflush(stdout) ;

   /*  now allocate memory for and read in and process split node faults... */

      if( 0 == nfnode )
         {
	  if (pamr_my_pid() == 0)
	    printf(
		"\nNo fault nodes; generation of element group #%d complete.\n",
		grp_ptr->group_num) ;
          return ;
         }
      fnode = (int *) calloc(nfnode,sizeof(int)) ;
      fvec = (real *) calloc(nfnode,7*sizeof(real)) ;
      fgrp = (int *) calloc(nfnode,sizeof(int)) ;
          
      read_slip( mesh, fnode , fvec , fgrp , type , nfnode , ndof ) ;
                     
      /* save this in attributes area.
         Make sure zeroed first! (at alloc point in inphase?)
      */

      /* Preserving number of split nodes */
      (grp_ptr->el_info)->nsplit = nfnode;


      for(i=0; i < nfnode; i++) {
        if (fnode[i] < 0) continue;
	global.splitn[NSPLITATTR*(fnode[i]-1)] = fvec[7*i];
	global.splitn[NSPLITATTR*(fnode[i]-1)+1] = fvec[7*i+1];
	global.splitn[NSPLITATTR*(fnode[i]-1)+2] = fvec[7*i+2];
	global.splitn[NSPLITATTR*(fnode[i]-1)+3] = fvec[7*i+3];
	global.splitn[NSPLITATTR*(fnode[i]-1)+4] = fvec[7*i+4];
	global.splitn[NSPLITATTR*(fnode[i]-1)+5] = fvec[7*i+5];
	global.splitn[NSPLITATTR*(fnode[i]-1)+6] = fvec[7*i+6];
	global.splitn[NSPLITATTR*(fnode[i]-1)+7] = 1;
      }

      temp_where = (int *) calloc(nfnode,2*sizeof(int)) ;
      temp_amount = (real *) calloc(nfnode,ndof*sizeof(real)) ;
      temp_grp = (int *) calloc(nfnode,sizeof(int)) ;
      
      if (temp_where == NULL || temp_amount == NULL) attempt = SPLIT_MEM  ;
      completion() ;
      
      tally = 0 ;
      for( nel=0 ; nel<numel ; nel++ )
         {
          el_pt = grp_ptr->el_data + nel ;
          for(i=0 ; i<nen ; i++)
             {
             /* careful not to double count degenerate quads! */
              if(type == BILIN && i==(nen-1) && el_pt->degen == 1) continue ;
              
              this_node = *(el_pt->ien + i) ;
                  j = splitfind(this_node,fnode,nfnode) ;
/* j is the zero-based index number of the searched-for split node input entry */

                  if(j != -1)
                     {
                      tally++ ;
					  temp_where = (int *) realloc(temp_where,2*sizeof(int)*tally) ;
					  temp_amount = (real *) realloc(temp_amount,ndof*sizeof(real)*tally) ;
					  temp_grp = (int *) realloc(temp_grp,sizeof(int)*tally) ;
					  if (temp_where == NULL || temp_amount == NULL) attempt = SPLIT_MEM  ;
                      
                      *(temp_where+2*(tally-1)) = nel ;         /* zero-based el# */
                      *(temp_where+2*(tally-1)+1) = this_node ; /* one-based node# */
                      *(temp_grp+(tally-1)) = *(fgrp+j) ;         /* hist group# */
                      
                      count = (*(el_pt->is_split)) + 1 ;
                      *(el_pt->is_split) = count ;
                      *(el_pt->is_split + count) = tally-1 ;
 /* flag and indices set for easy lookup of split nodes later during stress calc */

                      ivec[0] = *(fvec + 7*j) ;
                      ivec[1] = *(fvec + 7*j + 1) ; /* ivec is the direction perp to slip */
                      ivec[2] = *(fvec + 7*j + 2) ;
                      svec[0] = *(fvec + 7*j + 3) ;
                      svec[1] = *(fvec + 7*j + 4) ; /* svec is the slip vector direction */
                      svec[2] = *(fvec + 7*j + 5) ;
                      slip    = *(fvec + 7*j + 6) ; /* slip is the full slip amplitude */
                     
                      norm = (float)sqrt(ivec[0]*ivec[0] + ivec[1]*ivec[1] + ivec[2]*ivec[2]) ;
                      if((float)fabs(norm - ONE) > 1.0e-03)
                         {
                          squawk("ivec not normalized!\n");
                          printf("node #%d , splitnode #%d , instance #%d\n",
                                 this_node,j+1,tally) ;
                          printf("ivec = [ %g %g %g ]\n",ivec[0],ivec[1],ivec[2]) ;
                          printf("|ivec| = %g\n", norm );
                          attempt = IVEC_NORM ;
                         }
                      norm = sqrt(svec[0]*svec[0] + svec[1]*svec[1] + svec[2]*svec[2]) ;
                      if(fabs(norm - ONE) > 1.0e-03)
                         {
                          squawk("svec not normalized!\n");
                          printf("node #%d , splitnode #%d , instance #%d\n",
                                 this_node,j+1,tally) ;
                          printf("svec = [ %g %g %g ]\n",svec[0],svec[1],svec[2]) ;
                          printf("|svec| = %g\n", norm );
                          attempt = SVEC_NORM ;
                         }
                      if(fabs(ivec[0]*svec[0]+ivec[1]*svec[1]+ivec[2]*svec[2]) > 1.0e-03)
                         {
                          squawk("ivec and svec not perpendicular!\n");
                          attempt = ISVEC_PERP ;
                         }
                     
                      cross[0] = svec[1]*ivec[2] - svec[2]*ivec[1] ;
                      cross[1] = svec[2]*ivec[0] - svec[0]*ivec[2] ;
                      cross[2] = svec[0]*ivec[1] - svec[1]*ivec[0] ;
                      
                      xctr = ZERO ;
                      yctr = ZERO ;
                      zctr = ZERO ;
                      for(k=0 ; k<nen ; k++)
                         {
                          xctr += *(global.coords + nsd * ( *(el_pt->ien + k) - 1 ) ) ;
                          yctr += *(global.coords + nsd * ( *(el_pt->ien + k) - 1 ) + 1 ) ;
                          if(nsd == 3)
                          zctr += *(global.coords + nsd * ( *(el_pt->ien + k) - 1 ) + 2 ) ;
                         }
                      xctr = xctr/(real)nen - *(global.coords + nsd * ( this_node - 1 ) ) ;
                      yctr = yctr/(real)nen - *(global.coords + nsd * ( this_node - 1 ) + 1 ) ;
                      if(nsd == 3)
                      zctr = zctr/(real)nen - *(global.coords + nsd * ( this_node - 1 ) + 2 ) ;
                      
                      dot = cross[0]*xctr + cross[1]*yctr + cross[2]*zctr ;
                      if( dot > ZERO )
                         mult = slip * PT5 ;
                      else
                         mult = -slip * PT5 ;
                      switch( ndof )
                         {
                          case 1:
                              *(temp_amount+ndof*(tally-1)) = svec[2]*mult ;
                              break ;
                          case 2:
                              *(temp_amount+ndof*(tally-1)) = svec[0]*mult ;
                              *(temp_amount+ndof*(tally-1)+1) = svec[1]*mult ;
                              break ;
                          case 3:
                              *(temp_amount+ndof*(tally-1)) = svec[0]*mult ;
                              *(temp_amount+ndof*(tally-1)+1) = svec[1]*mult ;
                              *(temp_amount+ndof*(tally-1)+2) = svec[2]*mult ;
                              break ;
                         }

                     }
                  
             }     /* loop i over element nodes */
         }         /* loop nel over elements */
       completion() ;
         

      (grp_ptr->el_info)->nfnode = tally ; /* nfnode now contains the total number of instances! */
      (grp_ptr->el_info)->slip_list = temp_where ;
      (grp_ptr->el_info)->slip_val = temp_amount ;
      (grp_ptr->el_info)->slip_grp = temp_grp ;
      free( fnode ) ;
      free( fvec ) ;
      free( fgrp ) ;
      squawk("Split nodes all processed . . .\n") ;

      if (pamr_my_pid() == 0)
         printf("\nGeneration of element group #%d complete.\n",
             grp_ptr->group_num) ;
      fflush(stdout) ;
   }

/************************** end of gen_element *******************************/ 

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   next_element   ***************
*/

/*  INTERFACE:   */

   int  next_element(
                     int   *node_pointer , /* pointer to hold node */
                     int   *gen_pointer   /* pointer to gen parameter */
                    )

/*  RETURN VALUE:  number of the next node or object read in */
/*  DESCRIPTION:   */
/*
** Function next_element
** next_element reads the first part of a record in a double 
** null-terminated list, so the calling function may 
** catch the termination.
**
** The use of this function is not limited to finite elements, but is
** also employed for other lists, such as nodes and tractions.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

            {
             fscanf(in_file,"%d",node_pointer) ;  
             fscanf(in_file,"%d",gen_pointer) ;

             return(*node_pointer);
            }

/************************** end of next_element *******************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   read_surf   ***************
*/

/*  INTERFACE:   */

     void
     read_surf(
               pamr_Mesh m[], /* the mesh */
               int  *tally ,  /* counter for records read in */
               int  *temp_list ,  /* storage for surf location data */
               real  *temp_trac ,  /* storage for surf traction data */
               int  type ,    /* element type number */
               int  numel ,  /* number of elements */
               int  ndof    /* number of disp degrees of freedom */
              )

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine read_surf
** read_surf reads in surface traction records.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

   {
    int  nel , side , j , n ;
    real  trac[3] ;
    bool mine ;
    
    switch( type )
       {
        case BILIN:
        case SERENDIP:
        case TET:
        case TRILIN:
          while (next_element( &nel , &side ) !=  0 )
             {
              for(j=0;j<ndof;j++)
                fscanf(in_file,"%lf",trac+j) ;
                
              mine = pamr_my_element(m,nel) ;
              if(!mine) continue ; /* only keep those belonging to my proc */
              
              n = pamr_my_element_indx(m,nel) - 1 ; /* zero-based local */

              *(temp_list+2*(*tally)) = n ;
              *(temp_list+2*(*tally)+1) = side ;
              
              for(j=0;j<ndof;j++)
                *(temp_trac+ndof*(*tally)+j) = trac[j] ;
                
              (*tally)++ ;  /* increment the counter */
             }
          break ;

        case QUAKE:
          break ;
       }
   }
/************************** end of read_surf *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   read_buoy   ***************
*/

/*  INTERFACE:   */

     void
     read_buoy(
               pamr_Mesh m[], /* the mesh */
               int  *tally ,  /* counter for records read in */
               BUOY_DATA *buoy_ptr ,  /* buoyancy data structure */
               int  type ,    /* element type number */
               int  numel ,  /* number of elements */
               int  ndof    /* number of disp degrees of freedom */
              )

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine read_buoy
** read_buoy reads in buoyancy surface records.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

   {
    int  nel , side , j , n ;
    bool mine ;
    
    switch( type )
       {
        case BILIN:
        case SERENDIP:
        case TET:
        case TRILIN:
          while (next_element( &nel , &side ) !=  0 )
             {                
              mine = pamr_my_element(m,nel) ;
              if(!mine) continue ; /* only keep those belonging to my proc */
              
              n = pamr_my_element_indx(m,nel) - 1 ; /* zero-based local */

              *(buoy_ptr->buoy_list+2*(*tally)) = n ;
              *(buoy_ptr->buoy_list+2*(*tally)+1) = side ;
              
              (*tally)++ ;  /* increment the counter */
             }
          break ;

        case QUAKE:
          break ;
       }
   }
/************************** end of read_buoy *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   read_slip   ***************
*/

/*  INTERFACE:   */

     void
     read_slip( pamr_Mesh m[], /* the mesh */
               int  *fnode ,  /* global numbers of split nodes */
               real  *fvec ,  /* slip and plane vectors */
               int  *fgrp ,  /* history group # to which split node belongs */
               int  type ,    /* element type number */
               int  nfnode ,  /* number of split nodes */
               int  ndof    /* number of disp degrees of freedom */
              )

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine read_slip
** read_slip reads split-node records for symmetric fault 
** slip accumulation.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

   {
    int   n, i, j, k, itemp  ;
    real ftemp ;
    
    switch( type )
       {
        case BILIN:
        case SERENDIP:
        case TET:
        case TRILIN:
          for(n=0 ; n<nfnode ; n++)
             {
              fscanf(in_file,"%d%d%lf%lf%lf%lf%lf%lf%lf",
                fnode+n , fgrp+n , fvec+7*n , fvec+7*n+1 , fvec+7*n+2 ,
                 fvec+7*n+3 , fvec+7*n+4 , fvec+7*n+5 , fvec+7*n+6 ) ;
  /* node # , group # , ivec[0] , ivec[1] , ivec[2] , svec[0] , svec[1] , svec[2] , slip */

	      /* we need to change from global to local node numbers,
		 and then sort the entries into ascending node number order */

              *(fnode + n) = pamr_my_node_indx(m, *(fnode + n)) ;
             }
          break ;

        case QUAKE:
          break ;
       }

    /* now sort! not the most efficient, but I think it works...*/

    for(i=0 ; i<nfnode ; i++)
      {
        for(j=i+1 ; j<nfnode ; j++)
	  {
            if( *(fnode+j) < *(fnode+i) )
              {
		itemp = *(fnode+i) ;
		*(fnode+i) = *(fnode+j) ;
		*(fnode+j) = itemp ;
		
		itemp = *(fgrp+i) ;
		*(fgrp+i) = *(fgrp+j) ;
		*(fgrp+j) = itemp ;
		
		for(k=0 ; k<7 ; k++)
                  {
		    ftemp = *(fvec+7*i+k) ;
		    *(fvec+7*i+k) = *(fvec+7*j+k) ;
		    *(fvec+7*j+k) = ftemp ;
                  }
              }
	  }
      }

   }
/************************** end of read_slip *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   splitfind   ***************
*/

/*  INTERFACE:   */

 int splitfind(
               int   qnode ,  /* query node to search for */
               int  *fnode ,  /* global numbers of split nodes */
               int  nfnode    /* number of split nodes */
              )

/*  RETURN VALUE:   index number of found split node entry; -1 if not found */
/*  DESCRIPTION:   */
/*
** function splitfind
** splitfind performs an efficient bisection search on the list of split nodes 
** to see if qnode is among them.  If so, it returns the index number of
** that entry.  Otherwise, returns -1.  Assumes that the fnode list is
** provided sorted in ascending node number order!
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

{
 int  bot , mid , top ;
 
 bot = 0 ;
 top = nfnode-1 ;
 mid = top/2 ;
 
 if( !BETWEEN(fnode[bot],qnode,fnode[top]) )  return( -1 ) ;
 if(qnode == fnode[bot])  return( bot ) ;
 if(qnode == fnode[top])  return( top ) ;
 if(qnode == fnode[mid])  return( mid ) ;
 while((top-bot) > 2)
    {
     if( BETWEEN(fnode[bot],qnode,fnode[mid]) )
        {
         top = mid ;
         mid = bot + (top-bot)/2 ;
         if(qnode == fnode[mid])  return( mid ) ;
        }
     else
        {
         bot = mid ;
         mid = bot + (top-bot)/2 ;
         if(qnode == fnode[mid])  return( mid ) ;
        }
    }
 return( -1 ) ;
}
/************************** end of splitfind *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   load_element   ***************
*/

/*  INTERFACE:   */

  void
  load_element(
               ELEMENT_DATA  *el_pt  ,  /* ptr to current element struct */
               ELEMENT_INFO  *info ,   /* ptr to element info struct */
               int      nel_glob ,      /* one-based global element number */
               int      ien[] ,     /* holding for ien assignments */
               int      mat       /* one-based material index number */
              )

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine load_element 
** load_element loads element arrays with ien and equation data.
** The information for the element has been previously read into a 
** temporary array or generated. 
**
** This function checks for degeneracies of 2D elements,
** right-hand rule order of tet nodes (swapping when necessary),
** and fills in the ien and lm arrays for the element.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

    {
     COUNTER i , j , k , index , ieq , ndof , nsd ;
     int   node_number ;
     real s;
     int itmp;
     
     ndof = fe_sys.ndof ; 
     nsd = fe_sys.nsd ; 
     el_pt->nel =  nel_glob ;

/* load the ien array with one-based node numbers ...  */

     for(i=0 ; i<info->nen ; i++)
        {
         if( (*(el_pt->ien + i) = ien[i]) )
	      {
           continue ;
	      }
         else
	      {
           attempt = IEN_ERR*1000 + nel_glob ;
	      }
             /* Error if glob_to_local returns zero local node # */
        }
/* jwp: if tet, check triple product and reorder if necessary. TP should
be > 0 if tet is right-handed. Check for determinant >0 (later) at all
wt points .  
*/

     if(info->nen == 4 && nsd == 3) {
        s = XV(0,1)*XV(1,2)*XV(2,3) - XV(2,1)*XV(1,2)*XV(0,3) +
            XV(1,1)*XV(2,2)*XV(0,3) - XV(0,1)*XV(2,2)*XV(1,3) +
            XV(2,1)*XV(0,2)*XV(1,3) - XV(1,1)*XV(0,2)*XV(2,3);

        if(s < 0.0) 
        {
           itmp          = el_pt->ien[3];
           el_pt->ien[3] = el_pt->ien[2];
           el_pt->ien[2] = itmp;
        }
     }

     el_pt->degen = 0 ;
     switch( info->type )
        {
         case BILIN:
             el_pt->nint = 4 ;
             if( ien[2] == ien[3] )
                el_pt->degen = 1 ;
             break ;

         case TET:
             el_pt->nint = 4 ;
             if(mat < 0)
                {
                 mat = -mat ;
                 el_pt->degen = 1 ;  /* this is actually a "duo" element */
                }
             break ;

         case TRILIN:
             el_pt->nint = 6 ;
             if( ien[2] == ien[3] && ien[6] == ien[7] )
                el_pt->degen = 1 ; /* triangular prism */
             if( ien[2] == ien[3] && ien[4] == ien[5] && ien[4] == ien[6] && ien[4] == ien[7] )
                el_pt->degen = 2 ; /* tetrahedron */
             if(el_pt->degen != 0)  el_pt->nint = 8 ;
             break ;

         case SERENDIP:
             el_pt->nint = 9 ;
             if( ien[2] == ien[3] && ien[2] == ien[6] )
                el_pt->degen = 1 ;
             break ;
        }
     el_pt->mat = mat-1 ;        /* mat is stored in zero-based form  */

/* and load the lm array with zero-based equation numbers  */

     el_pt->is_bc = 0 ;
     index = 0 ;
     for( j = 0 ; j < info->nen ; j++)
        {
         node_number = *(el_pt->ien + j) ;
         for( k = 0 ; k < ndof ; k++)
             {
              ieq = *(global.id_pointer + ndof*(node_number-1) + k) ;
              *(el_pt->lm + index) = ieq ;
              if( ieq == (-1) )
                 el_pt->is_bc = 1 ;
              index++ ;
             }
        }
    }

/************************** end of load_element ******************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   gen_map   ***************
*/

/*  INTERFACE:   */

    void
    gen_map(pamr_Mesh m[])

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine gen_map
** gen_map fills the nodal activity map array.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

   {
    int		ndof = fe_sys.ndof;
    int		current_node, node_indx;
    int	    next_node;
    int	gen_inc ;
    int	end_node ;
    COUNTER	k , i ;
    int	bctype[3];
    
     
    /* the map array contains space for 
    up to 3 d.o.f. type designators */

    squawk("Start reading boundary info . . .\n") ;
    
    /* by default, make all entries free unless overidden */
    for(i=0 ; i<loc_sys.numnp ; i++)
       {
        for(k=0 ; k<ndof ; k++)
          *(global.eqtype+ndof*i+k) = 1 ;
       }
     
  /* now read in specified designators */
  
    while (next_element( &current_node , &gen_order ) !=  0 )
      {
       bool mine = pamr_my_node(m, current_node);
        for(k=0 ; k<ndof ; k++)
           fscanf(in_file,"%d",&bctype[k]) ;

        if( mine )
           {
            node_indx = pamr_my_node_indx(m, current_node) - 1 ; /* make it zero-based */
            for(k=0 ; k<ndof ; k++)
                *(global.eqtype+ndof*(node_indx)+k) = bctype[k] ;
           }

    if(1 == gen_order)  /* non-functional old code! */
       {
        next_node = current_node;
        gen_inc = 0;
        end_node = 0;
        fscanf(in_file,"%d%d",&gen_inc,&end_node);
        next_node = current_node ;
        while( next_node < end_node )
           {
            next_node += gen_inc ;
            if( mine )
               {
                last_node++ ;
                if(last_node > fe_sys.numnp)
                   {
                    attempt = NODES_ERR1 ;
                    continue ;
                   }
            
                for(k=0 ; k<ndof ; k++)
                    *(global.eqtype+ndof*(next_node-1)+k) = bctype[k] ;
               }
           }
       }
    
    if(gen_order == 2)  /* non-functional old code! */
       {
        int	x_inc, y_inc, end_x, end_y, x_offset, y_offset;
        fscanf(in_file,"%d%d%d%d",&x_inc,&y_inc,&end_x,&end_y);
        y_offset = -y_inc ;
        while( y_offset < end_y-current_node )
           {
            y_offset += y_inc ;

            x_offset = -x_inc ;
            while( x_offset < end_x-current_node )
               {
                x_offset += x_inc ;
                next_node = current_node + y_offset + x_offset ;
                if( next_node != current_node )
                   {
                    if( mine )
                       {
                        last_node++ ;
                        if(last_node > fe_sys.numnp)
                           {
                            attempt = NODES_ERR1 ;
                            continue ;
                           }
            
                        for(k=0 ; k<ndof ; k++)
                            *(global.eqtype+ndof*(next_node-1)+k) = bctype[k] ;
                       }
                   }
               }
           }
       }
      }  /* end while */
     
     completion() ; 

    squawk("Boundary code info complete.\n") ;
     
    }

/************************** end of gen_map ***********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   gen_real   ***************
*/

/*  INTERFACE:   */

    void
    gen_real(
             pamr_Mesh m[],      /* parallel mesh */
             real      *array ,  /* array of generated values */
             int        dim      /* dimensionality of grid */
            )

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine gen_real
** gen_real reads or generates floating point global array.
** Used to read or generate nodes, displacement bcs, or 
** velocity bcs.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {

      COUNTER i , j  ;
      int    l_node ;
      real   value ;
      bool   mine ;

      while (next_element( &node , &gen_order) != 0 )
            {
             if(0 == gen_order)
               {
		 if ((mine = pamr_my_node(m, node))) {
		   l_node = pamr_my_node_indx(m, node);
                 }
                for(i = 0 ; i < dim ; ++i)
                   {
                    fscanf(in_file,"%lf",&value) ;
                    if(mine)
                      *(array + (l_node-1)*dim + i) = value ;
                   } 
               }
             else
               {
                  for ( i = 0 ; i < gen_order ; ++i)
                    {
                       for( j = 0 ; j < dim ; ++j)
                           fscanf(in_file,"%lf",&temp[j][i]) ;
                    }
                  for ( i = 0 ; i < 3 ; ++i)
                    {
                     fscanf(in_file,"%d",&nodes_gen[i]) ;  
                     fscanf(in_file,"%d",&node_inc[i]) ;  
                    }
                  gen_isopar_grid(array,dim) ;
                  
               }
            }
    }

/************************** end of gen_real **********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   gen_isopar_grid   ***************
*/

/*  INTERFACE:   */

 void
 gen_isopar_grid(
                 real      *array ,  /* array of generated values */
                 int        dim      /* dimensionality of grid */
                )

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine gen_isopar_grid
** isopar_grid generates real data via isoparametric interpolation
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
      int      option = 3 ; /* set to gen over a volume */
      int       incr = 1 ;
      real dr  , ds  , dt    ,
             r  ,  s  ,  t    ;
      real dot_sh ()         ;

      int    l_node ;
      real   value ;
      COUNTER   i , j  ,  k   ,  l  ;

      if(2 == gen_order)
        {
         option = 1 ;       /*  set to gen over a line   */
         if(nodes_gen[0] != 0) incr = node_inc[0] + 1 ;
         if(nodes_gen[1] != 0) incr = node_inc[1] + 1 ;
         if(nodes_gen[2] != 0) incr = node_inc[2] + 1 ;
        }
      else if(4 == gen_order)
             {
              option = 2 ;     /*  set to gen over an area   */
              incr = 1   ;  /* this needs to be made more general */
             }
      else if(gen_order != 8) attempt = GEN_ERR ;


      dr = (nodes_gen[0] != 0) ?   TWO/ (real) nodes_gen[0] : ZERO ;
      ds = (nodes_gen[1] != 0) ?   TWO/ (real) nodes_gen[1] : ZERO ;
      dt = (nodes_gen[2] != 0) ?   TWO/ (real) nodes_gen[2] : ZERO ;

      nodes_gen[0]++ ;
      nodes_gen[1]++ ;
      nodes_gen[2]++ ;

      t = - ONE ;
      for ( k = 0 ; k < *(nodes_gen + 2) ; ++k )
          {
            s = -ONE ;
            for ( j = 0 ; j < *(nodes_gen + 1) ; ++j)
             {
              r = -ONE ;
              for (i = 0 ; i < *nodes_gen ; ++i)
               {

                gen_shape_grid(r,s,t,shape,option) ;
                l_node = node ;
                  for (l = 0 ; l < dim ; ++l)
                     {
                      value = dot_sh(shape,temp[l],gen_order);
                      if(l_node)
                        *(array + (l_node-1)*dim + l) = value ;
                     } 
                node += incr            ;
                r += dr                 ;

               }
              s += ds  ;
             }
           t += dt ;
          }
      }

/************************** end of gen_isopar_grid ***************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   gen_shape_grid   ***************
*/

/*  INTERFACE:   */

  void
  gen_shape_grid(
                 real      r  ,  /* isopar coordinate */
                 real      s  ,  /* isopar coordinate */
                 real      t  ,  /* isopar coordinate */
                 real      *shape  ,  /* shape function storage */
                 int       option      /* generation option flag */
                )

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine gen_shape_grid
** gen_shape_grid finds shape function for isoparametric generation.
** It appears to JWP to use bilinear and trilinear interpolation.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
      real r1   , r2   , s1   ,
            s2   , t1   , t2   ;
      real *shape1 = shape + 1 ;

      if(1 == option)
        {
         *shape1 = r*PT5 ;
         *shape     = PT5 - *shape1 ;
         *shape1 = PT5 + *shape1 ;
        }
      else if(2 == option)
           {
            r2 = r*PT5 ;
            r1 = PT5-r2 ;
            r2 += PT5   ;
            s2 = s*PT5  ;
            s1 = PT5-s2 ;
            s2 += PT5   ;
            *shape++ = r1*s1 ;
            *shape++ = r2*s1 ;
            *shape++ = r2*s2 ;
            *shape   = r1*s2 ;
           }
           else
           {
            r2 = r*PT5 ;
            r1 = PT5-r2 ;
            r2 += PT5   ;

            s2 = s*PT5  ;
            s1 = PT5-s2 ;
            s2 += PT5   ;

            t2 = t*PT5  ;
            t1 = PT5-t2 ;
            t2 += PT5   ;

            *shape++ = r1*s1*t1 ;
            *shape++ = r2*s1*t1 ;
            *shape++ = r2*s2*t1 ;
            *shape++ = r1*s2*t1 ;
            *shape++ = r1*s1*t2 ;
            *shape++ = r2*s1*t2 ;
            *shape++ = r2*s2*t2 ;
            *shape   = r1*s2*t2 ;
            }
     }

/************************** end of gen_shape_grid ****************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   dot_sh   ***************
*/

/*  INTERFACE:   */

    real  dot_sh(
                 real      *shape  ,     /* shape function storage */
                 real      *variable  ,  /* interpolate variable storage */
                 int        n            /* number to sum over */
                )

/*  RETURN VALUE:   dot product value */
/*  DESCRIPTION:   */
/*
** Function dot_sh
** dot_sh performs the dot product of the vector of shape functions
** (may be gradients of shape functions)
** with the vector "variable"
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

      {
       real sum = ZERO             ,
             *shape_end = shape + n ;

       while (shape < shape_end) sum += (*shape++)*(*variable++) ;

       return(sum) ;
      }

/************************** end of dot_sh ************************************/ 

