/*
***                           File generat.h                      ***
***                        GeoFEST version 4.5
*** Copyright (c) 2003, California Institute of Technology        ***
*** U.S.Sponsorship under NASA Contract NAS7-1260 is acknowledged ***
***
*** This software is designated for public release under JPL Task ***
*** Plan Number 83-6791 and may be publicly released after        ***
*** approval for release has been granted by GSFC Software        ***
*** Release  Authority
***
*** This file contains modules for generating and reading 
*** node and element data and related isoparametric tasks 
***
***          == gen_element ==
***          == next_element ==
***          == read_surf ==
***          == read_slip ==
***          == splitfind ==
***          == load_element ==
***          == gen_map ==
***          == gen_real ==
***          == gen_isopar_grid ==
***          == gen_shape_grid ==
***          == dot_sh ==
***
***/

/*
***GeoFEST version 4.5***
G. A. Lyzenga -- unified parallel/serial version and bug fixes 11/17/03
G. A. Lyzenga -- split node enhancements 8/16/02
G. A. Lyzenga -- modification to correctly relax hydrostatic stresses 6/15/00
J. W. Parker -- 3D support and deleted hypercube branches 9/16/98
G. A. Lyzenga  -- last modified on 3/03/95
Modification history:
2.1: fixes bug in boundary velocities and changes
     the way stresses are output.
2.2: implements split node faults
2.3: adds pass-thru comments
3.0: 3-D tetrahedra supported, hypercube branches deleted
3.1: hydrostatic relaxation changes
4.0: name change from visco to GeoFEST; correct implementation of 2-D out-of-
     plane stresses as well as hydrostatic relaxation; culling of remaining
     vestiges of hypercube stuff; one code without recompilation for both
     2-D and 3-D domains
4.1: Introduction of trilinear hexahedral bricks and PCG iterative solver option
4.2: Bug fixes; standard source prologues added (June 02)
4.3: Corrects inefficiency in split node algorithm and introduces new
     split node input format
4.4: Adds dummy calls to MPI and Pyramid to provide a single unified
     source for parallel and serial codes.  Also introduces a major
     algorithmic enhancement; stress equilibrium maintenance during visco
     elastic stepping for long-term stability.  Also more minor bug fixes.
4.5: Supports surface tractions and roller/spring duo options for tet elements.
     Provides enhanced conjugate gradient convergence monitoring.
 $Log: generat.h,v $
 Revision 1.1  2003/02/09 06:10:55  lyzenga
 *** empty log message ***


*/

#include <stdio.h>
#include <stdlib.h>
#include "mydefs.h"
#include "errcodes.h"
#include "finel.h"
#include <math.h>

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   gen_element   ***************
*/

/*  INTERFACE:   */

void
gen_element(
  pamr_Mesh m[],        /* parallel mesh */
  GROUP*	grp_ptr	/* pointer to current element group		*/
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine gen_element
** gen_element generates element storage and ien array
** It reads in element group information, individual elements,
** and surface traction records.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of gen_element *******************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   next_element   ***************
*/

/*  INTERFACE:   */

int next_element(
  int*	node_pointer,	/* pointer to hold node				*/
  int*	gen_pointer	/* pointer to gen parameter			*/
  );

/*  RETURN VALUE:  number of the next node or object read in */
/*  DESCRIPTION:   */
/*
** Function next_element
** next_element reads the first part of a record in a double 
** null-terminated list, so the calling function may 
** catch the termination.
**
** The use of this function is not limited to finite elements, but is
** also employed for other lists, such as nodes and tractions.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of next_element *******************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   read_surf   ***************
*/

/*  INTERFACE:   */

void
read_surf(
  pamr_Mesh m[], /* the mesh */
  int*	tally,		/* counter for records read in			*/
  int*	temp_list,	/* storage for surf location data		*/
  real*	temp_trac,	/* storage for surf traction data		*/
  int	type,		/* element type number				*/
  int	numel,		/* number of elements				*/
  int	ndof		/* number of disp degrees of freedom		*/
  );

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine read_surf
** read_surf reads in surface traction records.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of read_surf *********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   read_buoy   ***************
*/

/*  INTERFACE:   */

     void
     read_buoy(
               pamr_Mesh m[], /* the mesh */
               int  *tally ,  /* counter for records read in */
               BUOY_DATA *buoy_ptr ,  /* buoyancy data structure */
               int  type ,    /* element type number */
               int  numel ,  /* number of elements */
               int  ndof    /* number of disp degrees of freedom */
              );

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine read_buoy
** read_buoy reads in buoyancy surface records.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of read_buoy *********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   read_slip   ***************
*/

/*  INTERFACE:   */

     void
     read_slip( pamr_Mesh m[], /* the mesh */
               int  *fnode ,  /* global numbers of split nodes */
               real  *fvec ,  /* slip and plane vectors */
               int  *fgrp ,  /* history group # to which split node belongs */
               int  type ,    /* element type number */
               int  nfnode ,  /* number of split nodes */
               int  ndof    /* number of disp degrees of freedom */
              );

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine read_slip
** read_slip reads split-node records for symmetric fault 
** slip accumulation.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of read_slip *********************************/
/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   splitfind   ***************
*/

/*  INTERFACE:   */

 int splitfind(
               int   qnode ,  /* query node to search for */
               int  *fnode ,  /* global numbers of split nodes */
               int  nfnode    /* number of split nodes */
              );

/*  RETURN VALUE:   index number of found split node entry; -1 if not found */
/*  DESCRIPTION:   */
/*
** function splitfind
** splitfind performs an efficient bisection search on the list of split nodes 
** to see if qnode is among them.  If so, it returns the index number of
** that entry.  Otherwise, returns -1.  Assumes that the fnode list is
** provided sorted in ascending node number order!
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of splitfind *********************************/
/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   load_element   ***************
*/

/*  INTERFACE:   */

void
load_element(
  ELEMENT_DATA*	el_pt,		/* ptr to current element struct	*/
  ELEMENT_INFO*	info,		/* ptr to element info struct		*/
  int		nel_glob,	/* one-based global element number	*/
  int		ien[],		/* holding for ien assignments		*/
  int		mat		/* one-based material index number	*/
  );

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine load_element 
** load_element loads element arrays with ien and equation data.
** The information for the element has been previously read into a 
** temporary array or generated. 
**
** This function checks for degeneracies of 2D elements,
** right-hand rule order of tet nodes (swapping when necessary),
** and fills in the ien and lm arrays for the element.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of load_element ******************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   gen_map   ***************
*/

/*  INTERFACE:   */

void
gen_map(pamr_Mesh m[]);

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine gen_map
** gen_map fills the nodal activity map array.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of gen_map ***********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   gen_real   ***************
*/

/*  INTERFACE:   */

void
gen_real(
  pamr_Mesh m[], /* parallel mesh */
  real*	array,	/* array of generated values				*/
  int	dim	/* dimensionality of grid				*/
  );

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine gen_real
** gen_real reads or generates floating point global array.
** Used to read or generate nodes, displacement bcs, or 
** velocity bcs.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of gen_real **********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   gen_isopar_grid   ***************
*/

/*  INTERFACE:   */

void
gen_isopar_grid(
  real*	array,	/* array of generated values				*/
  int	dim	/* dimensionality of grid				*/
  );

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine gen_isopar_grid
** isopar_grid generates real data via isoparametric interpolation
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of gen_isopar_grid ***************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   gen_shape_grid   ***************
*/

/*  INTERFACE:   */

void
gen_shape_grid(
  real	r,	/* isopar coordinate					*/
  real	s,	/* isopar coordinate					*/
  real	t,	/* isopar coordinate					*/
  real*	shape,	/* shape function storage				*/
  int	option	/* generation option flag				*/
  );

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine gen_shape_grid
** gen_shape_grid finds shape function for isoparametric generation.
** It appears to JWP to use bilinear and trilinear interpolation.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of gen_shape_grid ****************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   dot_sh   ***************
*/

/*  INTERFACE:   */

real  dot_sh(
  real*	shape,		/* shape function storage			*/
  real*	variable,	/* interpolate variable storage			*/
  int	n		/* number to sum over				*/
  );

/*  RETURN VALUE:   dot product value */
/*  DESCRIPTION:   */
/*
** Function dot_sh
** dot_sh performs the dot product of the vector of shape functions
** (may be gradients of shape functions)
** with the vector "variable"
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of dot_sh ************************************/
