/*
***                           File inphase.c                      ***
***                        GeoFEST version 4.5
*** Copyright (c) 2003, California Institute of Technology        ***
*** U.S.Sponsorship under NASA Contract NAS7-1260 is acknowledged ***
***
*** This software is designated for public release under JPL Task ***
*** Plan Number 83-6791 and may be publicly released after        ***
*** approval for release has been granted by GSFC Software        ***
*** Release  Authority
***
*** This file contains subroutines which handle input, 
*** equation numbering, and memory allocation.
***
***          == input_phase ==
***          == matrix_alloc ==
***          == gen_number ==
***          == output_phase ==
***          == el_output ==
***          == locate_pt ==
***          == save_refine ==
***/

/*
***GeoFEST version 4.5***
G. A. Lyzenga -- unified parallel/serial version and bug fixes 11/17/03
G. A. Lyzenga -- split node enhancements 8/16/02
G. A. Lyzenga -- modification to correctly relax hydrostatic stresses 6/15/00
J. W. Parker -- 3D support and deleted hypercube branches 9/16/98
G. A. Lyzenga  -- last modified on 3/03/95
Modification history:
2.1: fixes bug in boundary velocities and changes
     the way stresses are output.
2.2: implements split node faults
2.3: adds pass-thru comments
3.0: 3-D tetrahedra supported, hypercube branches deleted
3.1: hydrostatic relaxation changes
4.0: name change from visco to GeoFEST; correct implementation of 2-D out-of-
     plane stresses as well as hydrostatic relaxation; culling of remaining
     vestiges of hypercube stuff; one code without recompilation for both
     2-D and 3-D domains
4.1: Introduction of trilinear hexahedral bricks and PCG iterative solver option
4.2: Bug fixes; standard source prologues added (June 02)
4.3: Corrects inefficiency in split node algorithm and introduces new
     split node input format
4.4: Adds dummy calls to MPI and Pyramid to provide a single unified
     source for parallel and serial codes.  Also introduces a major
     algorithmic enhancement; stress equilibrium maintenance during visco
     elastic stepping for long-term stability.  Also more minor bug fixes.
4.5: Supports surface tractions and roller/spring duo options for tet elements.
     Provides enhanced conjugate gradient convergence monitoring.
 $Log: inphase.c,v $
 Revision 1.4  2003/02/09 06:10:55  lyzenga
 *** empty log message ***

*/

/*
#include <stdio.h>
#include "mydefs.h"
#include "errcodes.h"
#define EXTERN extern
#include "finel.h"
*/

#define EXTERN extern
#include "inphase.h"
#include "utility.h"
#include "output_phase.h"

void completion(void);
void gen_map(pamr_Mesh m[]);
void
gen_real(
         pamr_Mesh m[], /* parallel mesh                */
         real*	array,	/* array of generated values	*/
         int	dim	/* dimensionality of grid	*/
        );
void
elgrp_loop(
           pamr_Mesh m[], /* parallel mesh */
           int	task	/* which element-level task to perform */
          );
void vrestart(pamr_Mesh m[]);
void
profile_diag(
	 PROFILE*	a,	/* ptr to assembled profile stiffness storage   */
	 int	profile_case	/* old flag for substructuring...		*/
            );
void
shape(
  ELEMENT_DATA*	el_pt,	/* ptr to current element struct		*/
  ELEMENT_INFO*	info,	/* ptr to element info struct			*/
  real*		sh_pt,	/* el storage to hold shape functions		*/
  real*		det_pt,	/* el storage to hold determinant		*/
  int		type,	/* element type number				*/
  int		nen,	/* number of element nodes			*/
  int		nint,	/* number of Gauss integration points		*/
  int		all	/* flag indicating we need derivatives also	*/
      );
      
real   dotsh() ;

static real
   sh_temp[256] , det_temp[9] ;

char  line[80] ;


/**********************************************************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   input_phase   ***************
*/

/*  INTERFACE:   */

void
input_phase(pamr_Mesh m[])

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine input_phase
** input_phase is the top-level input file reading function.  
** Most input-file records are read here or in routines 
** in generat.c that are called by this function.
** A few items are read by main.c as well.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

    {
     char CVS_Id[] = "@(#) CVS id $Id: inphase.c,v 1.4 2003/02/09 06:10:55 lyzenga Exp $, Revision: " REVNO;
     COUNTER  i , n , onoff ;
     int   numnp , ndof , nrates , nsd ; 
     char  simtask[10] ;
     
/* *************************** SIM CONTROL CODES ***************** */
/* user control codes: (1 or 0)
   [0] ELASTIC1
   [1] ELAS_OUT1
   [2] REFINE
   [3] REFINE_OUT
   [4] ELASTIC2
   [5] ELAS_OUT2
   [6] VISCO
*/

for(i=0; i<7; i++)
   {
    fscanf(in_file,"%s%d",simtask,&onoff) ;
    if(0 == strcmp(simtask, "ELASTIC1"))
       flow_sim[0]=onoff ;
    else if(0 == strcmp(simtask, "ELAS_OUT1"))
       flow_sim[1]=onoff ;
    else if(0 == strcmp(simtask, "REFINE"))
       flow_sim[2]=onoff ;
    else if(0 == strcmp(simtask, "REFINE_OUT"))
       flow_sim[3]=onoff ;
    else if(0 == strcmp(simtask, "ELASTIC2"))
       flow_sim[4]=onoff ;
    else if(0 == strcmp(simtask, "ELAS_OUT2"))
       flow_sim[5]=onoff ;
    else if(0 == strcmp(simtask, "VISCO"))
       flow_sim[6]=onoff ;
    else 
       {attempt = FLOW_CODES; completion();}
   }

/* *************************** GLOBAL INPUT PHASE ***************** */

     fscanf(in_file,"%d",&fe_sys.numnp) ;
     fscanf(in_file,"%d",&fe_sys.nsd) ;
     fscanf(in_file,"%d",&fe_sys.ndof) ;
     fscanf(in_file,"%d",&fe_sys.nrates) ;
     fscanf(in_file,"%d",&fe_sys.save_shape) ;
     fscanf(in_file,"%d",&fe_sys.solver) ;

     if (pamr_my_pid() == 0)
	printf("Number of nodes = %d\n\n",fe_sys.numnp) ;

     if (pamr_my_pid() == 0)
	printf(
     "nsd = %d  ;  ndof = %d  ;  nrates = %d  ;  save_shape = %d  ;  solver = %d\n\n",
      fe_sys.nsd , fe_sys.ndof , fe_sys.nrates , fe_sys.save_shape , fe_sys.solver ) ;

     if(pamr_num_procs() != 1)
        loc_sys.numnp = pamr_local_node_count(m);
     else
        loc_sys.numnp = fe_sys.numnp ;

     numnp = loc_sys.numnp;
     nsd = fe_sys.nsd ;
     ndof    = fe_sys.ndof    ;
     nrates  = fe_sys.nrates  ;
     
     global.coords = (real * ) calloc(numnp*nsd,sizeof(real)) ;
     if (global.coords == NULL) attempt = CRD_MEM ;
     completion() ;
     
     global.displ = (real * ) calloc(numnp*ndof,sizeof(real)) ;
     if (global.displ == NULL) attempt = CRD_MEM ;
     completion() ;
     
     global.del_displ = (real * ) calloc(numnp*ndof,sizeof(real)) ;
     if (global.del_displ == NULL) attempt = CRD_MEM ;
     completion() ;
     
     global.forv = (real * ) calloc(numnp*ndof,sizeof(real)) ;
     if (global.forv == NULL) attempt = CRD_MEM ;
     completion() ;
     
     global.eqtype = (int * ) calloc(numnp*ndof,sizeof(int)) ;
     if (global.eqtype == NULL) attempt = CRD_MEM ;
     completion() ;
     
     global.splitn = (real * ) calloc(numnp*NSPLITATTR,sizeof(real)) ;
     if (global.splitn == NULL) attempt = CRD_MEM ;
     completion() ;
     
     if( nrates > 0 )
        {
         global.bctime = (real * ) calloc(nrates,sizeof(real)) ;
         global.rate = (real * ) calloc(numnp*ndof*nrates,
                                         sizeof(real)) ;
         if (global.rate == NULL) attempt = CRD_MEM ;
        }
     completion() ;
     
/* smoothing is disabled --
     if( fe_sys.psmooth > 0 )
        {
         global.pnode = (real * ) calloc(numnp,sizeof(real)) ;
         global.yp = (real * ) calloc(numnp,sizeof(real)) ;
         if (global.yp == NULL) attempt = PSM_MEM ;
        }
     completion() ;
*/

     gen_map(m) ;  /*  fill the decomp map array  */
     squawk("Global b.c. map generated...\n") ;
         
     gen_number(m) ;  /* obtain local node and equation numbering */
        
     squawk("Equation numbering complete...\n") ; 
    
     gen_real( m,global.coords,nsd ) ;  /*  fill the coord array  */
     squawk("Coordinate generation complete...\n") ;
     
     gen_real( m,global.forv,ndof ) ; /*  fill the b. c. array  */
     squawk("Displacement boundary value generation complete...\n") ;
     
     for( n=0 ; n<nrates ; n++ )
        {
         fscanf(in_file,"%lf",global.bctime+n) ;
	 if (pamr_my_pid() == 0) {
	    printf("Reading in boundary velocity set #%d for time = %g...\n",
		n+1 , *(global.bctime+n) ) ;
	    fflush(stdout) ;
	    }
         gen_real( m,(global.rate + n*ndof*numnp),ndof ) ;
         /*  fill the velocity arrays  */
        }
     squawk("Velocity boundary rates generation complete...\n") ;
     
/* ************************ ELEMENT INPUT PHASE ***************** */

     fscanf(in_file,"%d",&global.n_group) ;
     if(global.n_group > MAX_GROUP)
         {
          squawk("Too many element groups\n") ;
          exit(1) ;
         }

     for(i=0 ; i<global.n_group ; i++)
         {
          groups[i].el_info =
             (ELEMENT_INFO *) calloc(1,sizeof(ELEMENT_INFO)) ;
          if (groups[i].el_info == NULL) attempt = ELI_MEM + i ;
          completion() ;

          groups[i].el_mat =
             (ELEMENT_MAT *) calloc(1,sizeof(ELEMENT_MAT)) ;
          if (groups[i].el_mat == NULL) attempt = MAT_MEM + i ;
          completion() ;
          
          groups[i].group_num = i+1 ;
         }
                  
     squawk("Starting element generation...\n") ;

     elgrp_loop( m, GENERATE ) ;   /* read in and generate element geometry */

     squawk("Doing shape functions...\n") ;

     elgrp_loop( m, SHAPE ) ;    /* calculate element shape functions
                                 for future use */

     squawk("Element generation complete.  Setting up linear algebra...\n") ;
     
/* ************************ EQUATION SETUP PHASE ***************** */

     matrix_alloc(m) ;      /*  set up storage and indexing for the 
                              skyline profile elimination solver OR
                              the preconditioned CG solver  */


     squawk("Equation setup completed.\n") ;

/* ************************ TIME INPUT PHASE ***************** */

     time_data.time = ZERO ;
     time_data.loc_nprt_nodes = ZERO ;
     time_data.loc_nprt_elem = ZERO ;
     
     fscanf(in_file,"%d%d",&time_data.nprt_nodes,&time_data.nprt_elem) ;

     if(time_data.nprt_nodes > 0)
        {
         time_data.node_list
            = ( int  * ) calloc(time_data.nprt_nodes,sizeof(int)) ;
         if (time_data.node_list == NULL) attempt = TGP_MEM ;
         completion() ;

         for( i = 0 ; i < time_data.nprt_nodes ; i++ )
            fscanf(in_file,"%d", time_data.node_list + i ) ;

/*   commenting this out -- going back to old (global) numbering of output
         for( i = 0 ; i < time_data.nprt_nodes ; i++ ) {
            int node_id;
            fscanf(in_file,"%d", &node_id ) ;
            if (pamr_my_node(m, node_id)) {
                time_data.node_list[time_data.loc_nprt_nodes] = pamr_my_node_indx(m, node_id);
                time_data.loc_nprt_nodes++;
            }
         }
*/
        }

     if(time_data.nprt_elem > 0)
        {
         time_data.elem_list
            = ( int  * ) calloc(2*time_data.nprt_elem,sizeof(int)) ;
         if (time_data.elem_list == NULL) attempt = TGP_MEM ;
         completion() ;

         for( i = 0 ; i < time_data.nprt_elem ; i++ )
            fscanf(in_file,"%d%d", time_data.elem_list + 2*i
                   , time_data.elem_list + 2*i + 1 ) ;

                                    /* group # and elem. # */

/*   commenting this out -- going back to old (global) numbering of output
         for( i = 0 ; i < time_data.nprt_elem ; i++ ) {
            int group_id, elem_id;
            fscanf(in_file,"%d%d", group_id, elem_id);
 
            if (pamr_my_element(m, elem_id)) {
                time_data.elem_list[2*time_data.loc_nprt_elem] = group_id ;
                time_data.elem_list[2*time_data.loc_nprt_elem + 1] = pamr_my_element_indx(m, elem_id) ;
                time_data.loc_nprt_elem++;
            }
         }
*/
        }

/*     fscanf(in_file,"%d%d%d%lf",&time_data.ntime_grp,&time_data.nreform,
                           &time_data.nbackup,&time_data.slip_interval) ;
          change to input # of fault groups instead of single repeat time                 
*/

     fscanf(in_file,"%d%d%d%d",&time_data.ntime_grp,&time_data.nreform,
                           &time_data.nbackup,&time_data.nfgrps) ;
     
     if(time_data.ntime_grp == 0)  /* this is an elastic-only run */
        {
         time_data.nprints = 1 ;
         time_data.prt_time
             = ( real  * ) calloc(time_data.nprints,sizeof(real)) ;
         *(time_data.prt_time) = ZERO ;
         fltgrp_ptr = (FLTGRP *) calloc(time_data.nfgrps,sizeof(FLTGRP)) ;
         squawk("No time data  generation (elastic only)...\n") ;
         return ;
        }
     
     /* the rest of the input required only for time-stepping runs... */
     
     time_data.maxstep
        = ( real  * ) calloc(time_data.ntime_grp,sizeof(real)) ;
     if (time_data.maxstep == NULL) attempt = TGP_MEM ;
     time_data.endtime
        = ( real  * ) calloc(time_data.ntime_grp,sizeof(real)) ;
     if (time_data.endtime == NULL) attempt = TGP_MEM ;
     time_data.nsteps
        = ( int  * ) calloc(time_data.ntime_grp,sizeof(int)) ;
     if (time_data.nsteps == NULL) attempt = TGP_MEM ;
     completion() ;
                 
     time_data.alpha
         = ( real  * ) calloc(time_data.ntime_grp,sizeof(real)) ;
     if (time_data.alpha == NULL) attempt = TGP_MEM ;
     completion() ;
                 
     time_data.delt 
         = ( real  * ) calloc(time_data.ntime_grp,sizeof(real)) ;
     if (time_data.delt == NULL) attempt = TGP_MEM ;
     completion() ;
                 
     for( i = 0 ; i < time_data.ntime_grp ; i++ )
        fscanf(in_file,"%lf%lf%lf",
             time_data.endtime + i, time_data.alpha + i , time_data.delt + i ) ;

     fscanf(in_file,"%d", &time_data.nprints ) ;
     if( time_data.nprints > 0 )
         {
          time_data.prt_time
             = ( real  * ) calloc(time_data.nprints,sizeof(real)) ;
          if (time_data.prt_time == NULL) attempt = TGP_MEM ;

          for( i = 0 ; i < time_data.nprints ; i++ )
            fscanf(in_file,"%lf", time_data.prt_time + i ) ;
/*	completion() ;	*/
         }
         
 /* now comes the input block for split node slip history groups...*/
 
     fltgrp_ptr = (FLTGRP *) calloc(time_data.nfgrps,sizeof(FLTGRP)) ;
     for(i=0 ; i<time_data.nfgrps ; i++)
        {
         fscanf(in_file,"%d",&n) ;
         fscanf(in_file,"%lf",&((fltgrp_ptr+n-1)->interval) ) ;
         fscanf(in_file,"%lf",&((fltgrp_ptr+n-1)->first_event) ) ;
        }

     squawk("Time data  generation complete...\n") ;
     
     time_data.restart = 0 ;
     fscanf(in_file,"%s",line) ;
     if(strcmp(line , "NO_RESTART") != 0 )
        {
         sprintf(time_data.startfile,"%s",line) ;
         time_data.restart = 1 ;
         vrestart(m) ;
        }

     time_data.sav_state = 0 ;
     fscanf(in_file,"%s",line) ;
     if(strcmp(line , "NO_SAVE") != 0 )
        {
         sprintf(time_data.savefile,"%s",line) ;
         time_data.sav_state = 1 ;
        }

    } /* end function */

/************************** end of input_phase *******************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   matrix_alloc   ***************
*/

/*  INTERFACE:   */

void
matrix_alloc(pamr_Mesh m[])

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine matrix_alloc
** allocates space for the sparse stiffness matrix _or_
** the PCG solver arrays
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

   {
    int   i ;
    
    switch(fe_sys.solver)
       {
        case DIRECT:
         stiff.neq = fe_sys.neq = loc_sys.neq ; /* only valid for single processor! */

         stiff.diag = (int  * ) calloc(fe_sys.neq,sizeof(int )) ;
         if (stiff.diag == NULL) attempt = DIAG_MEM ;
         completion() ;
         
         elgrp_loop( m, REORDER );
         elgrp_loop( m, COL_HT ) ;
         profile_diag( &stiff , INTERIOR ) ;
         
         printf("Total profile size is %d\n",stiff.size);

         stiff.matrix =
            (real * ) calloc(stiff.size,sizeof(real)) ;
         if (stiff.matrix == NULL) attempt = PROF_MEM ;
         completion() ;

         break;
/*---------------------------------------------------------------*/
         
        case PCG:
         pcg.neq = loc_sys.neq ;

         pcg.precond = (real  * ) calloc(loc_sys.neq,sizeof(real)) ;
         if (pcg.precond == NULL) attempt = PCG_MEM ;

         pcg.r = (real  * ) calloc(loc_sys.neq,sizeof(real)) ;
         if (pcg.r == NULL) attempt = PCG_MEM ;

         pcg.d = (real  * ) calloc(loc_sys.neq,sizeof(real)) ;
         if (pcg.d == NULL) attempt = PCG_MEM ;

         pcg.temp = (real  * ) calloc(loc_sys.neq,sizeof(real)) ;
         if (pcg.temp == NULL) attempt = PCG_MEM ;

         completion() ;
         
         pcg.hist = (real  * ) calloc(CGMAX,sizeof(real)) ;
/*
         char histname[80];
         sprintf(histname,"/data2/nortonc/src/GeoFEST/geofest/cghist.txt%d",pamr_my_pid()) ;
*/
	 /*         sprintf(histname,"../cghist.txt%d",pamr_my_pid()) ; */

	 /*         if((cgconv_file = fopen("cghist.txt","w")) == NULL) */
/*
         if((cgconv_file = fopen(histname,"w")) == NULL)
          {
           printf("Convergence history file couldn't be opened!\n") ;
           exit(1) ;
          }
*/
         global.permut = (int *) calloc(loc_sys.numnp+1,sizeof(int));
         for(i=0;i<=loc_sys.numnp;i++)
           global.permut[i] = i ;  /* null permutation */

         force.last_result = (real * ) calloc(loc_sys.neq,sizeof(real)) ;
         if (force.last_result == NULL) attempt = R_MEM ;

         force.first_result = (real * ) calloc(loc_sys.neq,sizeof(real)) ;
         if (force.first_result == NULL) attempt = R_MEM ;

         break;
       }
       
         force.full_rhs = (real * ) calloc(loc_sys.neq,sizeof(real)) ;
         if (force.full_rhs == NULL) attempt = R_MEM ;
         force.net_external = (real * ) calloc(loc_sys.neq,sizeof(real)) ;
         if (force.net_external == NULL) attempt = R_MEM ;
         completion() ;

   }

/************************** end of matrix_alloc ******************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   gen_number   ***************
*/

/*  INTERFACE:   */

void
gen_number(pamr_Mesh m[])

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine gen_number
** gen_number assigns equation numbers based on nodes and ndof
** Note: Constructed inverse id_pointer array to find global
**       1-based node numbers from equation numbers (in dot_par)
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

   {
    COUNTER   i , j , k , index ;
    int   eqtype[3] ;
    int   ieq_local, ieq_owned ;
    int   *id_pointer ;
    int   neq_local , ndof , numnp ;

     neq_local = 0 ;
     numnp = loc_sys.numnp ;
     ndof = fe_sys.ndof ; 

   
     for( i=0 ; i< numnp ; i++ )
         {
          index = i ;

          for(k=0 ; k<ndof ; k++)
             {
              eqtype[k] = *(global.eqtype+ndof*index+k) ;
             }
                      /* 1 = active , 0 = boundary  */

               for(k=0 ; k<ndof ; k++)
                  if(eqtype[k])  neq_local++ ;
              
         }   /* end loop over nodes */

     loc_sys.neq = neq_local ;
     squawk("Decomposed node and equation totals:\n") ;
     completion() ;
     printf ("IAM: %d\t local numnp = %d  local neq = %d\n",
	pamr_my_pid() , loc_sys.numnp, loc_sys.neq) ;
     fflush(stdout) ;
/*
     printf ("numnp = %d  neq = %d\n", loc_sys.numnp, loc_sys.neq) ;
     fflush(stdout) ;
*/
     
  /* we now will assign equation numbers */
  
  /* first, allocate space for the id array  */
  /* second, allocate space for the inverse id array  */
     global.id_pointer =
        (int * ) calloc(numnp*ndof,sizeof(int)) ;
     if (global.id_pointer == NULL) attempt = ID_MEM ;
     global.owned_eqs =
        (int * ) calloc(loc_sys.neq,sizeof(int)) ;
     if (global.owned_eqs == NULL) attempt = ID_MEM ;
     completion() ;

     id_pointer = global.id_pointer ;
     
  /* now assign equations ...    (equations start at zero)  */
   
     ieq_local = 0 ;
     ieq_owned = 0 ;
     
     for(j=0 ; j<numnp*ndof ; j++) *(id_pointer + j) = (-1) ;
     
     for( i=0 ; i<numnp ; i++ )
         {
          index = pamr_my_node_gid( m, i+1) ; /* return global index */

          for(k=0 ; k<ndof ; k++)
             {
              eqtype[k] = *(global.eqtype+ndof*i+k) ;
             }

          for(k=0 ; k<ndof ; k++)
             {
              if(eqtype[k])
                 {
                  *(id_pointer + ndof*(i) + k) =
                     ieq_local ;
                  if (!pamr_my_partition_node( m, index ) ||
                       pamr_my_node_leader(m, index)) {
                       global.owned_eqs[ieq_owned] = ieq_local;
                       ieq_owned++ ;
                  }
                  ieq_local++ ;
                 }
             }
         }   /* end loop over global node number */
         
     loc_sys.neq_owned = ieq_owned;

   }

/************************** end of gen_number ********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   output_phase   ***************
*/

/*  INTERFACE:   */

   void
   output_phase(
                pamr_Mesh m[], /* parallel mesh */
                int    quake   /* flag to indicate quake or not */
               )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine output_phase
** output_phase writes out requested data at a quake 
** or scheduled time.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

    {
     COUNTER  i , n , k , ntot , index , ndof , numnp , numnp_all , nsd ;
     int  mine , print , curr_node , count , prt_node ;
     static real  outbuf[9] ;
     real  time , test ;
     MPI_Status  status ;
         
     int owned_nodes, out_num, tot_num, j, s_size, tmp_sum, indx, global_node;
     int *rcounts, *displs;
     OUT_NODE_DATA *s_data, *r_data;

     ndof = fe_sys.ndof ;
     numnp = loc_sys.numnp ;
     numnp_all = fe_sys.numnp ;
     nsd = fe_sys.nsd ;
     time = time_data.time ;
     
     print = 0 ;
     for(n=0 ; n<time_data.nprints ; n++) {
         test = *(time_data.prt_time + n) ;
         if( test >= ZERO && test <= time ) {
             *(time_data.prt_time + n) = (-ONE) ;
             print = 1 ;
             break ;
         }
     }

     if( !(print || quake) )
         return ;

     if (pamr_my_pid() == 0) {
	if(quake)
	    printf("Performing quake output at time = %f ...\n",time) ;
	else
	    printf("Performing scheduled output at time = %f ...\n",time) ;
     }

     if(time_data.nprt_nodes == -1)
        for (global_node = 1, owned_nodes = 0; global_node <= numnp_all; global_node++)
           {
            if (pamr_my_owned_node(m,global_node))
              owned_nodes++ ;
           }
     else if(time_data.nprt_nodes > 0)
        for (n = 0, owned_nodes = 0; n < time_data.nprt_nodes; n++)
           {
            global_node = *(time_data.node_list + n) ;
            if (pamr_my_owned_node(m,global_node))
              owned_nodes++ ;
           }
     else
        owned_nodes = 0 ;

  /* Allocate buffers for MPI_GATHERV() */

     rcounts = calloc(pamr_num_procs(),sizeof(int));
     displs  = calloc(pamr_num_procs(),sizeof(int));

  /* Gather number of owned nodes from each processor... */

     out_num = owned_nodes;
     MPI_Gather(&out_num, 1, MPI_INT, rcounts, 1, MPI_INT, 0, MPI_COMM_WORLD);

  /* Compute total number of owned nodes and displacements for MPI_Gatherv() */

     for (i = 0, tot_num = 0; i < pamr_num_procs(); i++)
          tot_num = tot_num + rcounts[i];
     displs[0] = 0; j = 1;
     for (i = 1; i < pamr_num_procs(); i++) {
          for(k = 0, tmp_sum = 0; k < j; k++)
              tmp_sum = tmp_sum + rcounts[k];
          displs[i] = tmp_sum;
          j = j + 1;
     }

  /* Allocate transport buffers for MPI_Gatherv() and initialize */

     s_data  = calloc(owned_nodes, sizeof(OUT_NODE_DATA));     /* Owned Nodes */

     if (pamr_my_pid() == 0 && pamr_num_procs() != 1)
         r_data = calloc(tot_num, sizeof(OUT_NODE_DATA));
     else
         r_data = NULL ;

     if (pamr_my_pid() == 0) {
	fprintf(out_file,
	"\n Global coordinates & displacements & delt displacements \n") ;
	fprintf(out_file,"\n Simulation time = %f ; step size = %g\n",
	    time,time_data.dt) ;
	if(quake)
	  fprintf(out_file,"***Fault Rupture*** at t=%g\n",time) ;
	fprintf(out_file,
 "\n          coordinates          displacements       del_displacements\n") ;
     }
     
      if (time_data.nprt_nodes == (-1))
	      ntot = numnp_all ;
      else
          ntot = time_data.nprt_nodes ;
             
      for(i=0, indx = 0 ; i<ntot ; i++) {
	/*	  completion() ; */
          mine = 0 ;
          if(time_data.nprt_nodes == (-1))
             curr_node = i+1 ;
          else
             curr_node = *(time_data.node_list+i) ;

/* test to see if this node is in the local processor... */

	  mine = pamr_my_owned_node(m,curr_node) ;

          if(mine) { /* the node is in my processor */
	     index = pamr_my_node_indx(m,curr_node) - 1 ;
	     prt_node = curr_node ;

	     out_node_data_init(&s_data[indx], curr_node, 1,
				(global.coords + nsd*index),
				(global.displ + ndof*index),
				(global.del_displ + ndof*index)) ;
	     ++indx;
	  }

      }        

  /* Continue initialization for MPI_Gatherv() */

     s_size = out_num * sizeof(OUT_NODE_DATA);

     for (i = 0; i < pamr_num_procs(); i++) {
          displs[i]  = displs[i]  * sizeof(OUT_NODE_DATA);
          rcounts[i] = rcounts[i] * sizeof(OUT_NODE_DATA);
     }

/*
     printf("IAM: %d\t OwnedNodes: %d\t Total: %d\n",pamr_my_pid(), owned_nodes, tot_num);
     if (pamr_my_pid() == 2)
              out_node_data_fprintf_list( out_file, r_data, tot_num );
              out_node_data_fprintf_list( stdout, s_data, owned_nodes );
     pamr_finalize(true); exit(0);

     for(i = 0; i < pamr_num_procs(); i++) {
       printf("IAM: %d\t Rcounts: %d\t Displ: %d", pamr_my_pid(), rcounts[i], displs[i]);
     }
     printf("\n");
     pamr_finalize(true); exit(0);
*/

     if(pamr_num_procs() != 1)
        MPI_Gatherv( (void *)s_data, s_size, MPI_BYTE, (void *)r_data, rcounts, displs,
                  MPI_BYTE, 0, MPI_COMM_WORLD );
     else
        r_data = s_data ;

/*
     printf("IAM: %d\t OwnedNodes: %d\t Total: %d\n",pamr_my_pid(), owned_nodes, tot_num);
     if (pamr_my_pid() == 0)
         out_node_data_fprintf_list( out_file, r_data, tot_num );
         out_node_data_fprintf_list( stdout, r_data, tot_num );
     pamr_finalize(true); exit(0);
*/

     if (pamr_my_pid() == 0) {
         out_node_data_fprintf_list( out_file, r_data, tot_num );
         fflush(out_file) ;
     }

     out_node_data_destroy_list( s_data, owned_nodes );
     if (pamr_my_pid() == 0)
         out_node_data_destroy_list( r_data, tot_num );

     free( s_data );
     if (pamr_my_pid() == 0 && pamr_num_procs() != 1)  free( r_data );
     free( rcounts );
     free( displs );

     elgrp_loop(m, OUTPUT) ;
     squawk("Output step finished.\n") ;

    }

/************************** end of output_phase ******************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   el_output   ***************
*/

/*  INTERFACE:   */

      void
      el_output(
		pamr_Mesh	m[],	/* parallel mesh */
                GROUP		*grp_ptr/* pointer to current element group */
               )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine el_output
** el_output prints out element stress information.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
      int g_numel , nint , nel , use_list , n , str_size , inode1 , inode2 ,
	  ipt , j , ng , ntot , nsd , mine , prt_n , count , index ;
      real   xpt[3][10] , str_ctr[9] ;
      ELEMENT_DATA   *el_pt ;
      MPI_Status  status ;

      int owned_nodes, out_num, tot_num, i, k, s_size, tmp_sum, indx ;
      int *rcounts, *displs;
      OUT_ELEMENT_DATA *s_data, *r_data;

      /* Allocate buffers for MPI_GATHERV() */

      rcounts = calloc(pamr_num_procs(),sizeof(int));
      displs  = calloc(pamr_num_procs(),sizeof(int));

      /* Gather number of owned elements from each processor... */

      if(pamr_num_procs() > 1)
         {
          if(time_data.nprt_elem == -1)
            out_num = pamr_local_element_count(m);
          else if(time_data.nprt_elem > 0)
             {
              for (n = 0, out_num = 0; n < time_data.nprt_elem; n++)
                 {
                  if(pamr_my_element(m , *(time_data.elem_list + 2*n + 1)))
                     out_num++ ;
                 }
             }
          else
             out_num = 0 ;
         }
      else /* single processor */
         {
          if(time_data.nprt_elem == -1)
             out_num =  (grp_ptr->el_info)->numel ;
          else
             out_num = time_data.nprt_elem ;
         }
      
      
      MPI_Gather(&out_num, 1, MPI_INT, rcounts, 1, MPI_INT, 0, MPI_COMM_WORLD);

      /* Compute total number of owned elements and displacements for MPI_Gatherv() */

      for (i = 0, tot_num = 0; i < pamr_num_procs(); i++)
   	   tot_num = tot_num + rcounts[i];
      displs[0] = 0; j = 1;
      for (i = 1; i < pamr_num_procs(); i++) {
	   for(k = 0, tmp_sum = 0; k < j; k++)
	       tmp_sum = tmp_sum + rcounts[k];
	   displs[i] = tmp_sum;
	   j = j + 1;
      }

      /* Allocate transport buffers for MPI_Gatherv() and initialize */

      s_data  = calloc(out_num, sizeof(OUT_ELEMENT_DATA));

      if (pamr_my_pid() == 0 && pamr_num_procs() != 1)
          r_data  = calloc(tot_num, sizeof(OUT_ELEMENT_DATA));
      else
          r_data = NULL ;

      nsd = fe_sys.nsd ;
      g_numel = (grp_ptr->el_info)->g_numel ;
      str_size = (grp_ptr->el_info)->stress_size ;

     if (pamr_my_pid() == 0)
	{
	 fprintf(out_file,
	   "\n Element stresses  -- element group #%d \n",grp_ptr->group_num) ;
	 fprintf(out_file,"\n Simulation time = %f ; step size = %g\n",
	   time_data.time,time_data.dt) ;
	 fprintf(out_file,
      "\n coordinates                          stresses . . .->      el.#\n") ;
     if(str_size == 6)
	 fprintf(out_file,
      "   X             Y            Z       sig_XX        sig_YY      sig_ZZ      sig_XY        sig_XZ      sig_YZ\n") ;
     
     else if(str_size == 4)
	 fprintf(out_file,
      "   X             Y       sig_XX        sig_YY      sig_XY        sig_ZZ\n") ;
     
	fflush(out_file) ;
	}

      
      switch((grp_ptr->el_info)->type)
          {
           case BILIN:
           case SERENDIP:
           case TET:
           case TRILIN:
               if(time_data.nprt_elem == (-1))
                  {
                   ntot = g_numel ;
                   use_list = 0 ;
                  }
               else
                  {
                   ntot = time_data.nprt_elem ;
                   use_list = 1 ;
                  }
               for ( nel = 0, indx = 0 ; nel < ntot ; nel++)
                   {
                    if( use_list )
                       {
                        if( *(time_data.elem_list + 2*nel) !=
                            grp_ptr->group_num ) continue ;
                        ng = *(time_data.elem_list + 2*nel + 1) ;
                        n = ng ;
                       }
                    else
                       {
                        ng = nel+1 ;
                        n = ng ;
                       }
                    if( n == 0 ) continue ;
                    
/* test to see if this element is in the local processor... */

                  mine = pamr_my_element(m,n) ;
                  
                  if(mine)  /* the element is in my processor */
                     {
                      index = pamr_my_element_indx(m,n) - 1 ;
                      prt_n = n ;

                      el_pt = grp_ptr->el_data + index ;

                      nint = el_pt->nint ;
                      
                      if((grp_ptr->el_info)->type == TET && el_pt->degen == 1)
                           /* this is a "duo" element - no integration points  */
                         {
                          for(j=0 ; j<str_size ; j++)
                             str_ctr[j] = *(el_pt -> stress + j) ;
                             
                          inode1 = *(el_pt->ien    ) - 1 ;  /* zero-based node index */
                          inode2 = *(el_pt->ien + 1) - 1 ;
                          for(j=0 ; j < nsd  ; j++)
                             str_ctr[str_size+j] =  PT5 * (*(global.coords+inode1*nsd+j) +
                                       *(global.coords+inode2*nsd+j) );
   	   	                     out_element_data_init(&s_data[indx], prt_n, 1, str_size,
                                            &str_ctr[str_size],
                                            &str_ctr[0]);
                          ++indx;
                          continue ;
                         }
                      
                      locate_pt( el_pt , grp_ptr->el_info , xpt ) ;

                      for(j=0 ; j<str_size ; j++)
                          str_ctr[j] = ZERO ;  /* midpoint stresses */

                      for(j=0 ; j < nsd ; j++)
                          xpt[j][9] = ZERO ;  /* midpoint location */

                      for (ipt = 0 ; ipt < nint ; ipt++)
                        {
                           for(j=0; j < nsd; j++)
                               xpt[j][9] += xpt[j][ipt] / ((real) nint) ;

                           for(j=0 ; j<str_size ; j++)
                               str_ctr[j] +=
                               *((el_pt -> stress) + ipt*str_size + j)  / ((real) nint) ;
                        }

                      for(j=0 ; j < nsd; j++)
   		          str_ctr[str_size+j] = xpt[j][9] ;
		      /* the last three slots contain the coords */
                     
   	   	      out_element_data_init(&s_data[indx], prt_n, 1, str_size,
                                            &str_ctr[str_size],
                                            &str_ctr[0]);
                      ++indx;
		     }

		}

  /* Continue initialization for MPI_Gatherv() */

     s_size = out_num * sizeof(OUT_ELEMENT_DATA);

     for (i = 0; i < pamr_num_procs(); i++) {
          displs[i]  = displs[i]  * sizeof(OUT_ELEMENT_DATA);
          rcounts[i] = rcounts[i] * sizeof(OUT_ELEMENT_DATA);
     }

     if(pamr_num_procs() != 1)
        MPI_Gatherv( (void *)s_data, s_size, MPI_BYTE, (void *)r_data, rcounts, displs,
                  MPI_BYTE, 0, MPI_COMM_WORLD );
     else
        r_data = s_data ;


     if (pamr_my_pid() == 0) {
         out_element_data_fprintf_list( out_file, r_data, tot_num , str_size );
         fflush(out_file) ;
     }

     out_element_data_destroy_list( s_data, owned_nodes );
     if (pamr_my_pid() == 0)
         out_element_data_destroy_list( r_data, tot_num );

     free( s_data );
     if (pamr_my_pid() == 0 && pamr_num_procs() != 1)  free( r_data );
     free( rcounts );
     free( displs );

          break;
          }
    }

/************************** end of el_output *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   locate_pt   ***************
*/

/*  INTERFACE:   */

      void
      locate_pt(
                ELEMENT_DATA  *el_pt  ,  /* ptr to current element struct */
                ELEMENT_INFO  *info ,   /* ptr to element info struct */
                real   xpt[3][10]     /* return array of computed coords */
               )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine locate_pt
** locate_pt maps the integration points into physical space.
** Global coordinates are placed in xpt[j]
** Called by el_output element stress output function (above).
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

  {
   int    j , ipt , save_shape , type , nint , nen , nsd , nsdp1 ;
   real  *shape_ptr , *det_ptr , *sh ;

   save_shape = fe_sys.save_shape ;
   nen = info->nen ;
   type = info->type ;
   nint = el_pt->nint ;
   nsd = fe_sys.nsd ;
   nsdp1 = nsd + 1 ;
   
   if( !save_shape )
        {
         shape_ptr = sh_temp ;
         det_ptr = det_temp ;
         shape( el_pt , info , shape_ptr , det_ptr , type ,
                nen , nint , 0 ) ;
        }
   else
        {
         shape_ptr = el_pt -> shape ;
        }

     for (ipt = 0 ; ipt < nint ; ipt++)
          {
           sh = shape_ptr + ipt*nen*nsdp1 ;

           for(j=0; j < nsd ; j++)
              xpt[j][ipt] = dotsh( global.coords ,
                  nsd , j , sh , el_pt , info , nen , nsd , 0 ) ;
          }

  }

/************************** end of locate_pt *********************************/ 

/*
   *************   ROUTINE:   save_refine   ***************
*/

/*  INTERFACE:   */

      void
      save_refine(
                  pamr_Mesh m[]
                 )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine save_refine
** writes out the current (refined) mesh in
** the form of a GeoFEST-readable input file
** for future use or adaptation
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
{
 GROUP		*grp_ptr; /* pointer to current element group */
 ELEMENT_DATA  *el_pt;
 BUOY_DATA      *buoy_ptr ;

 int i , j , k , nel ;
 
 newmesh_file = fopen("refined.dat", "w") ;
 fprintf(newmesh_file,"outfilenamehere.out\n") ;
 fprintf(newmesh_file,"Comment line #1 *\n") ;
 fprintf(newmesh_file,"Comment line #2 *\n") ;
 /* insert default control codes to run the elastic problem on the new mesh */
 fprintf(newmesh_file,"ELASTIC1  1\nELAS_OUT1  1\n") ;
 fprintf(newmesh_file,"REFINE  0\nREFINE_OUT  0\n") ;
 fprintf(newmesh_file,"ELASTIC2  0\nELAS_OUT2  0\nVISCO  0\n\n") ;

 fprintf(newmesh_file,"%d\n\n",fe_sys.numnp) ;
 fprintf(newmesh_file,"%d   %d   %d   %d   %d\n\n",fe_sys.nsd,fe_sys.ndof,
                        fe_sys.nrates,fe_sys.save_shape,fe_sys.solver) ;
 for(i=0;i<fe_sys.numnp;i++)
    {
     fprintf(newmesh_file,"%d   0     %d   %d   %d\n",i+1,
             global.eqtype[i*fe_sys.ndof], global.eqtype[i*fe_sys.ndof+1], 
             global.eqtype[i*fe_sys.ndof+2]) ;
    }
 fprintf(newmesh_file,"0  0\n\n") ;

 for(i=0;i<fe_sys.numnp;i++)
    {
     fprintf(newmesh_file,"%d   0     %f   %f   %f\n",i+1,
             global.coords[i*fe_sys.nsd], global.coords[i*fe_sys.nsd+1], 
             global.coords[i*fe_sys.nsd+2]) ;
    }
 fprintf(newmesh_file,"0  0\n\n") ;
 /* for now, we don't have refined forv boundary values... */
 fprintf(newmesh_file,"0  0\n\n\n") ;

  for(i=0 ; i<global.n_group ; i++)
	 {
	  fprintf(newmesh_file,"%d\n",i+1) ;
	  grp_ptr = &groups[i];
	  fprintf(newmesh_file,"%d\n",(grp_ptr->el_info)->g_numel) ;
	  fprintf(newmesh_file,"%d   %d   %d   %d   %d\n", (grp_ptr->el_info)->type,
	          (grp_ptr->el_info)->numat, (grp_ptr->el_info)->numsuf, 
	          (grp_ptr->el_info)->numbuoy, (grp_ptr->el_info)->nfnode ) ;
      for(j=0 ; j< (grp_ptr->el_info)->numat ; j++)
        {
	     fprintf(newmesh_file,"%f   %f   %f   %f      %f   %f   %f\n", 
	             *((grp_ptr->el_mat)->mu+j) , *((grp_ptr->el_mat)->lambda+j) ,
	             *((grp_ptr->el_mat)->visc+2*j) , *((grp_ptr->el_mat)->visc+2*j+1) ,
	             *((grp_ptr->el_mat)->bforce+fe_sys.ndof*j) , *((grp_ptr->el_mat)->bforce+fe_sys.ndof*j+1) ,
	             *((grp_ptr->el_mat)->bforce+fe_sys.ndof*j+2)) ;
	    }
	  fprintf(newmesh_file,"\n") ;
  
	  for(nel=0 ; nel < (grp_ptr->el_info)->g_numel ; nel++)
		 {
		  el_pt = grp_ptr->el_data + nel ;
		  fprintf(newmesh_file,"%d   0   %d       %d   %d   %d   %d\n",
		          nel+1 , (el_pt->mat)+1 , *(el_pt->ien) , *(el_pt->ien + 1) ,
		          *(el_pt->ien + 2) , *(el_pt->ien + 3)) ;
		 }
       fprintf(newmesh_file,"0  0\n\n") ;

      if((grp_ptr->el_info)->numsuf > 0)
         {
          for(j=0 ; j< (grp_ptr->el_info)->numsuf ; j++)
             {
              fprintf(newmesh_file,"%d   %d         %f   %f   %f\n", 
                      1 + *((grp_ptr->el_info)->surf_list) , *((grp_ptr->el_info)->surf_list+1) ,
                      *((grp_ptr->el_info)->surf_trac) , *((grp_ptr->el_info)->surf_trac+1) ,
                      *((grp_ptr->el_info)->surf_trac+2) ) ;
             }
         fprintf(newmesh_file,"0  0\n\n") ;
         }

      if((grp_ptr->el_info)->numbuoy > 0)
         {
          for(j=0 ; j< (grp_ptr->el_info)->numbuoy ; j++)
             {
              buoy_ptr = (grp_ptr->el_info)->buoy + j ;
              fprintf(newmesh_file,"%d     %f   %f   %f     %f\n", 
                      buoy_ptr->numface , buoy_ptr->upvec[0] , buoy_ptr->upvec[1] ,
                      buoy_ptr->upvec[2] , buoy_ptr->rho_g) ;
              for(k=0 ; k < buoy_ptr->numface ; k++)
                 {
                  fprintf(newmesh_file,"%d    %d\n", 
                          1 + *(buoy_ptr->buoy_list) , 
                          *(buoy_ptr->buoy_list+1)  ) ;
                 }
              fprintf(newmesh_file,"0  0\n\n") ;
             }

         }

      if((grp_ptr->el_info)->nfnode > 0)
         {
          for(j=0 ; j< (grp_ptr->el_info)->nfnode ; j++)
             {
             }
         }

	 }
          
 fclose(newmesh_file) ;
}
/************************** end of save_refine *********************************/ 

