/*
***                           File inphase.h                      ***
***                        GeoFEST version 4.5
*** Copyright (c) 2003, California Institute of Technology        ***
*** U.S.Sponsorship under NASA Contract NAS7-1260 is acknowledged ***
***
*** This software is designated for public release under JPL Task ***
*** Plan Number 83-6791 and may be publicly released after        ***
*** approval for release has been granted by GSFC Software        ***
*** Release  Authority
***
*** This file contains subroutines which handle input, 
*** equation numbering, and memory allocation.
***
***          == input_phase ==
***          == matrix_alloc ==
***          == gen_number ==
***          == output_phase ==
***          == el_output ==
***          == locate_pt ==
***          == save_refine ==
***/

/*
***GeoFEST version 4.5***
G. A. Lyzenga -- unified parallel/serial version and bug fixes 11/17/03
G. A. Lyzenga -- split node enhancements 8/16/02
G. A. Lyzenga -- modification to correctly relax hydrostatic stresses 6/15/00
J. W. Parker -- 3D support and deleted hypercube branches 9/16/98
G. A. Lyzenga  -- last modified on 3/03/95
Modification history:
2.1: fixes bug in boundary velocities and changes
     the way stresses are output.
2.2: implements split node faults
2.3: adds pass-thru comments
3.0: 3-D tetrahedra supported, hypercube branches deleted
3.1: hydrostatic relaxation changes
4.0: name change from visco to GeoFEST; correct implementation of 2-D out-of-
     plane stresses as well as hydrostatic relaxation; culling of remaining
     vestiges of hypercube stuff; one code without recompilation for both
     2-D and 3-D domains
4.1: Introduction of trilinear hexahedral bricks and PCG iterative solver option
4.2: Bug fixes; standard source prologues added (June 02)
4.3: Corrects inefficiency in split node algorithm and introduces new
     split node input format
4.4: Adds dummy calls to MPI and Pyramid to provide a single unified
     source for parallel and serial codes.  Also introduces a major
     algorithmic enhancement; stress equilibrium maintenance during visco
     elastic stepping for long-term stability.  Also more minor bug fixes.
4.5: Supports surface tractions and roller/spring duo options for tet elements.
     Provides enhanced conjugate gradient convergence monitoring.
 $Log: inphase.h,v $
 Revision 1.1  2003/02/09 06:10:55  lyzenga
 *** empty log message ***

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "mydefs.h"
#include "errcodes.h"
#include "finel.h"
#include <mpi.h>
#include "output_phase.h"

/**********************************************************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   input_phase   ***************
*/

/*  INTERFACE:   */

void
input_phase(pamr_Mesh m[]);

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine input_phase
** input_phase is the top-level input file reading function.
** Most input-file records are read here or in routines
** in generat.c that are called by this function.
** A few items are read by main.c as well.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of input_phase *******************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   matrix_alloc   ***************
*/

/*  INTERFACE:   */

void
matrix_alloc(pamr_Mesh m[]);

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine matrix_alloc
** matrix allocates space for the sparse stiffness matrix _or_
** the PCG solver arrays
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of matrix_alloc ******************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   gen_number   ***************
*/

/*  INTERFACE:   */

void
gen_number(pamr_Mesh m[]);

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine gen_number
** gen_number assigns equation numbers based on nodes and ndof
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of gen_number ********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   output_phase   ***************
*/

/*  INTERFACE:   */

void
output_phase(
  pamr_Mesh m[], /* parallel mesh */
  int	quake	/* flag to indicate quake or not			*/
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine output_phase
** output_phase writes out requested data at a quake
** or scheduled time.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of output_phase ******************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   el_output   ***************
*/

/*  INTERFACE:   */

void
el_output(
  pamr_Mesh	m[],	/* parallel mesh				*/
  GROUP*	grp_ptr	/* pointer to current element group		*/
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine el_output
** el_output prints out element stress information.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of el_output *********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   locate_pt   ***************
*/

/*  INTERFACE:   */

void
locate_pt(
  ELEMENT_DATA*	el_pt,	/* ptr to current element struct		*/
  ELEMENT_INFO*	info,	/* ptr to element info struct			*/
  real	xpt[3][10]	/* return array of computed coords		*/
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine locate_pt
** locate_pt maps the integration points into physical space.
** Global coordinates are placed in xpt[j]
** Called by el_output element stress output function (above).
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of locate_pt *********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   save_refine   ***************
*/

/*  INTERFACE:   */

      void
      save_refine(
                  pamr_Mesh m[]
                 );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine save_refine
** writes out the current (refined) mesh in
** the form of a GeoFEST-readable input file
** for future use or adaptation
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of save_refine *********************************/ 
