/*
***                           File main.c                         ***
***                        GeoFEST version 4.5g
*** Copyright (c) 2003, California Institute of Technology        ***
*** U.S.Sponsorship under NASA Contract NAS7-1260 is acknowledged ***
***
*** This software is designated for public release under JPL Task ***
*** Plan Number 83-6791 and may be publicly released after        ***
*** approval for release has been granted by GSFC Software        ***
*** Release  Authority
***
*** This file contains the program main entry point, the main
*** task driver, and modules for driving high-level functions
*** and interactions with the operator:
***
***          == main ==
***          == elastic ==
***          == time_step ==
***          == clear_stiff ==
***          == elgrp_loop ==
***          == node_load ==
***          == accumulate ==
***          == completion ==
***          == wrt_save ==
***          == vrestart ==
***          == el_save ==
***          == check_monitor ==
***          == get_refine ==
***          == save_attributes ==
***          == update_phase ==
***          == terminate ==
***/

/*
***GeoFEST version 4.5g***
G. A. Lyzenga -- unified parallel/serial version and bug fixes 11/17/03
G. A. Lyzenga -- split node enhancements 8/16/02
G. A. Lyzenga -- modification to correctly relax hydrostatic stresses 6/15/00
J. W. Parker -- 3D support and deleted hypercube branches 9/16/98
G. A. Lyzenga  -- last modified on 3/03/95
Modification history:
2.1: fixes bug in boundary velocities and changes
     the way stresses are output.
2.2: implements split node faults
2.3: adds pass-thru comments
3.0: 3-D tetrahedra supported, hypercube branches deleted
3.1: hydrostatic relaxation changes
4.0: name change from visco to GeoFEST; correct implementation of 2-D out-of-
     plane stresses as well as hydrostatic relaxation; culling of remaining 
     vestiges of hypercube stuff; one code without recompilation for both
     2-D and 3-D domains
4.1: Introduction of trilinear hexahedral bricks and PCG iterative solver option
4.2: Bug fixes; standard source prologues added (June 02)
4.3: Corrects inefficiency in split node algorithm and introduces new
     split node input format
4.4: Adds dummy calls to MPI and Pyramid to provide a single unified
     source for parallel and serial codes.  Also introduces a major
     algorithmic enhancement; stress equilibrium maintenance during visco
     elastic stepping for long-term stability.  Also more minor bug fixes.
4.5: Supports surface tractions and roller/spring duo options for tet elements.
     Provides enhanced conjugate gradient convergence monitoring. 
 $Log: main.c,v $
 Revision 1.11  2005/05/02 23:38:23  jwparker
 Charles says he fixed an inadvertent modification that did more graphic
 related output but stopped early.  So this should be the working 4.5g.

 Revision 1.10  2005/04/25 23:39:26  jwparker
 This version is the Milestone G version which ran on Cosmos, and also the one
 we must post on OpenChannel for any verification. Further work should compare
 this with the previous revision, which ran on Los Angeles and so was
 further developed towards automatic threshholding for AMR.  There is also
 a beta version of 4.5 plus buoyancy and fault physics advances I think, yet
 to be checked in.

 This version may only be used on parallel computers due to AMR hard-wiring.

 Revision 1.8  2004/04/05 21:15:23  jwparker
 This is the update version 4.5 with several improvements over the
 milestone F code. These include single source file for parallel
 and sequential use (supported by Sequential_Library dummy calls
 to mimic MPI functions), physics based correction for stress
 equilibirium in long-term viscous solutions,enhanced convergence
 monitoring, and support for surface tractions and roller/spring duo
 options for tet elements.

 Revision 1.4  2003/02/09 06:10:55  lyzenga
 *** empty log message ***

*/

#define EXTERN
#include "main.h"
#include "utility.h"
#include "solver.h"
#include "inphase.h"
#include "generat.h"
#include "stiff.h"
#include "strain.h"
void input_phase(pamr_Mesh m[]);

/*
#define OK		0
#define KILL		1
#define SAVE_STOP	2
#define SAVE_GO		3
#define REPORT		4
#define NCODES		5
*/
typedef enum {
  OK,
  KILL,
  SAVE_STOP,
  SAVE_GO,
  REPORT,
  NCODES
  } CODE;

   char  line[80] , fullname[80] ;
   int   step_count ;
   int   add_int() ;
   char  *mon_codes[NCODES] =
          { "OK" , "KILL" , "SAVE_STOP" , "SAVE_GO" , "REPORT" } ;
       
/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   main   ***************
*/

/*  INTERFACE:   */
     int
     main(
          int argc ,      /* number of input arguments */
          char *argv[]    /* input argument strings */
         )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine main
** main prints a banner, 
** interprets up to two args as input and output files,
** scans two comment lines in the input, 
** and divides (and times) the processing
** into input, elastic solution, time-stepping, and output.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/


     {
       char CVS_Id[] = "@(#) CVS id $Id: main.c,v 1.11 2005/05/02 23:38:23 jwparker Exp $, Revision: " REVNO;
       int iam, nproc;
       pamr_Mesh m[2];
	/* Evidently, MPICH relies upon MPI_INIT to start up GeoFEST	*/
	/* on all of the [virtual] processors except the first one	*/
	/* but Pyramid-3D calls the Fortran version of MPI_INIT		*/
	/* which doesn't pass GeoFEST command line options		*/
	/* so we must call the C version of MPI_INIT here first.	*/
	MPI_Init(&argc, &argv);
	/* MPI_Init removes the MPICH mpirun command line options	*/
	/* leaving only the GeoFEST command line options.		*/
	/* Pyramid-3D checks whether MPI is already initialized		*/
	/* and will not attempt to initialize it again.			*/
       pamr_init(&argc, &argv, m, 1);
       iam = pamr_my_pid();
       nproc = pamr_num_procs();
/*
{ for (int j = 0; j < argc; ++j)
  fprintf(stdout, "%argv[%d] = %s\n", j, argv[j]);
  }
*/
       bugflag = 0 ;
       if (iam == 0) {
         printf("\n\n\n************************************************\n\n");
         printf("Welcome to the Geophysical Finite Element Simulation Tool "
		"(GeoFEST " REVNO ") (%s)\n", argv[0]); fflush(stdout);
         printf("CPU time = %f\nLoading Data At: ", (double)clock()/CLOCKS_PER_SEC);
       }
        pamr_current_time();

        num_element_terms = 1;
        num_face_terms = 4;   /* flag(0=def,1=trac,2=buoy), val1, val2, val3 */
        num_node_terms = 16;
        pamr_define_mesh_terms(&m[0], num_element_terms, num_face_terms, 0, num_node_terms); /* Elems, Faces, Edges, Nodes */
        pamr_define_mesh_terms(&m[1], num_element_terms, num_face_terms, 0, num_node_terms); /* Elems, Faces, Edges, Nodes */
 
	/* The first GeoFEST command line option is the input file name.*/
	{ const char*	inputFileName = (1 < argc)? argv[1]: "input.dat";
	  in_file = fopen(inputFileName, "r");
	  if (NULL != in_file) {
	    /* Pyramid-3D loads its own version of the input file.	*/
	    sprintf(fullname, "%s.jpl", inputFileName);
	    pamr_load_mesh_serial(&m[0], fullname);
	    pamr_repartition(&m[0]);
	    }
	  else {
	    fprintf(stderr, "Couldn't open input file: %s!\n", inputFileName);
	    attempt = FILE_OPEN;
	    }
	  completion();
	  }
	/* Read output file name into string line from the input file.	*/
	fscanf(in_file, "%s", line);
	/* The second GeoFEST command line option is			*/
       	/* the output directory path.					*/
	{ const char*	outputDirectoryPath = (2 < argc)? argv[2]: ".";
	  if (0 == pamr_my_pid()) {
	    sprintf(fullname, "%s/%s", outputDirectoryPath, line);
	    out_file = fopen(fullname, "w");
	    if (NULL == out_file) {
	      fprintf(stderr, "Couldn't open output file %s!\n", fullname);
	      attempt = FILE_OPEN;
	      }
	    else {
	      fprintf(stdout, "Opening output file %s ...\n", fullname);
	      }
	    }
	  completion();
	  if (0 == pamr_my_pid()) {
	    sprintf(fullname, "%s/%s", outputDirectoryPath, CGFILE);
	    cgconv_file = fopen(fullname, "w");
	    if (NULL == cgconv_file) {
	      fprintf(stderr, "Couldn't open convergence history file %s!\n",
		  fullname);
	      attempt = FILE_OPEN;
	      }
	    }
	  completion();
	  }


/*  *** read and echo the pass-thru comment lines *** */

       fscanf(in_file,"%[^*]*\n",line) ;
       if (pamr_my_pid() == 0) {
	printf("%s\n",line) ;
	fprintf(out_file,"%s\n",line) ;
	}

       fscanf(in_file,"%[^*]*\n",line) ;
       if(pamr_my_pid() == 0) {
	printf("%s\n",line) ;
	fprintf(out_file,"%s\n",line) ;
	}

       input_phase(&m[0]) ;
       completion() ;
       if(flow_sim[0] == 0) terminate(iam) ;

if (iam == 0) {
  printf("CPU time = %f\nStarting Elastic Soln At: ",
      (double)clock()/CLOCKS_PER_SEC);
  }
pamr_current_time();

       if( !time_data.restart )
          {
           elastic(&m[0]) ;


           if(flow_sim[1] == 1) output_phase(&m[0], 0) ;

      if(flow_sim[2] == 1) /*refinement is desired *****************/
          {
           squawk("Starting Save Attributes.\n") ;

           save_attributes(&m[0]) ;

           squawk("Starting Adaptive Refinement.\n") ;

           get_refine(&m[0], &m[1]) ;

           squawk("Starting Update Phase.\n") ;

           update_phase(&m[0], &m[1]) ;

           if(flow_sim[3] == 1) 
              save_refine(&m[1]) ;

	   /*
pamr_visualize(&m[1], "Landers16M.plt");
pamr_finalize(TRUE); exit(EXIT_SUCCESS);
           output_phase(&m[1], 0) ;
pamr_finalize(TRUE); exit(EXIT_SUCCESS);
	   */
          if(flow_sim[4] == 1) /* solution on refined mesh is desired */
              {
               elastic(&m[1]) ;
               if(flow_sim[5] == 1)
                  output_phase(&m[1], 0) ;
              }


	   /*
pamr_visualize(&m[1], "Stepover1.plt");
pamr_finalize(TRUE); exit(EXIT_SUCCESS);
	   */
          }  /* end refinement option ***************************/

          }

if(flow_sim[6] == 0) terminate(iam) ;

if (iam == 0) {
               printf("CPU time = %f\nStarting Time_Step Soln At: ",
               (double)clock()/CLOCKS_PER_SEC);
              }
pamr_current_time();

       if(flow_sim[2] == 1)
          time_step(&m[1]) ;
       else
          time_step(&m[0]) ;
          
       if(time_data.sav_state)
          wrt_save(&m[1]) ;
          
        terminate(iam) ;

     } 
/****************************** end of main **********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   elastic   ***************
*/

/*  INTERFACE:   */

void
elastic(pamr_Mesh m[])

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine elastic
** elastic performs an elastic solution of the 
** finite element problem.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
    {
     int  i ;
    
     squawk("Starting elastic solution...\n") ;
     time_data.elastic = 1 ;
     clear_real(global.displ,loc_sys.numnp*fe_sys.ndof) ;
     clear_real(global.del_displ,loc_sys.numnp*fe_sys.ndof) ;
     clear_real(force.full_rhs,loc_sys.neq) ;
     clear_real(force.net_external,loc_sys.neq) ;
     clear_stiff() ;
     squawk("Storage arrays cleared...\n") ;

   /* only use PCG solver in parallel option! */
     if (DIRECT == fe_sys.solver && pamr_num_procs() != 1)
        {
         attempt = WRONG_SOLVER ;
         completion() ;
        }

     elgrp_loop( m, BC_ELAS ) ;
     squawk("BC's computed...\n") ;
     time_data.do_slip = 1 ; /* obsolete; replaced by groups below */
     printf("Number of fault slip groups = %d\n",time_data.nfgrps) ;
     
     if(time_data.ntime_grp == 0)
        {
         squawk("No time stepping; slip all faults now...\n") ;
         for(i=0 ; i<time_data.nfgrps ; i++)
            {
             (fltgrp_ptr+i)->due_now = 1 ; /* slip this fault at t=0 */
             (fltgrp_ptr+i)->first_event = ZERO ;
             /* in elastic-only problems, slip all faults */
            }
        }
     
     for(i=0 ; i<time_data.nfgrps ; i++)
        {
         (fltgrp_ptr+i)->prev_num = 0 ;
         
         if( (fltgrp_ptr+i)->first_event == ZERO)
            {
             (fltgrp_ptr+i)->due_now = 1 ; /* slip this fault at t=0 */
             printf("Fault #%d slip occurring at t = 0 \n",i+1) ;
            }
    /* A fault with a first event time of zero will move during the elastic step;
       it will then not move again until a full repeat interval has elapsed */
         else
            {
             (fltgrp_ptr+i)->due_now = 0 ;
             (fltgrp_ptr+i)->prev_num = -1 ;
    /* a fault with a non-zero first event time will not move during the elastic
       step, but will at the first event time and every repeat interval thereafter */
            }
        }
     elgrp_loop( m, FORMS_ELAS ) ;
     squawk("Stiffness computed...\n") ;
     elgrp_loop( m, RHS_ELAS ) ;
     squawk("RHS done...\n") ;


     /* Globalize HERE on force.full_rhs and update pcg.precond (preconditioner) */
     
     globalize( m, force.full_rhs, 0 );
     
     if (PCG == fe_sys.solver)
        {
		 globalize( m, pcg.precond, 0 );
	
		 /* Finish forming preconditioner */
		 { int ii;
		   for (ii = 0; ii < loc_sys.neq; ii++) {
			   *(pcg.precond+ii) = ONE / *(pcg.precond+ii);
		   }
		 }
        }

     node_load(m,ZERO,YES) ;

     squawk("Calling SOLVER\n") ;

     if (DIRECT == fe_sys.solver)
        solver( m, FACBACK ) ;
     else if (PCG == fe_sys.solver)
        solver( m, ITER ) ;
     accumulate() ;
     elgrp_loop( m, E_STRESS ) ;
     squawk("Elastic solution complete.\n") ;
    }
/*************************** end of elastic **********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   time_step   ***************
*/

/*  INTERFACE:   */

void
time_step(pamr_Mesh m[])

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine time_step
** time_step computes the visco-elastic time stepping solutions
** to the finite element problem.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
    {
         COUNTER  i , eventnum ;

         int new_step = FALSE ;
         real time = time_data.time ;

	 completion() ;

         output_phase(m, 0) ;

         if (0 == time_data.ntime_grp)  return;
         time_data.elastic = 0 ;

         squawk("Starting time-stepping algorithm...\n") ;

         force.incr_rhs = (real * ) calloc(loc_sys.neq,sizeof(real)) ;
         clear_real(force.full_rhs,loc_sys.neq) ;

         if (NULL == force.incr_rhs) attempt = R_MEM ;
         completion() ;

         step_count = 0 ;

{ int	time_group = 0;
for ( time_group = 0 ; time_group < time_data.ntime_grp ; time_group++)
   {
	real	dt	= *(time_data.delt + time_group);
	real	alpha	= *(time_data.alpha + time_group);
	real	endtime	= *(time_data.endtime + time_group);
	int	step	= -1;
	time_data.nsteps[time_group] = 0;
	time_data.nsub = 1;
		     
	while (time_data.time < endtime)
	   {
	    int	reform	= NO;
	    real	tstep	= dt/(real)time_data.nsub;
	    ++time_data.nsteps[time_group];
	    ++step;
	    ++step_count;
	    time += tstep;
	    time_data.dt = tstep;
	    time_data.alpha_delt = tstep*alpha;
	    time_data.time = time;
	    
	    time_data.do_slip = 0;
		 for(i=0 ; i<time_data.nfgrps ; i++)
			{
			 (fltgrp_ptr+i)->due_now = 0 ;
			 eventnum = (int)((time - (fltgrp_ptr+i)->first_event)/(fltgrp_ptr+i)->interval) ;

			 if( eventnum > (fltgrp_ptr+i)->prev_num && eventnum >= 0 
			       && time >= (fltgrp_ptr+i)->first_event )
                {
                 time_data.do_slip = 1 ;
                 (fltgrp_ptr+i)->due_now = 1 ;
                 ((fltgrp_ptr+i)->prev_num)++ ;
                 if(pamr_my_pid() == 0)
		   printf("Scheduled fault #%d slip occurring at t = %f \n",i+1,time) ;
                }
			}

	    if( time_data.do_slip || 
		((alpha != ZERO) && (step % time_data.nreform == 0)) )
	       {
			clear_stiff() ;
			elgrp_loop( m, FORMS_STEP ) ;
			reform = YES ;
           }
  /*   postpone this clearing until after buoyancy calc is done
		clear_real(global.del_displ,
		           loc_sys.numnp*fe_sys.ndof) ;
  */         
     new_step = FALSE ;
     if(fe_sys.nrates == 0)
       {
        clear_real(force.incr_rhs,loc_sys.neq) ;
        clear_real(global.forv , loc_sys.numnp*fe_sys.ndof) ;
       }
     else
	   { int	n = 0;
		 for(n=0 ; n<fe_sys.nrates ; n++)
			{
			 real test = *(global.bctime + n) ;
			 if(ZERO <= test && test <= time)
				{
				 new_step = TRUE ;
				 move_real(global.rate + n*loc_sys.numnp*fe_sys.ndof
							, global.forv , loc_sys.numnp*fe_sys.ndof) ;
				 *(global.bctime + n) = (-ONE) ;
				 break ;
				}
			}
	   }
	   
     if(reform || new_step)  /* only redo the rhs if the bc or stiffness have changed */
       {
        clear_real(force.incr_rhs,loc_sys.neq) ;
        elgrp_loop( m, BC_STEP ) ;
       }

     incr_real(force.incr_rhs , force.full_rhs , loc_sys.neq) ;
       /* need to preserve the equil correction already in full_rhs from last step */


		time_data.max_strain = ZERO ;
	
		elgrp_loop( m, RHS_STEP ) ;
                
     /* Globalize HERE on force.full_rhs;
        also update M (preconditioner) ONLY IF reform == YES */

	globalize( m, force.full_rhs, 0 );

	/* Finish forming preconditioner */
	if(reform == YES && PCG == fe_sys.solver)
	  { int ii;
		  globalize( m, pcg.precond, 0 ) ;
		  for (ii = 0; ii < loc_sys.neq; ii++) {
			*(pcg.precond+ii) = ONE / *(pcg.precond+ii);
		  }
      }
     
/*   now we (post-globalize) add any nodal load forces  (GL)  */
     node_load(m,time,reform) ;

	 clear_real(global.del_displ,
		           loc_sys.numnp*fe_sys.ndof) ;

	{ int new_nsub = time_data.nsub;
	  if (time_data.max_strain > ZERO)
	   {
	    real	ratio = time_data.max_strain*5.0*time_data.nsub;
	    double	exp = log(ratio)/log(TWO);
	    exp = (double)((int)exp);
	    new_nsub = MIN ((int) (0.01 + pow(2.0, exp + 1.0)), MAXSUB);
	    new_nsub = MAX (1, new_nsub);
	   }

	  else
	   {
	    new_nsub = time_data.nsub ;
	   }

	  if (new_nsub != time_data.nsub)
	   {
	    int  failed = 0 ;  /* temporarily not implemented */
            if( !failed )
	    time -= tstep ;
	    time_data.time = time ;
	    if(pamr_my_pid() == 0)
              printf("Changing nsub from %d to %d at time = %f\n", 
	    time_data.nsub, new_nsub, time) ;
	    time_data.nsub = new_nsub ;
	    step = (-1) ;
	    continue ;    /* roll back a time step */
	  }
	}
                            
                 if (PCG == fe_sys.solver)
                     solver( m, ITER ) ;
                 else if(reform)
                     solver( m, FACBACK ) ;
                 else
                     solver( m, BACK ) ;
                               
                  accumulate() ;
                  elgrp_loop( m, V_STRESS ) ;
  
          /* perform stress equilibrium correction */
                  clear_real(force.full_rhs,loc_sys.neq) ;
                  elgrp_loop( m, EQUIL ) ;
                  incr_real(force.net_external , force.full_rhs , loc_sys.neq) ;

                  output_phase(m, 0) ;

                  check_monitor(m, step_count+1) ;
                  
                  fflush( stdout ) ;


		     }     /* end loop within time group */
             }   /* end loop over time groups  */
  }                
}
             
/*************************** end of time_step ********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   clear_stiff   ***************
*/

/*  INTERFACE:   */

void
clear_stiff(void)

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine clear_stiff
** clear_stiff nulls the stiffness matrix memory.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

{
    switch(fe_sys.solver)
       {
        case DIRECT:
         clear_real(stiff.matrix,stiff.size) ;
         break;
         
        case PCG:
         clear_real(pcg.precond,pcg.neq) ;
         break;
       }

}
/*************************** end of clear_stiff ******************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   elgrp_loop   ***************
*/

/*  INTERFACE:   */

void
     elgrp_loop(
                pamr_Mesh m[],      /* parallel mesh */
                int  task      /* which element-level task to perform */
               )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine elgrp_loop
** elgrp_loop invokes a particular task that relies on 
** element structures and affects all the elements in 
** every element group.
**
** Supported tasks are
** GENERATE: read or generate element data, fill ien, lm arrays
** OUTPUT: compute and write out element centroid stresses
** SHAPE: compute parent-space shape functions and gradients
** FORMS_ELAS: compute elastic stiffness matrix
** FORMS_STEP: compute VE single-step stiffness matrix
** RHS_ELAS: compute elastic right-hand side vector for FE problem
** RHS_STEP: compute VE right-hand side vector for FE problem
** BC_ELAS:  compute boundary-condition terms for elastic FE problem
** BC_STEP: compute boundary-condition terms for VE FE single step
** COL_HT: compute the column heights for profile stiffness storage
** E_STRESS: compute the stresses from elastic solution
** V_STRESS: compute the stresses from a VE step solution
** DUMP: dump information needed for restarts
** RESTORE: read and restore solution vector and info for a restart
** FAIL_CHECK: check for stress-driven fault failure
** SMOOTH_BEGIN: perform pressure smoothing set-up
** SMOOTH_END: apply pressure-smoothing function from coefficients
** REORDER: analyze adjacency, find profile-optimizing node permutation
** EQUIL: compute equilibrium stress corrections for visco steps
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

  {
   int	i = 0;
   
   for(i=0 ; i<global.n_group ; i++)
      {
       switch( task )
               {
                case GENERATE:
                    gen_element(m, &groups[i] ) ;
                    break ;
                case OUTPUT:
                    el_output(m, &groups[i] ) ;
                    break ;
                case SHAPE:
                    p_shape( &groups[i] ) ;
                    completion() ;
                    break ;
                case FORMS_ELAS:
                    form_stiff( &groups[i] , ELASTIC ) ;
                    completion() ;
                    break ;
                case FORMS_STEP:
                    form_stiff( &groups[i] , VISCO ) ;
                    completion() ;
                    break ;
                case RHS_ELAS:
                    form_rhs( &groups[i] , ELASTIC ) ;
                    completion() ;
                    break ;
                case RHS_STEP:
                    form_rhs( &groups[i] , VISCO ) ;
                    completion() ;
                    break ;
                case BC_ELAS:
                    form_bc( &groups[i] , ELASTIC ) ;
                    completion() ;
                    break ;
                case BC_STEP:
                    form_bc( &groups[i] , VISCO ) ;
                    completion() ;
                    break ;
                case COL_HT:
                    colht( &groups[i] , &stiff ) ;
                    completion() ;
                    break ;
                case E_STRESS:
                    form_stress(&groups[i] , ELASTIC) ;
		    completion() ;
                    break ;
                case V_STRESS:
                    form_stress(&groups[i] , VISCO) ;
		    completion() ;
                    break ;
                case ESPROD:
                    estiffprod( &groups[i] ) ;
                    break ;
                case DUMP:
                    el_save(&groups[i] , DUMP) ;
                    break ;
                case RESTORE:
                    el_save(&groups[i] , RESTORE) ;
                    break ;
                case REORDER:
                    reorder( &groups[i] );
                    completion() ;
                    break;
                case EQUIL:
                    form_equil( &groups[i] );
                    break;
                default:
                    attempt = BAD_TASK ;
                    completion() ;
                    break ;
               }

      }  /* end loop */
  }
/*************************** end of elgrp_loop *******************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   node_load   ***************
*/

/*  INTERFACE:   */

void
     node_load(
               pamr_Mesh m[],      /* parallel mesh */
               real  time ,   /* current simulation time */
               int reform     /* flag indicating stiffness reform status */
              )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine node_load
** node_load applies node-based forces that may accumulate
** with time.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

    {
     real   val ;
     int  k,  id , new_step ;
     int pgnode; /* one-based permuted node numbers */
    
     int	ndof	= fe_sys.ndof;
     int	numnp	= loc_sys.numnp;
    if(ZERO == time)          /* the ELASTIC case... */
       {
        /*  put in free nodal forces, if any  */
    {  int	i = 0;
        for(i=0 ; i<numnp ; i++)
           {
            for(k=0 ; k<ndof ; k++)
               {
                pgnode = global.permut[i+1];
                id = *(global.id_pointer+(pgnode-1)*ndof+k) ;
                val = *(global.forv+i*ndof+k) ;
                if(val != ZERO && id != (-1))
                  {
                   *(force.full_rhs + id) += val ; /* Due to movement of call */
                   *(force.net_external + id) -= val ;
                  }
               }
           }
     }

       }
    
  else {		/* the TIME-STEPPING case... */
  
 
/* these tasks need to be performed  post-globalize in parallel code (GL) */

    { int	i = 0;
        for(i=0 ; i<numnp ; i++)
           {
        	int	k = 0;
            for(k=0 ; k<ndof ; k++)
               {
                int pgnode = global.permut[i+1];
                int id = *(global.id_pointer+(pgnode-1)*ndof+k) ;
                real val = *(global.forv+i*ndof+k) * time_data.dt ;
                if(val != ZERO && id != (-1))
                  {
                   *(force.full_rhs + id) += val ; /* Due to movement of call */
                   *(force.net_external + id) -= val ;
                  }
               }
           }
    }
    

       }
    }
/*************************** end of node_load ********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   accumulate   ***************
*/

/*  INTERFACE:   */

void
accumulate(void)

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine accumulate
** accumulate updates the total displacement based on the 
** single-step increment.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

    {
     COUNTER   i=0 ;

     for(i=0 ; i<loc_sys.numnp*fe_sys.ndof ; i++)
        {
         *(global.displ + i) += *(global.del_displ + i) ;
        }     
    }

/*************************** end of accumulate *******************************/ 

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   completion   ***************
*/

int attempt = 0;			/* global variable definition	*/

/*  INTERFACE:   */

void
completion(void)

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine completion
** completion syncronizes multiple processors. 
** Keeps track of incurred errors in each processor.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

    {
     int   all = 0 , contrib = attempt , ierror ;
     
     if(pamr_num_procs() == 1)
        all = attempt ;
     else
        ierror =
         MPI_Allreduce((void *)&contrib, (void *)&all, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
        
     if(all)
        {
         squawk("An error occurred:\n") ;
         printf("IAM: %d\t  error code = %d\n", pamr_my_pid(), attempt) ;
	     fflush(stdout);
	     pamr_finalize(TRUE) ;
         exit(EXIT_FAILURE) ;
        }

    }

/*************************** end of completion *******************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   wrt_save   ***************
*/

/*  INTERFACE:   */
     void
     wrt_save(pamr_Mesh m[])


/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine wrt_save
** wrt_save dumps state to disk for restarts.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

    {
     static real  outbuf[4] ;
         
     COUNTER	ndof	= fe_sys.ndof;
     COUNTER	numnp	= fe_sys.numnp;
     real		time	= time_data.time;
     
     if (NULL == (dump_file = fopen(time_data.savefile, "w")))
          {
           printf("Dump file couldn't be opened!\n") ;
           return ;
          }

     printf("Saving simulation at time = %f to disk...\n",time) ;
     fprintf(dump_file,"%f\n",time) ;
     
  { COUNTER	i = 0;
     for(i=0 ; i<numnp ; i++)
                 {
                  COUNTER	index = i;
                  COUNTER	k = 0;
                  for(k=0 ; k<ndof ; k++)
                         {
                          outbuf[k] =
                             *(global.displ + ndof*index + k) ;
                         }
                            
                  fprintf(dump_file,"%d",i+1) ;
                  for(k=0 ; k<ndof ; k++)
                     {
                      fprintf(dump_file,"  %f",outbuf[k]) ;
                     }
                  fprintf(dump_file,"\n") ;
                 }        
    }
             fflush(dump_file) ;
             elgrp_loop( m, DUMP) ;
             fclose(dump_file) ;
             printf("Simulation dump complete.\n") ;
    }

/*************************** end of wrt_save *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   vrestart   ***************
*/

/*  INTERFACE:   */

void
vrestart(pamr_Mesh m[])

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** vrestart restores state from previous dump and 
** continues simulation.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

    {
     static real  outbuf[4] ;
         
     COUNTER	ndof = fe_sys.ndof;
     COUNTER	numnp = fe_sys.numnp;
     real		time = time_data.time;
     
     if (NULL == (dump_file = fopen(time_data.startfile, "r"))) 
          {
           printf("Restart file %s couldn't be opened!\n",time_data.startfile) ;
           exit(1) ;
          }

     fscanf(dump_file,"%lf\n",&time) ;
     time_data.time = time ;
     printf("Restarting simulation at time = %f ...\n",time) ;
     
  { COUNTER	i = 0;
     for(i=0 ; i<numnp ; i++)
        {
         int	n = 0;
         fscanf(dump_file,"%d",&n) ;
          { COUNTER	k = 0;
                  for(k=0 ; k<ndof ; k++)
                     {
                      fscanf(dump_file,"%lf",outbuf+k) ;
                     }
	      }
          { COUNTER	index = i;
        	COUNTER   k = 0;
            for(k=0 ; k<ndof ; k++)
				 {
				  *(global.displ + ndof*index + k) =
					 outbuf[k] ;
				 }
          }        
      }
    }

	 elgrp_loop( m, RESTORE) ;
	 fclose(dump_file) ;
	 printf("Simulation restoration complete.\n") ;
    }

/*************************** end of vrestart *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   el_save   ***************
*/

/*  INTERFACE:   */
     void
     el_save(
             GROUP   *grp_ptr ,       /* pointer to current element group */
             int     code             /* code specifying which action */
            )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine el_save
** el_save dumps or restores element data.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

{
 if (DUMP == code) 
     {
	  int numel = (grp_ptr->el_info)->numel;
	  int	str_size = (grp_ptr->el_info)->stress_size;
      fflush(dump_file) ;
      
      switch((grp_ptr->el_info)->type)
          {
           case BILIN:
           case SERENDIP:
           case TET:
      { int		nel = 0;
               for ( nel = 0 ; nel < numel ; nel++)
                   {
				    int		n = nel + 1;
				    ELEMENT_DATA*	el_pt = grp_ptr->el_data + n - 1;
				    int		nint = el_pt->nint;
				    int		ipt = 0;
                    fflush(dump_file) ;

                    for (ipt = 0 ; ipt < nint ; ipt++)
                       {
	                    int		j = 0;
                        for(j=0 ; j<str_size ; j++)
                           fprintf( dump_file,"%f  ",
                            *((el_pt -> stress) + ipt*str_size + j) ) ;
                        fprintf(dump_file,"\n") ;
                       }
                   }
	}
               fflush(dump_file) ;
               break ;
               
          }
     }

else if(RESTORE == code)
     {
      int	numel = (grp_ptr->el_info)->numel;
      int	str_size = (grp_ptr->el_info)->stress_size;
      
      switch((grp_ptr->el_info)->type)
          {
           case BILIN:
           case SERENDIP:
      { int		nel = 0;
               for ( nel = 0 ; nel < numel ; nel++)
                   {
					int		n = nel + 1;
					ELEMENT_DATA*	el_pt = grp_ptr->el_data + n - 1;
					int		nint = el_pt->nint;
					int		ipt = 0;
                    for (ipt = 0 ; ipt < nint ; ipt++)
                       {
                       	int		j = 0;
                        for(j=0 ; j<str_size ; j++)
                           {
	                        real	str_temp;
                            fscanf( dump_file,"%lf",
                                    &str_temp ) ;
                            if(n != 0)
                               {
                                el_pt = grp_ptr->el_data + n - 1 ;
                                *((el_pt -> stress) + ipt*str_size + j) =
                                  str_temp ;
                               }
                           }
                       }
                   }
	}
               break ;
               
          }
     }

    }

/**************************** end of el_save *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   check_monitor   ***************
*/

/*  INTERFACE:   */
     void
     check_monitor(
                   pamr_Mesh m[],  /* parallel mesh */
                   int   nstep    /* count of steps for backup */
                  )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine check_monitor
** check_monitor allows user run intervention via strings stored
** in the file monitor.fem.
**
** Supported options:
** OK: take no action - continue running normally
** REPORT: report cuurent time step to stdout and continue running normally
** KILL: close files and abort
** SAVE_STOP: write state, then abort
** SAVE_GO:  write dumps as requested in input, and continue
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

    {
     int  flag = 0 ;
     int  n = 0 ;
     
     if(nstep % time_data.nbackup == 0)
       {
	if (pamr_my_pid() == 0)
           printf("Simulation running normally at time = %f ; step size = %f\n",
             time_data.time , time_data.dt) ;
        flag = 1 ;
       }
     n = NCODES;
     if(NULL == (dump_file = fopen("monitor.fem", "r")))
          {
           return ;
          }

     fscanf(dump_file,"%s",line) ;
  { int	i = 0;
     for(i=0 ; i<NCODES ; i++)
        {
         if(0 == strcmp(line, mon_codes[i]))
            {
             n = i ;
             break ;
            }
        }
    }

     switch( n )
        {
         case OK:
            fclose(dump_file) ;
            if(flag && time_data.sav_state)
               {
		wrt_save(m) ;
		squawk("Run backed up and continuing...\n") ;
               }
            return ;
            break ;

         case REPORT:
	    if(pamr_my_pid() == 0)
	       {
		rewind(dump_file) ;
		fprintf(dump_file,"OK\n") ;
		printf("Simulation running normally at time = %f ; step size = %f\n",
                time_data.time , time_data.dt) ;
	       }
	    fclose(dump_file) ;
            return ;
            break ;

         case KILL:
            fclose(in_file) ;
	    if(pamr_my_pid() == 0)
	       {
		fclose(out_file) ;
		rewind(dump_file) ;
		fprintf(dump_file,"OK\n") ;
printf("CPU time = %f\nTime at Forced Termination: ",
    (double)clock()/CLOCKS_PER_SEC);
            }
	    fclose(dump_file) ;
            pamr_current_time();
            squawk("Run aborted immediately by monitor request.\n") ;
            pamr_finalize(TRUE) ;
            exit(EXIT_SUCCESS) ;
            break ;

         case SAVE_STOP:
	    if(pamr_my_pid() == 0)
	       {
		rewind(dump_file) ;
		fprintf(dump_file,"OK\n") ;
	       }
	    fclose(dump_file) ;
            if(time_data.sav_state)
               wrt_save(m) ;
            else
               {
                sprintf(time_data.savefile,"save.def") ;
                wrt_save(m) ;
               }
	    fclose(in_file) ;
	    if(pamr_my_pid() == 0)
                fclose(out_file) ;
            squawk("Run saved and halted by monitor request.\n") ;
	    pamr_finalize(TRUE) ;
            exit(EXIT_SUCCESS) ;
            break ;

         case SAVE_GO:
	    if(pamr_my_pid() == 0)
	       {
		rewind(dump_file) ;
		fprintf(dump_file,"OK\n") ;
	       }
	    fclose(dump_file) ;
            if(time_data.sav_state)
               wrt_save(m) ;
            else
               {
                sprintf(time_data.savefile,"save.def") ;
                wrt_save(m) ;
               }
            squawk("Run backed up and continuing...\n") ;
            return ;
            break ;

         default:
            fclose(dump_file) ;
            return ;
            break ;
        }
    }

/************************** end of check_monitor *****************************/ 

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*                                                                                                                                  
*************   ROUTINE:   get_refine   ***************                                                                          
*/

/*  INTERFACE:   */

void
get_refine(pamr_Mesh m_in[], pamr_Mesh m_out[])

     /*  RETURN VALUE:  -none- */
     /*  DESCRIPTION:   */
     /*                                                                                                                                  
     ** Routine get_refine                                                                                                               
     ** performs adaptive mesh refinement based on strain energy.                                                                        
     ** Only works for one elment group.                                                                                                 
     **/
     /*   EOP   */
     /*---------------------------------------------------------------------------*/
{
  int i, j;
  int *error_elements;  /* list of marked elements for refinement */
  GROUP *grp_ptr;       /* pointer to current element group */

  grp_ptr = &groups[0];
  error_elements = (int *) calloc((grp_ptr->el_info)->numel, sizeof(int));

  for ( j = 0; j < (grp_ptr->el_info)->numel; j++ ) {
    if ( grp_ptr->el_data[j].str_energy > REFINE_THRESHOLD ) {
      error_elements[j] = 1;
    }
    else {
      error_elements[j] = 0;
    }
  } /* mark elements for refinement */

  pamr_mark_refinement(m_in, m_out, error_elements);
  pamr_logical_amr(m_in);
  pamr_repartition(m_in);
  pamr_physical_amr(m_in, m_out);

  free(error_elements);

  squawk("Adaptive mesh refinement complete.\n") ;
}
/*************************** end of get_refine **********************************/

/*
*************   ROUTINE:   save_attributes   ***************                                                                          
*/

/*  INTERFACE:   */

void
save_attributes(pamr_Mesh m[])

     /*  RETURN VALUE:  -none- */
     /*  DESCRIPTION:   */
     /*                                                                                                                                  
     ** Routine save_attributes

     ** saves GeoFEST problem description values as PYRAMID attributes

     ** Only works for one elment group.                                                                                                 
     **/
     /*   EOP   */
     /*---------------------------------------------------------------------------*/
{
  int i, j;
  /*
  real *eqtype_x, *eqtype_y, *eqtype_z;
  real *bx, *by, *bz, *sz, *sy, *sz, *slip;
  */
  int   ndof     = fe_sys.ndof;
  real *splitn_data = calloc((NSPLITATTR+ndof),sizeof(real));
  real *material_data = calloc(1,sizeof(real));
  real *facet_data = calloc(4,sizeof(real));
  GROUP *grp_ptr;       /* pointer to current element group */

  int	numnp	= loc_sys.numnp;
  int *fnode;
  int  nel , side , face_indx, pamr_side ;
  real *fvec;
  grp_ptr = &groups[0];
  BUOY_DATA  *buoy_ptr  ;
  int  *buoy_list , numface ;

  /* Set equation type attribute for nodes */
  for ( i = 0; i < numnp; i++ ) {

       /*
       pamr_map_mesh_node_terms( m, i, &eqtype_x, &eqtype_y, &eqtype_z,
                                 &bx, &by, &bz, &sx, &sy, &sz, &slip );
       */

       splitn_data[0] = (real) global.eqtype[i*ndof];
       splitn_data[1] = (real) global.eqtype[i*ndof+1];
       splitn_data[2] = (real) global.eqtype[i*ndof+2];
       splitn_data[3] = (real) global.splitn[NSPLITATTR*i+0];
       splitn_data[4] = (real) global.splitn[NSPLITATTR*i+1];
       splitn_data[5] = (real) global.splitn[NSPLITATTR*i+2];
       splitn_data[6] = (real) global.splitn[NSPLITATTR*i+3];
       splitn_data[7] = (real) global.splitn[NSPLITATTR*i+4];
       splitn_data[8] = (real) global.splitn[NSPLITATTR*i+5];
       splitn_data[9] = (real) global.splitn[NSPLITATTR*i+6];
       splitn_data[10] = (real) global.splitn[NSPLITATTR*i+7];

       /*
       printf("BIAM: %d SZ: %lf Slip: %lf\n",pamr_my_pid(), splitn_data[9],  splitn_data[10]);
       */

       pamr_set_mesh_node_terms( m, i+1, splitn_data ); /* mesh, node_indx, data_array */

  }

  /* Set attribute for elements*/
  for ( i = 0; i < (grp_ptr->el_info)->numel; i++ ) {
        material_data[0] = (real) (grp_ptr->el_data[i].mat + 1); /* now 1-based to match input file */

        pamr_set_mesh_element_terms( m, i+1, material_data ); /* Note Pyramid wants an array[1] */

  }

  /* gl: Set attributes for facets */
  /*
     Loop over surface forces and assign facet_data array (already allocated above)
     pamr_set_mesh_face_terms( m, face_indx, facet_data );
  */
  
    if( (grp_ptr->el_info)->numsuf != 0 )
       for ( i = 0 ; i < (grp_ptr->el_info)->numsuf ; i++ )
          {
           nel = *((grp_ptr->el_info)->surf_list + 2*i) ;
           side = *((grp_ptr->el_info)->surf_list + 2*i + 1) ;
           
           if(side == 1) pamr_side = 4 ;
           else if(side == 2) pamr_side = 3 ;
           else if(side == 3) pamr_side = 2 ;
           else if(side == 4) pamr_side = 1 ;
           
           face_indx = pamr_face_indx(m, nel, pamr_side) ;
           
	facet_data[0] = (real) 1 ;  /* denotes an active traction facet */
	facet_data[1] = *((grp_ptr->el_info)->surf_trac + ndof*i) ;
	facet_data[2] = *((grp_ptr->el_info)->surf_trac + ndof*i+1) ;
	facet_data[3] = *((grp_ptr->el_info)->surf_trac + ndof*i+2) ;

           pamr_set_mesh_face_terms( m, face_indx, facet_data ) ;

          }

  /*
     Now repeat the looping procedure for buoyancy surfaces (if any)
  */

    if( (grp_ptr->el_info)->numbuoy != 0 )
       for ( i = 0; i < (grp_ptr->el_info)->numbuoy; i++ )
          {
           buoy_ptr = (grp_ptr->el_info)->buoy + i ;
           buoy_list = buoy_ptr->buoy_list ;
		   numface = buoy_ptr->numface ;
		   for(j=0 ; j<numface ; j++)
			   {
				nel = *(buoy_list + 2*j) ;
				side = *(buoy_list + 2*j + 1) ;

           if(side == 1) pamr_side = 4 ;
           else if(side == 2) pamr_side = 3 ;
           else if(side == 3) pamr_side = 2 ;
           else if(side == 4) pamr_side = 1 ;
           
           face_indx = pamr_face_indx(m, nel, pamr_side) ;
           
			    facet_data[0] = (real) 2 ;  /* denotes an active buoyancy facet */
			    facet_data[1] = (real) (i+1) ; /* 1-based horizon index */
			    facet_data[2] = 0 ; /* empty for buoyancy case */
			    facet_data[3] = 0 ;
				
                pamr_set_mesh_face_terms( m, face_indx, facet_data ) ;
			   }
	
          }


  free(splitn_data);
  free(material_data);
  free(facet_data);
  squawk("save_attributes complete.\n") ;
}
/*************************** end of save_attributes **********************************/

/*
*************   ROUTINE:   update_phase   ***************                                                                          
*/

/*  INTERFACE:   */

void
update_phase(pamr_Mesh m_coarse[], pamr_Mesh m_refine[])

     /*  RETURN VALUE:  -none- */
     /*  DESCRIPTION:   */
     /*

     ** Routine update_phase

     ** comment on updating

     ** Only works for one elment group.

     **/
     /*   EOP   */
     /*---------------------------------------------------------------------------*/
{

GLOBAL_DATA  old_global ;
SYSTEM_INFO  old_fe_sys, old_loc_sys ;
GROUP *grp_ptr, *old_grp_ptr, *old_groups;
ELEMENT_DATA   *el_pt ;
BUOY_DATA      *buoy_ptr , *old_buoy_ptr ;
real *node_terms = calloc(num_node_terms,sizeof(real));
real *temp_amount, *fvec, ivec[3], svec[3], slip, norm,
      xctr, yctr, zctr;
real mat[1], ftemp, cross[3], dot, mult, length_sqrd;
int numnp, nsd, ndof, nrates, i, j, k, jj, g_numel, numel, numat, type,
    nel, nfnode, plastic, nint, ncount, nen, nee,
    stress_size, dbar_size, nsf, is_split, itemp, ien_current[8],
    tally, this_node, count, numfaces, numbuoy, numsuf, side, kk, index;
int inquire[4] ;
int *fnode, *temp_where, *buoy_list;
real facet_data[4];

     old_groups = (GROUP *) calloc(MAX_GROUP,sizeof(GROUP));
     if (old_groups == NULL) attempt = CRD_MEM ;
     completion() ;

     for (i=0; i < global.n_group; i++ ) {
         old_groups[i] = groups[i]; /* structure copy */
     }
     old_global = global; /* structure copy */
     old_fe_sys = fe_sys; /* structure copy */
     old_loc_sys = loc_sys; /* structure copy */

     fe_sys.numnp = pamr_total_node_count(m_refine);

     /* fe_sys.nsd,ndof,nrates,save_shape,solver are correct at this point */

     if (pamr_my_pid() == 0)
	printf("Number of nodes = %d\n\n",fe_sys.numnp) ;

     if(pamr_num_procs() != 1)
        loc_sys.numnp = pamr_local_node_count(m_refine);
     else
        loc_sys.numnp = fe_sys.numnp ;

     numnp = loc_sys.numnp;
     nsd = fe_sys.nsd ;
     ndof = fe_sys.ndof    ;
     nrates = fe_sys.nrates  ;

     global.coords = (real * ) calloc(numnp*nsd,sizeof(real)) ;
     if (global.coords == NULL) attempt = CRD_MEM ;
     completion() ;
     
     global.displ = (real * ) calloc(numnp*ndof,sizeof(real)) ;
     if (global.displ == NULL) attempt = CRD_MEM ;
     completion() ;
     
     global.del_displ = (real * ) calloc(numnp*ndof,sizeof(real)) ;
     if (global.del_displ == NULL) attempt = CRD_MEM ;
     completion() ;
     
     global.forv = (real * ) calloc(numnp*ndof,sizeof(real)) ;
     if (global.forv == NULL) attempt = CRD_MEM ;
     completion() ;
     
     global.eqtype = (int * ) calloc(numnp*ndof,sizeof(int)) ;
     if (global.eqtype == NULL) attempt = CRD_MEM ;
     completion() ;
     
     global.splitn = (real * ) calloc(numnp*NSPLITATTR,sizeof(real)) ;
     if (global.splitn == NULL) attempt = CRD_MEM ;
     completion() ;
     
     if( nrates > 0 )
        {
         old_global.bctime = global.bctime;
	  /* global.bctime = (real * ) calloc(nrates,sizeof(real)) ; */
/* Plan to make node rate an attribute...
         global.rate = (real * ) calloc(numnp*ndof*nrates,
                                         sizeof(real)) ;
         if (global.rate == NULL) attempt = CRD_MEM ;
*/
        }
     completion() ;

     /* Work in progress.... set node_terms
        Replaces gen_map() in input_phase */
     /* node_terms are real averaged values in PYRAMID by default so
        0.0 implies fixed while other values imply free
        Alternative is to use coarse mesh with refined mesh references */

     for (i = 0; i < numnp; i++) {
          pamr_get_mesh_node_terms(m_refine, i+1, node_terms);
          global.eqtype[i*ndof] = (node_terms[0] == 0.0) ? 0 : 1;
          global.eqtype[i*ndof+1] = (node_terms[1] == 0.0) ? 0 : 1;
          global.eqtype[i*ndof+2] = (node_terms[2] == 0.0) ? 0 : 1;
     }

     squawk("Global b.c. map generated...\n") ;

     gen_number(m_refine) ;  /* obtain local node and equation numbering */
        
     squawk("Equation numbering complete...\n") ; 

     /* gen_real to be replaced by filling of node coordinates 
        must be placed in global.coords */

     pamr_mesh_coordinates(m_refine, global.coords);

     /* B.C. stuff will go in here, but requires some thinking */

     /* global.n_group is okay as it comes from old values */

     /* Replace elgrp_loop(generate) with a new loop that gets
        pyramid data 
        Split node inheritance has to go here too.
     */


     for(i=0 ; i<global.n_group ; i++)
         {
          grp_ptr = &groups[i];
          old_grp_ptr = &old_groups[i];

	   /* This code follows input_phase */
          grp_ptr->el_info =
             (ELEMENT_INFO *) calloc(1,sizeof(ELEMENT_INFO)) ;
          if (grp_ptr->el_info == NULL) attempt = ELI_MEM + i ;
          completion() ;

	  /* Be careful not to free groups[i].el_mat pointer */
          grp_ptr->el_mat = old_grp_ptr->el_mat;
          
          grp_ptr->group_num = i+1 ; /* 1-based indexing */

	   /* This code follows gen_element */
          g_numel = pamr_total_element_count(m_refine);

          if(pamr_num_procs() != 1) numel = pamr_local_element_count(m_refine) ;
          else numel = g_numel ;

          (grp_ptr->el_info)->numel = numel ;
          (grp_ptr->el_info)->g_numel = g_numel;
          type = (grp_ptr->el_info)->type = (old_grp_ptr->el_info)->type;
          numat = (grp_ptr->el_info)->numat = (old_grp_ptr->el_info)->numat;
          numbuoy = (grp_ptr->el_info)->numbuoy = (old_grp_ptr->el_info)->numbuoy;
          (grp_ptr->el_info)->buoy = (BUOY_DATA *) calloc(numbuoy,sizeof(BUOY_DATA)) ;
          for(j=0 ; j<numbuoy ; j++)
             {  /* Now we transfer the unchanged parts of the buoy data over */
              buoy_ptr = (grp_ptr->el_info)->buoy + j ;
              old_buoy_ptr = (old_grp_ptr->el_info)->buoy + j ;
              buoy_ptr->upvec[0] = old_buoy_ptr->upvec[0] ;
              buoy_ptr->upvec[1] = old_buoy_ptr->upvec[1] ;
              buoy_ptr->upvec[2] = old_buoy_ptr->upvec[2] ;
              buoy_ptr->rho_g    = old_buoy_ptr->rho_g ;
             }

	  printf ("IAM: %d\t local numel = %d\n", pamr_my_pid() , numel) ;
          if (pamr_my_pid() == 0)
	      printf("Global number of elements = %d\n\n",g_numel) ;
          
	  fflush(stdout) ;

          grp_ptr->el_data =
              (ELEMENT_DATA *) calloc(numel,sizeof(ELEMENT_DATA)) ;
          if (NULL == grp_ptr->el_data) attempt = ELD_MEM ;

          completion() ;

          nen = (grp_ptr->el_info)->nen = (old_grp_ptr->el_info)->nen;

          ndof = fe_sys.ndof ;
          nsd = fe_sys.nsd ;
          nee = (grp_ptr->el_info)->nee = nen*ndof ;
          (grp_ptr->el_info)->el_size = nee*nee ;
     
          stress_size = (grp_ptr->el_info)->stress_size =
                        (old_grp_ptr->el_info)->stress_size;
          dbar_size   = (grp_ptr->el_info)->dbar_size =
 	                (old_grp_ptr->el_info)->dbar_size;

         if( type == BILIN || type == SERENDIP || type == TET || type == TRILIN )
           {
            nsf = ndof ;
           }
     
          /*  we have already set element material properties 
               above via pointer-copy . . .  */

	 for( nel=0 ; nel<numel ; nel++ )
	   {
	     el_pt = grp_ptr->el_data + nel ;

	     /*  now allocate memory for the ien and lm arrays... */

	     el_pt->ien = (int *) calloc(nen,sizeof(int)) ;
	     if (NULL == el_pt->ien) attempt = IEN_MEM ;

	     el_pt->lm = (int *) calloc(nee,sizeof(int)) ;
	     if (NULL == el_pt->lm) attempt = LM_MEM ;

	   }

	 completion() ;

         /* generate element structures */
	 for( nel=0 ; nel<numel ; nel++ )
	   {
           /* Note Pyramid wants an array[1] */
             pamr_get_mesh_element_terms( m_refine, nel+1, &mat[0] );
             for ( j = 0 ; j < nen ; ++j) {
  	          ien_current[j] = pamr_element_node_indx(m_refine, nel+1, j+1);
	     }
	     /*
	     if (pamr_my_pid() == 4) {
	     printf("%d %d %d %d %d\n", nel+1,ien_current[0], ien_current[1],
		    ien_current[2],ien_current[3]);
	     }
	     */
              el_pt = grp_ptr->el_data + nel ;
              /* Passing local index nel that may not ever be used */
              load_element(el_pt,grp_ptr->el_info,
                           nel+1,ien_current,(int)mat[0]) ;

	      /*
              if (el_pt->mat != 0) printf("IAM: %d Mat: %d\n",pamr_my_pid(),el_pt->mat);
	      */

              plastic = *( (grp_ptr->el_mat) -> plastic + el_pt->mat ) ;
              nint = el_pt->nint ;

         if( type == BILIN || type == SERENDIP || type == TET || type == TRILIN)
            {
             el_pt->stress =
                ( real * ) calloc( nint*stress_size , sizeof(real)) ;
             if (NULL == el_pt->stress) attempt = STRESS_MEM  ;
             
             el_pt->dilat =
                ( real * ) calloc( nint , sizeof(real)) ;
             if (NULL == el_pt->dilat) attempt = STRESS_MEM  ;

             el_pt->is_split =
                ( int * ) calloc( 8 , sizeof(int)) ;
  /* there should never be more than 7 split nodes in an element, I hope */              
             if (NULL == el_pt->is_split) attempt = STRESS_MEM  ;
             *(el_pt->is_split) = 0 ; /* initialize just to be sure */

             if( plastic )
                {
                 el_pt->bta =
                 ( real * ) calloc( nint*stress_size , sizeof(real)) ;
                 if (NULL == el_pt->bta) attempt = BTA_MEM  ;
            
                 el_pt->dbar =
                 ( real * ) calloc( nint*dbar_size , sizeof(real)) ;
                 if (NULL == el_pt->dbar) attempt = DBAR_MEM  ;

                }
            }
   /*  if PCG is being used, allocate memory for element stiffness storage */
         if(PCG == fe_sys.solver)
            {
             el_pt->stiff =
                ( real * ) calloc( (grp_ptr->el_info)->el_size , sizeof(real)) ;
             if (NULL == el_pt->stiff) attempt = ELSTIFF_MEM  ;
            }

	   }
      completion() ;

      /* Here goes inheritance of surface tractions and buoyancy forces */
      /* 1. number of local faces
         2. identify and count number of faces as buoy or trac
         3. allocate space for these
         4. store values and owner element/face from pyramid into surf_trac
            and surf_list (grp_ptr->el_info)->surf_trac
      */

      for(j=0 ; j<numbuoy ; j++)
         ((grp_ptr->el_info)->buoy + j)->numface = 0 ; /* zero all the buoy facet counts */
      numsuf = 0;
      numfaces = pamr_local_face_count(m_refine);

      for (k = 0; k < numfaces; k++) {
	    pamr_get_mesh_face_terms(m_refine, k+1, facet_data);
        if ( (int) facet_data[0] == 1)
            numsuf++;
	    else if ( (int) facet_data[0] == 2)
	     {
     	    for(j=0 ; j<numbuoy ; j++)
     	       {
     	        if((int) facet_data[1] == j+1)
     	          (((grp_ptr->el_info)->buoy + j)->numface)++ ;
     	       }
     	 }
      }

      (grp_ptr->el_info)->surf_list = (int *) calloc(2*numsuf, sizeof(int));
      (grp_ptr->el_info)->surf_trac = (real *) calloc(ndof*numsuf, sizeof(real));
      for(j=0 ; j<numbuoy ; j++)
     	       {
     	        buoy_ptr = (grp_ptr->el_info)->buoy + j ;
     	        k = buoy_ptr->numface ;
     	        buoy_ptr->buoy_list = (int *) calloc(k,2*sizeof(int)) ;
     	        buoy_ptr->tally = 0 ; /* a counter to be used below */
     	       }

      for (k = 0, j = 0; k < numfaces; k++) {
	pamr_get_mesh_face_terms(m_refine, k+1, facet_data);
        if ( (int) facet_data[0] == 1)
           {           
	        *((grp_ptr->el_info)->surf_trac + ndof*j) = facet_data[1];
	        *((grp_ptr->el_info)->surf_trac + ndof*j+1) = facet_data[2];
	        *((grp_ptr->el_info)->surf_trac + ndof*j+2) = facet_data[3];
	        
	        pamr_element_face_side_indx( m_refine, k+1, inquire ) ;
	        
	        nel = inquire[0] ;
	        if(inquire[1]==1) side = 4 ;
	        else if(inquire[1]==2) side = 3 ;
	        else if(inquire[1]==3) side = 2 ;
	        else if(inquire[1]==4) side = 1 ;
	        
	        *((grp_ptr->el_info)->surf_list + 2*j) = nel ;
	        *((grp_ptr->el_info)->surf_list + 2*j + 1) = side ;
            j++;
           }
         else if ( (int) facet_data[0] == 2 &&
		   pamr_my_face_leader(m_refine, pamr_my_face_gid(m_refine, k+1)) )
           {
            pamr_element_face_side_indx( m_refine, k+1, inquire ) ;
	        nel = inquire[0] ;
	        if(inquire[1]==1) side = 4 ;
	        else if(inquire[1]==2) side = 3 ;
	        else if(inquire[1]==3) side = 2 ;
	        else if(inquire[1]==4) side = 1 ;

            kk = (int) facet_data[1] - 1 ;  /* which buoyancy group this is */
            buoy_ptr = (grp_ptr->el_info)->buoy + kk ;
            index = buoy_ptr->tally ;
            buoy_list = buoy_ptr->buoy_list ;
            *(buoy_list + 2*index) = nel ;
            *(buoy_list + 2*index + 1) = side ;
            (buoy_ptr->tally)++ ;
            }
     }



         } /* end Group Loop */
         
     /* handle split nodes.  Determine count, allocate space, fill space. */

     for (i = 0, ncount = 0; i < numnp; i++) {
          pamr_get_mesh_node_terms(m_refine, i+1, node_terms);
	  /* length_sqrd detects non-unit vectors resulting from taking the
             average between fault planes. This tries to ensure new nodes
             are not created off-plane. */
          length_sqrd = node_terms[6]*node_terms[6] +
                        node_terms[7]*node_terms[7] +
	                node_terms[8]*node_terms[8];
          if (length_sqrd > 0.9 && node_terms[10] > 0.9)
	    ncount++;
	  /*
	  is_split = (node_terms[10] > 0.9) ? 1 : 0;
	  ncount += is_split;
	  */
     }

     nfnode = ncount;

     fnode = ( int * ) calloc( nfnode , sizeof(int)) ;
             if (NULL == fnode) attempt = ELSTIFF_MEM  ;

     fvec = ( real * ) calloc( nfnode , 7*sizeof(real)) ;
             if (NULL == fvec) attempt = ELSTIFF_MEM  ;

     for (i = 0, ncount = 0; i < numnp; i++) {
          pamr_get_mesh_node_terms(m_refine, i+1, node_terms);
	  /* length_sqrd detects non-unit vectors resulting from taking the
             average between fault planes. This tries to ensure new nodes
             are not created off-plane. */
          length_sqrd = node_terms[6]*node_terms[6] +
                        node_terms[7]*node_terms[7] +
	                node_terms[8]*node_terms[8];
          if (length_sqrd > 0.9 && node_terms[10] > 0.9) {
	    /*
 	      ncount++;
	  is_split = (node_terms[10] > 0.9) ? 1 : 0;
	  if(is_split) {}
	    */
	    fnode[ncount] = i+1; /* fnode is a 1-based local node number */
	      for(j = 0; j < 7; j++)
		      fvec[7*ncount+j] = node_terms[j+3];
	      ncount ++;
	  }
     }
/* these need to be sorted so splitfind function will work correctly */
    for(i=0 ; i<nfnode ; i++)
      {
        for(j=i+1 ; j<nfnode ; j++)
	  {
            if( *(fnode+j) < *(fnode+i) )
              {
		itemp = *(fnode+i) ;
		*(fnode+i) = *(fnode+j) ;
		*(fnode+j) = itemp ;
		for(k=0 ; k<7 ; k++)
                  {
		    ftemp = *(fvec+7*i+k) ;
		    *(fvec+7*i+k) = *(fvec+7*j+k) ;
		    *(fvec+7*j+k) = ftemp ;
                  }
              }
	  }
      }

    /* ProcID, GnodeID, X, Y, Z, Bvecx, Bvecy, Bvecz, Svecx, Svecy, Svecz */
    /*
    printf("IAM: %d Old/New NFNODES: %d %d\n",pamr_my_pid(), (old_grp_ptr->el_info)->nfnode, nfnode);
    */

    /* cdn
    for (i=0; i < nfnode; i++){
      printf("%d %d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf\n",
	     pamr_my_pid(), pamr_my_node_gid( m_refine, i+1), global.coords[3*(fnode[i]-1)],
             global.coords[3*(fnode[i]-1)+1],global.coords[3*(fnode[i]-1)+2],
             fvec[7*i], fvec[7*i+1], fvec[7*i+2], fvec[7*i+3], fvec[7*i+4], fvec[7*i+5], fvec[7*i+6]);
    }
    */
      /* save this in attributes area. Prepares for *next* refinement.
         Make sure zeroed first! (at alloc point in inphase?)
      */

      /* Preserving number of split nodes */
      (grp_ptr->el_info)->nsplit = nfnode;

      for(i=0; i < nfnode; i++) {
        if (fnode[i] < 0) continue; /* probably  never happens, but may in gen_element */
	global.splitn[NSPLITATTR*(fnode[i]-1)] = fvec[7*i];
	global.splitn[NSPLITATTR*(fnode[i]-1)+1] = fvec[7*i+1];
	global.splitn[NSPLITATTR*(fnode[i]-1)+2] = fvec[7*i+2];
	global.splitn[NSPLITATTR*(fnode[i]-1)+3] = fvec[7*i+3];
	global.splitn[NSPLITATTR*(fnode[i]-1)+4] = fvec[7*i+4];
	global.splitn[NSPLITATTR*(fnode[i]-1)+5] = fvec[7*i+5];
	global.splitn[NSPLITATTR*(fnode[i]-1)+6] = fvec[7*i+6];
	global.splitn[NSPLITATTR*(fnode[i]-1)+7] = 1;
      }

      printf("Old/New Split Nodes: %d %d\n", (old_grp_ptr->el_info)->nsplit,
      (grp_ptr->el_info)->nsplit);

      temp_where = (int *) calloc(nfnode,2*sizeof(int)) ;
      temp_amount = (real *) calloc(nfnode,ndof*sizeof(real)) ;
      if (temp_where == NULL || temp_amount == NULL) attempt = SPLIT_MEM  ;
      completion() ;
      
      tally = 0 ;
      for( nel=0 ; nel<numel ; nel++ )
         {
          el_pt = grp_ptr->el_data + nel ;
          for(i=0 ; i<nen ; i++)
             {
              this_node = *(el_pt->ien + i) ;
                  j = splitfind(this_node,fnode,nfnode) ;
/* j is the zero-based index number of the searched-for split node input entry */

                  if(j != -1)
                     {
                      tally++ ;
					  temp_where = (int *) realloc(temp_where,2*sizeof(int)*tally) ;
					  temp_amount = (real *) realloc(temp_amount,ndof*sizeof(real)*tally) ;
					  if (temp_where == NULL || temp_amount == NULL) attempt = SPLIT_MEM  ;
                      
                      *(temp_where+2*(tally-1)) = nel ;         /* zero-based el# */
                      *(temp_where+2*(tally-1)+1) = this_node ; /* one-based node# */
                      
                      count = (*(el_pt->is_split)) + 1 ;
                      *(el_pt->is_split) = count ;
                      *(el_pt->is_split + count) = tally-1 ;
 /* flag and indices set for easy lookup of split nodes later during stress calc */

                      ivec[0] = *(fvec + 7*j) ;
                      ivec[1] = *(fvec + 7*j + 1) ; /* ivec is the direction perp to slip */
                      ivec[2] = *(fvec + 7*j + 2) ;
                      svec[0] = *(fvec + 7*j + 3) ;
                      svec[1] = *(fvec + 7*j + 4) ; /* svec is the slip vector direction */
                      svec[2] = *(fvec + 7*j + 5) ;
                      slip    = *(fvec + 7*j + 6) ; /* slip is the full slip amplitude */
                     
                      norm = (float)sqrt(ivec[0]*ivec[0] + ivec[1]*ivec[1] + ivec[2]*ivec[2]) ;
                      if((float)fabs(norm - ONE) > 1.0e-03)
                         {
                          squawk("ivec not normalized!\n");
                          printf("node #%d , splitnode #%d , instance #%d\n",
                                 this_node,j+1,tally) ;
                          printf("ivec = [ %g %g %g ]\n",ivec[0],ivec[1],ivec[2]) ;
                          printf("|ivec| = %g\n", norm );
                          attempt = IVEC_NORM ;
                         }
                      norm = sqrt(svec[0]*svec[0] + svec[1]*svec[1] + svec[2]*svec[2]) ;
                      if(fabs(norm - ONE) > 1.0e-03)
                         {
                          squawk("svec not normalized!\n");
                          printf("node #%d , splitnode #%d , instance #%d\n",
                                 this_node,j+1,tally) ;
                          printf("svec = [ %g %g %g ]\n",svec[0],svec[1],svec[2]) ;
                          printf("|svec| = %g\n", norm );
                          attempt = SVEC_NORM ;
                         }
                      if(fabs(ivec[0]*svec[0]+ivec[1]*svec[1]+ivec[2]*svec[2]) > 1.0e-03)
                         {
                          squawk("ivec and svec not perpendicular!\n");
                          attempt = ISVEC_PERP ;
                         }
                     
                      cross[0] = svec[1]*ivec[2] - svec[2]*ivec[1] ;
                      cross[1] = svec[2]*ivec[0] - svec[0]*ivec[2] ;
                      cross[2] = svec[0]*ivec[1] - svec[1]*ivec[0] ;
                      
                      xctr = ZERO ;
                      yctr = ZERO ;
                      zctr = ZERO ;
                      for(k=0 ; k<nen ; k++)
                         {
                          xctr += *(global.coords + nsd * ( *(el_pt->ien + k) - 1 ) ) ;
                          yctr += *(global.coords + nsd * ( *(el_pt->ien + k) - 1 ) + 1 ) ;
                          if(nsd == 3)
                          zctr += *(global.coords + nsd * ( *(el_pt->ien + k) - 1 ) + 2 ) ;
                         }
                      xctr = xctr/(real)nen - *(global.coords + nsd * ( this_node - 1 ) ) ;
                      yctr = yctr/(real)nen - *(global.coords + nsd * ( this_node - 1 ) + 1 ) ;
                      if(nsd == 3)
                      zctr = zctr/(real)nen - *(global.coords + nsd * ( this_node - 1 ) + 2 ) ;
                      
                      dot = cross[0]*xctr + cross[1]*yctr + cross[2]*zctr ;
                      if( dot > ZERO )
                         mult = slip * PT5 ;
                      else
                         mult = -slip * PT5 ;
                      switch( ndof )
                         {
                          case 1:
                              *(temp_amount+ndof*(tally-1)) = svec[2]*mult ;
                              break ;
                          case 2:
                              *(temp_amount+ndof*(tally-1)) = svec[0]*mult ;
                              *(temp_amount+ndof*(tally-1)+1) = svec[1]*mult ;
                              break ;
                          case 3:
                              *(temp_amount+ndof*(tally-1)) = svec[0]*mult ;
                              *(temp_amount+ndof*(tally-1)+1) = svec[1]*mult ;
                              *(temp_amount+ndof*(tally-1)+2) = svec[2]*mult ;
                              break ;
                         }

                     }
                  
             }     /* loop i over element nodes */
         }         /* loop nel over elements */
       completion() ;
         

      (grp_ptr->el_info)->nfnode = tally ; /* nfnode now contains the total number of instances! */
      (grp_ptr->el_info)->slip_list = temp_where ;
      (grp_ptr->el_info)->slip_val = temp_amount ;
      free( fnode ) ;
      free( fvec ) ;
      squawk("Split nodes all processed . . .\n") ;

      if (pamr_my_pid() == 0)
         printf("\nGeneration of element group #%d complete.\n",
             grp_ptr->group_num) ;
      fflush(stdout) ;
     
      elgrp_loop(m_refine, SHAPE);

      free(force.full_rhs);
      free(force.incr_rhs);
      free(force.last_result);
      free(force.first_result);
      free(force.net_external);
      free(pcg.precond);
      free(pcg.r);
      free(pcg.d);
      free(pcg.temp);
      free(pcg.hist);
      matrix_alloc(m_refine);

     /* nprt_nodes needs to be treated reasonably (where set?) */


  /* Thoughts

  1. Retrieve global number of nodes from PYRAMID pamr_global_node_count(m_in)
           tot_nodes = PAMR_TOTAL_NODE_COUNT(m_in)
           tot_elements = PAMR_TOTAL_ELEMENT_COUNT(m_in)

  2. Retrieve local number of nodes for allocations pamr_local_node_count(m_in)

    function PAMR_LOCAL_NODE_COUNT( m_in ) RESULT(nc)
    function PAMR_GET_LOCAL_NODES( m_in ) RESULT(terms)

     - global node IDs returned

  3. Set numnp = pamr_local_node_count(m_in)

  4. Set nsd = fe_sys.nsd , ndof = fe_sys.ndof

  5. Allocate (as real pointers) new_coords, new_displ new_del_displ, new_forv
     (int*) new_eqtype - when taken from attributes of size numnp*nsd and numnp*ndof

  6. Not yet supporting nrates > 0

  7. Fill new_eqtype array
  * Copy attributes from coarse mesh to refined mesh appropriately
     - Branch for newly created nodes and fill in new eqtype values as a function
       of nodes on the shared/split edge

  8. Call gen_number to determine local eqn. numbers

  9. Fill new_coords with values from coarse mesh

    function PAMR_MESH_COORDINATES( m_in ) RESULT(nc)
      real, dimension(5,this%loc_nodes) :: nc

    function PAMR_LOCAL_NODE_COORDS( m_in ) RESULT(nc)
      real, dimension(4,this%loc_nodes) :: nc

    - Ordered by node index. Another command can return coordinates based
      on elements if needed...

 10. [Future] Fill in displacement boundaries new_forv

 11. Allocations for elements

    function PAMR_LOCAL_ELEMENT_COUNT( m_in ) RESULT(ec)
    function PAMR_GET_LOCAL_ELEMENTS( m_in ) RESULT(terms)

    - Global element IDs returned

 12. Acquire new connectivity information

     - Can come as local data from PAMR_VISUALIZE in the format:
       Two separate arrays of data can be returned.
       node1 x,y,z
       noden x,y,z
       elem1 n1, n2, n3, n4
       elemm n1, n2, n3, n4

    - Or one can use the global IDs of nodes and elements

      function PAMR_MESH_CONNECTION( m_in ) RESULT(nc)
        integer, dimension(5,this%loc_elements) :: nc

      function PAMR_LOCAL_NODE_GIDS( m_in ) RESULT(ngids)
        integer, dimension(this%loc_nodes) :: ngids

      function PAMR_ELEMENT_NODE_CONNECT( m_in ) RESULT(nc)
         real, dimension(4,this%loc_elements) :: nc


 13. Fill in material index (copy from coarse mesh for old elements and
     determine from parent element for new elements)
     - is this inherited from the coarse mesh automatically by default.

     * Coarse/Parent mesh information is transported under repartitioning.
       Refined/Child mesh needs to inquire data from coarse/parent mesh.

 14. Treat split nodes and generate related element records

 15. Return and call elastic

 16. Generate Tecplot graphics

    call PAMR_VISUALIZE( m_in, "mesh.plt" )                                                                                   
  */
/* PRETTY SURE ALL THESE LINES ARE DUPLICATED AND SHOULD BE DELETED to XXX */
/* We agree and code was deleted */

    squawk("Starting Cleanup...\n");

    for(i=0 ; i<global.n_group ; i++)
        {
            old_grp_ptr = &old_groups[i];
            numel = (old_grp_ptr->el_info)->numel;
	    for( nel=0 ; nel<numel ; nel++ )
	    {
	       el_pt = old_grp_ptr->el_data + nel ;
               free(el_pt->shape);
               free(el_pt->deter);
               free(el_pt->stiff); /* assuming pcg */
               plastic = *( (old_grp_ptr->el_mat) -> plastic + el_pt->mat );
               if(plastic) free(el_pt->bta); 
               if(plastic) free(el_pt->dbar); 
               free(el_pt->is_split);
               free(el_pt->dilat);
               free(el_pt->stress);
               free(el_pt->lm);
               free(el_pt->ien);
            } 
 
             /* (do *NOT* free old_grp_ptr->el_mat: 
                 we pass that table by pointer.) */

            free(old_grp_ptr->el_data);

             /* (do *NOT* free temp_amount, temp_where: 
                 Those are ptr copied to el_info->slip_val, slip_list.
                 Note also that fnode, fvec are freed above.) */

            free(old_grp_ptr->el_info->slip_val);
            free(old_grp_ptr->el_info->slip_list);
            free(old_grp_ptr->el_info);
        }

/*if (nrates > 0)  (this isn't yet supported, but will be done soon)
 (this isn't yet supported, but will be done soon)
free (old_global.rate);
free (old_global.bctime);
*/
    squawk("Starting Final Free...\n");

     free(old_global.owned_eqs);
     free(old_global.id_pointer);
     free(old_global.permut);
     free(old_global.splitn);
     free(old_global.eqtype);
     free(old_global.forv);
     free(old_global.del_displ);
     free(old_global.displ);
     free(old_global.coords);
     free(old_groups);
     free(node_terms);
     globalize(m_coarse, (real *) NULL, 1);
}
/*************************** end of update_phase **********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*                                                                                                                                  
*************   ROUTINE:   terminate   ***************                                                                          
*/

/*  INTERFACE:   */

void
terminate(int iam)

     /*  RETURN VALUE:  -none- */
     /*  DESCRIPTION:   */
     /*                                                                                                                                  
     ** Routine terminate                                                                                                               
     ** Normal exit of the program from simulation flow                                                                        
     **/
     /*   EOP   */
     /*---------------------------------------------------------------------------*/
{
       fclose(in_file) ;
       
       if (iam == 0) {
	                  fclose(out_file) ;
	                 }
       squawk("Normal program termination.\n") ;

       if (iam == 0) {
                      printf("CPU time = %f\nFinishing Time_Step Soln At: ",
                      (double)clock()/CLOCKS_PER_SEC);
                     }
       pamr_current_time();

	   pamr_finalize(TRUE);

	   exit( 0 );

}
/*************************** end of terminate **********************************/


int MAIN__(void) { return 0; }
void mpi_init_(int ierror) { }

