/*
***                           File main.h                         ***
***                        GeoFEST version 4.5
*** Copyright (c) 2003, California Institute of Technology        ***
*** U.S.Sponsorship under NASA Contract NAS7-1260 is acknowledged ***
***
*** This software is designated for public release under JPL Task ***
*** Plan Number 83-6791 and may be publicly released after        ***
*** approval for release has been granted by GSFC Software        ***
*** Release  Authority
***
*** This file contains the program main entry point, the main
*** task driver, and modules for driving high-level functions
*** and interactions with the operator:
***
***          == main ==
***          == elastic ==
***          == time_step ==
***          == clear_stiff ==
***          == elgrp_loop ==
***          == node_load ==
***          == accumulate ==
***          == completion ==
***          == wrt_save ==
***          == vrestart ==
***          == el_save ==
***          == check_monitor ==
***/

/*
***GeoFEST version 4.5***
G. A. Lyzenga -- unified parallel/serial version and bug fixes 11/17/03
G. A. Lyzenga -- split node enhancements 8/16/02
G. A. Lyzenga -- modification to correctly relax hydrostatic stresses 6/15/00
J. W. Parker -- 3D support and deleted hypercube branches 9/16/98
G. A. Lyzenga  -- last modified on 3/03/95
Modification history:
2.1: fixes bug in boundary velocities and changes
     the way stresses are output.
2.2: implements split node faults
2.3: adds pass-thru comments
3.0: 3-D tetrahedra supported, hypercube branches deleted
3.1: hydrostatic relaxation changes
4.0: name change from visco to GeoFEST; correct implementation of 2-D out-of-
     plane stresses as well as hydrostatic relaxation; culling of remaining
     vestiges of hypercube stuff; one code without recompilation for both
     2-D and 3-D domains
4.1: Introduction of trilinear hexahedral bricks and PCG iterative solver option
4.2: Bug fixes; standard source prologues added (June 02)
4.3: Corrects inefficiency in split node algorithm and introduces new
     split node input format
4.4: Adds dummy calls to MPI and Pyramid to provide a single unified
     source for parallel and serial codes.  Also introduces a major
     algorithmic enhancement; stress equilibrium maintenance during visco
     elastic stepping for long-term stability.  Also more minor bug fixes.
4.5: Supports surface tractions and roller/spring duo options for tet elements.
     Provides enhanced conjugate gradient convergence monitoring.
 $Log: main.h,v $
 Revision 1.1  2003/02/09 06:10:55  lyzenga
 *** empty log message ***

*/

#include <stdio.h>
#include <stdlib.h>
#include "mydefs.h"
#include "finel.h"
#include "errcodes.h"
#include <math.h>
#include <time.h>

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   elastic   ***************
*/

/*  INTERFACE:   */

void
elastic(
  pamr_Mesh m[]      /* parallel mesh */
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine elastic
** elastic performs an elastic solution of the
** finite element problem.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/*************************** end of elastic **********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   time_step   ***************
*/

/*  INTERFACE:   */

void
time_step(
  pamr_Mesh m[]
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine time_step
** time_step computes the visco-elastic time stepping solutions
** to the finite element problem.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/*************************** end of time_step ********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   clear_stiff   ***************
*/

/*  INTERFACE:   */

void
clear_stiff(
  void
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine clear_stiff
** clear_stiff nulls the stiffness matrix memory.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/*************************** end of clear_stiff ******************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   elgrp_loop   ***************
*/

/*  INTERFACE:   */

void
elgrp_loop(
  pamr_Mesh m[], /* parallel mesh */
  int	task	/* which element-level task to perform			*/
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine elgrp_loop
** elgrp_loop invokes a particular task that relies on
** element structures and affects all the elements in
** every element group.
**
** Supported tasks are
** GENERATE: read or generate element data, fill ien, lm arrays
** OUTPUT: compute and write out element centroid stresses
** SHAPE: compute parent-space shape functions and gradients
** FORMS_ELAS: compute elastic stiffness matrix
** FORMS_STEP: compute VE single-step stiffness matrix
** RHS_ELAS: compute elastic right-hand side vector for FE problem
** RHS_STEP: compute VE right-hand side vector for FE problem
** BC_ELAS:  compute boundary-condition terms for elastic FE problem
** BC_STEP: compute boundary-condition terms for VE FE single step
** COL_HT: compute the column heights for profile stiffness storage
** E_STRESS: compute the stresses from elastic solution
** V_STRESS: compute the stresses from a VE step solution
** DUMP: dump information needed for restarts
** RESTORE: read and restore solution vector and info for a restart
** FAIL_CHECK: check for stress-driven fault failure
** SMOOTH_BEGIN: perform pressure smoothing set-up
** SMOOTH_END: apply pressure-smoothing function from coefficients
** REORDER: analyze adjacency, find profile-optimizing node permutation
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/*************************** end of elgrp_loop *******************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   node_load   ***************
*/

/*  INTERFACE:   */

void
node_load(
  pamr_Mesh m[], /* parallel mesh */
  real	time,	/* current simulation time				*/
  int	reform	/* flag indicating stiffness reform status		*/
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine node_load
** node_load applies node-based forces that may accumulate
** with time.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/*************************** end of node_load ********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   accumulate   ***************
*/

/*  INTERFACE:   */

void
accumulate(
  void
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine accumulate
** accumulate updates the total displacement based on the
** single-step increment.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/*************************** end of accumulate *******************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   completion   ***************
*/

/*  INTERFACE:   */

void
completion(
  void
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine completion
** completion syncronizes multiple processors.
** Keeps track of incurred errors in each processor.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/*************************** end of completion *******************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   wrt_save   ***************
*/

/*  INTERFACE:   */

void
wrt_save(
  pamr_Mesh m[]
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine wrt_save
** wrt_save dumps state to disk for restarts.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/*************************** end of wrt_save *********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   vrestart   ***************
*/

/*  INTERFACE:   */

void
vrestart(
	 pamr_Mesh m[]
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** vrestart restores state from previous dump and
** continues simulation.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/*************************** end of vrestart *********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   el_save   ***************
*/

/*  INTERFACE:   */

void
el_save(
  GROUP*	grp_ptr,	/* pointer to current element group	*/
  int		code		/* code specifying which action		*/
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine el_save
** el_save dumps or restores element data.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/**************************** end of el_save *********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   check_monitor   ***************
*/

/*  INTERFACE:   */

     void
     check_monitor(
                   pamr_Mesh m[], /* parallel mesh */
                   int   nstep    /* count of steps for backup */
                  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine check_monitor
** check_monitor allows user run intervention via strings stored
** in the file monitor.fem.
**
** Supported options:
** OK: close the dump file (?)
** KILL: close files and abort
** SAVE_STOP: write state, then abort
** SAVE_GO:  write dumps as requested in input, and continue
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of check_monitor *****************************/
/*  INTERFACE:   */

void
get_refine(pamr_Mesh m_in[], pamr_Mesh m_out[]);

     /*  RETURN VALUE:  -none- */
     /*  DESCRIPTION:   */
     /*                                                                                                                                  
     ** Routine get_refine                                                                                                               
     ** performs adaptive mesh refinement based on strain energy.                                                                        
     ** Only works for one elment group.                                                                                                 
     **/
     /*   EOP   */
     /*---------------------------------------------------------------------------*/

/*  INTERFACE:   */

void
save_attributes(pamr_Mesh m[]);

     /*  RETURN VALUE:  -none- */
     /*  DESCRIPTION:   */
     /*                                                                                                                                  
     ** Routine save_attributes

     ** saves GeoFEST problem description values as PYRAMID attributes

     ** Only works for one elment group.                                                                                                 
     **/
     /*   EOP   */
     /*---------------------------------------------------------------------------*/

/*  INTERFACE:   */

void
update_phase(pamr_Mesh m_coarse[], pamr_Mesh m_refine[]);

     /*  RETURN VALUE:  -none- */
     /*  DESCRIPTION:   */
     /*

     ** Routine update_phase

     ** comment on updating

     ** Only works for one elment group.

     **/
     /*   EOP   */
     /*---------------------------------------------------------------------------*/

/*  INTERFACE:   */

void
terminate(int iam);

     /*  RETURN VALUE:  -none- */
     /*  DESCRIPTION:   */
     /*

     ** Routine terminate

     ** Normal exit of the program from simulation flow

     **/
     /*   EOP   */
     /*---------------------------------------------------------------------------*/
