/*
***                         File solver.c                         ***
***                        GeoFEST version 4.5
*** Copyright (c) 2003, California Institute of Technology        ***
*** U.S.Sponsorship under NASA Contract NAS7-1260 is acknowledged ***
***
*** This software is designated for public release under JPL Task ***
*** Plan Number 83-6791 and may be publicly released after        ***
*** approval for release has been granted by GSFC Software        ***
*** Release  Authority
***
*** This file contains modules for the sparse matrix storage,
*** substructure reduction or other solvers of the matrix system:
***
***          == solver ==
***          == addstiff ==
***          == addfor ==
***          == estiffprod ==
***          == reorder ==
***          == genrcm ==
***          == rcm ==
***          == degree ==
***          == fnroot ==
***          == rootls ==
***          == colht ==
***          == profile_diag ==
***          == factor ==
***          == full_back ==
***          == pcg_loop ==
***          == converged ==
***          == put_soln ==
***          == globalize ==
***          == dot_par ==
***/

/*
***GeoFEST version 4.5***
G. A. Lyzenga -- unified parallel/serial version and bug fixes 11/17/03
G. A. Lyzenga -- split node enhancements 8/16/02
G. A. Lyzenga -- modification to correctly relax hydrostatic stresses 6/15/00
J. W. Parker -- 3D support and deleted hypercube branches 9/16/98
G. A. Lyzenga  -- last modified on 3/03/95
Modification history:
2.1: fixes bug in boundary velocities and changes
     the way stresses are output.
2.2: implements split node faults
2.3: adds pass-thru comments
3.0: 3-D tetrahedra supported, hypercube branches deleted
3.1: hydrostatic relaxation changes
4.0: name change from visco to GeoFEST; correct implementation of 2-D out-of-
     plane stresses as well as hydrostatic relaxation; culling of remaining
     vestiges of hypercube stuff; one code without recompilation for both
     2-D and 3-D domains
4.1: Introduction of trilinear hexahedral bricks and PCG iterative solver option
4.2: Bug fixes; standard source prologues added (June 02)
4.3: Corrects inefficiency in split node algorithm and introduces new
     split node input format
4.4: Adds dummy calls to MPI and Pyramid to provide a single unified
     source for parallel and serial codes.  Also introduces a major
     algorithmic enhancement; stress equilibrium maintenance during visco
     elastic stepping for long-term stability.  Also more minor bug fixes.
4.5: Supports surface tractions and roller/spring duo options for tet elements.
     Provides enhanced conjugate gradient convergence monitoring.
 $Log: solver.c,v $
 Revision 1.4  2003/02/09 06:10:56  lyzenga
 *** empty log message ***

*/

/*
#include <stdio.h>
#include "mydefs.h"
#include "errcodes.h"
#define EXTERN extern
#include "finel.h"
#include <math.h>
*/
#define EXTERN extern
#include "solver.h"
#include "utility.h"

void elgrp_loop(
                pamr_Mesh m[],  /* the mesh */
                int	task	/* which element-level task to perform */
               );

#define ES(i,j) *(es + nee*(j) + (i))

static int   first_time = TRUE ;

int compare_integers (const int *a, const int *b) {
  return (int) (*a - *b);
}

/*
static int   did_it = 0 ;
*/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   solver   ***************
*/

/*  INTERFACE:   */

     void
     solver( pamr_Mesh m[], /* the mesh */
            int  code     /* which solver option to use */
           )

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine solver
** solver performs single right-hand-side matrix solution,
** currently using factorization and back substitution on 
** single processor OR using PCG iteration.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

    {
     char CVS_Id[] = "@(#) CVS id $Id: solver.c,v 1.4 2003/02/09 06:10:56 lyzenga Exp $, Revision: " REVNO;

     switch(code)
        {
         case ITER:
           pcg_loop(m);
           break;
           
         case FACBACK:
         case BACK:
          if( code == FACBACK )
           factor(stiff.matrix , stiff.diag , fe_sys.neq ) ;
     
           full_back(stiff.matrix , force.full_rhs , stiff.diag ,
               fe_sys.neq) ;     
           break;
        }

     put_soln( &force , global.id_pointer ) ;
    }

/************************** end of solver ************************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   addstiff   ***************
*/

/*  INTERFACE:   */

   void
   addstiff(
            PROFILE  *a ,  /* ptr to assembled profile stiffness storage */
            real  *es ,        /* ptr to element stiffness storage */
            ELEMENT_INFO  *info ,   /* ptr to element info struct */
            ELEMENT_DATA  *el_pt    /* ptr to current element struct */
           )

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine addstiff
** addstiff adds element stiffness to global profile array,
** or to element storage if using PCG solver
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
     
      {
       int  i_eq , j_eq , i , j , iadd , nee ;
       real  *stiff , *M ;

       nee = info->nee ;
       
       switch(fe_sys.solver)
		  {
		   case DIRECT:
		   for (  j = 0 ; j < nee ; j++ )
			  {
			   if( (j_eq = (int) el_pt->lm[j]) != -1 )
				 {
				  for( i = 0 ; i <= j ; i++ )
					  {
					   if( (i_eq = (int) el_pt->lm[i]) != -1 )
						 {
						  if(j_eq >= i_eq)
								{
								 iadd = (int) *((a->diag) + j_eq ) -
										j_eq + i_eq ;
								 *((a->matrix)+iadd) +=  ES(i,j) ;
								 }
							  else
								  {
									iadd = (int) *((a->diag) + i_eq ) -
												   i_eq + j_eq ;
								   *((a->matrix)+iadd) +=  ES(i,j) ;
								  } 
						 
						 }
					 }
				  }
			   }
		   break ;
		   
		   case PCG:
		   stiff = el_pt->stiff ;
		   M = pcg.precond ;
		   for (  j = 0 ; j < nee ; j++ )
			   {
				  for( i = 0 ; i <= j ; i++ )
					 {
                      *(stiff + nee*j + i) = ES(i,j) ; /* store element stiffness */
                      
                      if( i == j && ( (i_eq = (int) el_pt->lm[i]) != -1 ) )
                         {
         /* this is a diagonal element -- use in preconditioner */
			   *(M+i_eq) += ES(i,j) ;

		   /* Replaced by direct value of partial contributions
                      to be inverted later....
                          if(*(M+i_eq) == ZERO)
                             {
                              *(M+i_eq) = ONE/ES(i,j) ;
                             }
                          else
                             {
                              *(M+i_eq) = ONE/((ONE/ *(M+i_eq)) + ES(i,j)) ;
                             }
		   */

                         }
					 }
			   }
		   break ;
	
		  } /* end switch */
			   
                 
   }

/************************** end of addstiff **********************************/ 

 
/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   addfor   ***************
*/

/*  INTERFACE:   */

     void
     addfor(
            real  *dest ,  /* destination assembled rhs vector */
            real  *rhs ,        /* source element rhs vector */
            ELEMENT_INFO  *info ,   /* ptr to element info struct */
            ELEMENT_DATA  *el_pt    /* ptr to current element struct */
           )

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine addfor
** addfor assembles the global r.h.s. vector.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
      COUNTER      i    ;
      int          k    ;


           for ( i = 0 ; i < info->nee ; i++ )
               {
                k = el_pt->lm[i] ;
                if( k != (-1) )
                   {
                     *(dest + k) += rhs[i] ;
                   }
               }
     }

/************************** end of addfor ************************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   estiffprod   ***************
*/

/*  INTERFACE:   */

 void
 estiffprod(
            GROUP   *grp_ptr    /* pointer to current element group */
           )

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine estiffprod
** calculates the product of a given vector with stiffness in element storage
**    t = A * d
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

  {
   int     numel , nel , nee , neq , i , j , i_eq , j_eq ;
   real    *t , *d , *stiff, *piton, *crampon ;
   int *llm;
   ELEMENT_DATA   *el_pt ,*el_end;

   numel = (grp_ptr->el_info)->numel ;
   nee = (grp_ptr->el_info)->nee ;
   neq = fe_sys.neq ;
   t = pcg.temp ; /* destination vector */
   d = pcg.d ;  /* multiplicand vector */
   el_pt = grp_ptr->el_data;
   el_end = el_pt + numel; 
   for (  ; el_pt < el_end ; el_pt++ )
      {
       stiff = el_pt->stiff ;
       llm = el_pt->lm;
       if(el_pt->is_bc == 1 || nee != 12)
          {
          for (  j = 0 ; j < nee ; j++ )
             {
              if( (j_eq = llm[j]) != -1 )
                 {
                  for( i = 0 ; i < j ; i++ )
                     {
                      if( (i_eq =  llm[i]) != -1 )
                         {
                          *(t+i_eq) += *(d+j_eq) * *(stiff + nee*j + i) ;
                          *(t+j_eq) += *(d+i_eq) * *(stiff + nee*j + i) ;
                         }
                     } /* i loop */
                  *(t+j_eq) += *(d+j_eq) * *(stiff + nee*j + j) ;
                 }
             } /* j loop */
          } else {
                     /* Low-operation in-line commands for special case
                        of no boundary, nee==12

                       Following the older code, the assumptions here are
                       a) stiff is allocated as a square, 12x12
                       b) only the upper "half" gets used (i <= j)
                       c) matrix is to be conceived as running down columns,
                          then across rows.  So "i" marks row.
                      */
                  real d0, d1, d2, d3, d4, d5, d6, d7, d8, d9, d10, d11;

                  
                  *(t+llm[0]) += (d0 = *(d+llm[0])) * *(stiff)
                              +  (d1 = *(d+llm[1])) * *(stiff +  nee) 
                              +  (d2 = *(d+llm[2])) * *(stiff + nee*2) 
                              +  (d3 = *(d+llm[3])) * *(stiff + nee*3) 
                              +  (d4 = *(d+llm[4])) * *(stiff + nee*4) 
                              +  (d5 = *(d+llm[5])) * *(stiff + nee*5) 
                              +  (d6 = *(d+llm[6])) * *(stiff + nee*6) 
                              +  (d7 = *(d+llm[7])) * *(stiff + nee*7) 
                              +  (d8 = *(d+llm[8])) * *(stiff + nee*8) 
                              +  (d9 = *(d+llm[9])) * *(stiff + nee*9) 
                              +  (d10 = *(d+llm[10])) * *(stiff + nee*10) 
                              +  (d11 = *(d+llm[11])) * *(stiff + nee*11); 

                    /* 
                       piton, crampon: pointers to landmark locations within the 
                       stiffness triangular matrix.

                       piton marks the top of the current column.

                       crampon marks the left edge of the current row.
                    */

                  piton = stiff + nee;
                  crampon = stiff + 1;
                  *(t+llm[1]) += d0 * *piton
                              + d1 * *(piton + 1) 
                              + d2 * *(crampon + nee*2 ) 
                              + d3 * *(crampon + nee*3 )
                              + d4 * *(crampon + nee*4 )
                              + d5 * *(crampon + nee*5 )
                              + d6 * *(crampon + nee*6 )
                              + d7 * *(crampon + nee*7 )
                              + d8 * *(crampon + nee*8 )
                              + d9 * *(crampon + nee*9 )
                              + d10 * *(crampon + nee*10 )
                              + d11 * *(crampon + nee*11 );
                  piton += nee;
                  crampon++;
                  *(t+llm[2]) += d0 * *piton
                              + d1 * *(piton + 1) 
                              + d2 * *(piton + 2) 
                              + d3 * *(crampon + nee*3 )
                              + d4 * *(crampon + nee*4 )
                              + d5 * *(crampon + nee*5 )
                              + d6 * *(crampon + nee*6 )
                              + d7 * *(crampon + nee*7 )
                              + d8 * *(crampon + nee*8 )
                              + d9 * *(crampon + nee*9 )
                              + d10 * *(crampon + nee*10 )
                              + d11 * *(crampon + nee*11 );
                  piton += nee;
                  crampon++;
                  *(t+llm[3]) += d0 * *piton
                              + d1 * *(piton + 1) 
                              + d2 * *(piton + 2) 
                              + d3 * *(piton + 3)
                              + d4 * *(crampon + nee*4 )
                              + d5 * *(crampon + nee*5 )
                              + d6 * *(crampon + nee*6 )
                              + d7 * *(crampon + nee*7 )
                              + d8 * *(crampon + nee*8 )
                              + d9 * *(crampon + nee*9 )
                              + d10 * *(crampon + nee*10 )
                              + d11 * *(crampon + nee*11 );
                  piton += nee;
                  crampon++;
                  *(t+llm[4]) += d0 * *piton
                              + d1 * *(piton + 1) 
                              + d2 * *(piton + 2) 
                              + d3 * *(piton + 3)
                              + d4 * *(piton + 4)
                              + d5 * *(crampon + nee*5 )
                              + d6 * *(crampon + nee*6 )
                              + d7 * *(crampon + nee*7 )
                              + d8 * *(crampon + nee*8 )
                              + d9 * *(crampon + nee*9 )
                              + d10 * *(crampon + nee*10 )
                              + d11 * *(crampon + nee*11 );
                  piton += nee;
                  crampon++;
                  *(t+llm[5]) += d0 * *piton
                              + d1 * *(piton + 1) 
                              + d2 * *(piton + 2) 
                              + d3 * *(piton + 3)
                              + d4 * *(piton + 4)
                              + d5 * *(piton + 5)
                              + d6 * *(crampon + nee*6 )
                              + d7 * *(crampon + nee*7 )
                              + d8 * *(crampon + nee*8 )
                              + d9 * *(crampon + nee*9 )
                              + d10 * *(crampon + nee*10 )
                              + d11 * *(crampon + nee*11 );
                  piton += nee;
                  crampon++;
                  *(t+llm[6]) += d0 * *piton
                              + d1 * *(piton + 1) 
                              + d2 * *(piton + 2) 
                              + d3 * *(piton + 3)
                              + d4 * *(piton + 4)
                              + d5 * *(piton + 5)
                              + d6 * *(piton + 6)
                              + d7 * *(crampon + nee*7 )
                              + d8 * *(crampon + nee*8 )
                              + d9 * *(crampon + nee*9 )
                              + d10 * *(crampon + nee*10 )
                              + d11 * *(crampon + nee*11 );
                  piton += nee;
                  crampon++;
                  *(t+llm[7]) += d0 * *piton
                              + d1 * *(piton + 1) 
                              + d2 * *(piton + 2) 
                              + d3 * *(piton + 3)
                              + d4 * *(piton + 4)
                              + d5 * *(piton + 5)
                              + d6 * *(piton + 6)
                              + d7 * *(piton + 7)
                              + d8 * *(crampon + nee*8 )
                              + d9 * *(crampon + nee*9 )
                              + d10 * *(crampon + nee*10 )
                              + d11 * *(crampon + nee*11 );
                  piton += nee;
                  crampon++;
                  *(t+llm[8]) += d0 * *piton
                              + d1 * *(piton + 1) 
                              + d2 * *(piton + 2) 
                              + d3 * *(piton + 3)
                              + d4 * *(piton + 4)
                              + d5 * *(piton + 5)
                              + d6 * *(piton + 6)
                              + d7 * *(piton + 7)
                              + d8 * *(piton + 8)
                              + d9 * *(crampon + nee*9 )
                              + d10 * *(crampon + nee*10 )
                              + d11 * *(crampon + nee*11 );
                  piton += nee;
                  crampon++;
                  *(t+llm[9]) += d0 * *piton
                              + d1 * *(piton + 1) 
                              + d2 * *(piton + 2) 
                              + d3 * *(piton + 3)
                              + d4 * *(piton + 4)
                              + d5 * *(piton + 5)
                              + d6 * *(piton + 6)
                              + d7 * *(piton + 7)
                              + d8 * *(piton + 8)
                              + d9 * *(piton + 9)
                              + d10 * *(crampon + nee*10 )
                              + d11 * *(crampon + nee*11 );
                  piton += nee;
                  crampon++;
                  *(t+llm[10]) += d0 * *piton
                              + d1 * *(piton + 1) 
                              + d2 * *(piton + 2) 
                              + d3 * *(piton + 3)
                              + d4 * *(piton + 4)
                              + d5 * *(piton + 5)
                              + d6 * *(piton + 6)
                              + d7 * *(piton + 7)
                              + d8 * *(piton + 8)
                              + d9 * *(piton + 9)
                              + d10 * *(piton + 10)
                              + d11 * *(crampon + nee*11 );
                  piton += nee;
                  crampon++;
                  *(t+llm[11]) += d0 * *piton
                              + d1 * *(piton + 1) 
                              + d2 * *(piton + 2) 
                              + d3 * *(piton + 3)
                              + d4 * *(piton + 4)
                              + d5 * *(piton + 5)
                              + d6 * *(piton + 6)
                              + d7 * *(piton + 7)
                              + d8 * *(piton + 8)
                              + d9 * *(piton + 9)
                              + d10 * *(piton + 10)
                              + d11 * *(piton + 11);

          }
      } /* nel loop */

  }

/************************** end of estiffprod ********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   reorder   ***************
*/

/*  INTERFACE:   */

    void
    reorder(
            GROUP   *grp_ptr    /* pointer to current element group */
           )

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine reorder
** reorder uses ien information to build adjacency information and call
** permutation optimizing routines for minimizing matrix profile.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

{
   int *id;
   int index,
       *icol,
       *irow,
       *llist,
       *msk,
       *xls,
       *invper,
       nnodes = loc_sys.numnp,
       max_nz = MAX_CONNECT*nnodes,
       lfree,
       nel,
       idof,ndof,
       next_icol,
       num_col,
       total_nonzeros,
       pgnode,
       next_eq,
       this_link, next_link, ref_lnode, ref_gnode, aux_lnode, aux_gnode;

   ELEMENT_INFO  *info ;
   ELEMENT_DATA  *el_pt ;

/* FILE *tfp; */

   info = grp_ptr->el_info ;

/* use ien data to fill irow, icol arrays */

   global.permut = (int *) calloc(nnodes+1,sizeof(int));
   icol = (int *) calloc(max_nz,sizeof(int));
   irow = (int *) calloc(nnodes + 2,sizeof(int));
   llist = (int *) calloc(2*max_nz,sizeof(int));
   invper = (int *) calloc(nnodes+1,sizeof(int));
   msk = (int *) calloc(nnodes+1,sizeof(int));
   xls = (int *) calloc(nnodes+1,sizeof(int));


   lfree = nnodes + 1;

/* compile adjacency information from every element. . . */
   for( nel = 0 ;  nel < info->numel ; nel++ )
   {
      el_pt = grp_ptr->el_data + nel ;

/* . . . every pair of nodes for this element . . . */
/* NOTE CHECK FOR ZERO-BASED VS 1-BASED LISTS (Local, global nodes ) */

      for( ref_lnode = 0; ref_lnode < info->nen; ref_lnode++)
      {
         ref_gnode = el_pt->ien[ref_lnode];
         for( aux_lnode = 0; aux_lnode < info->nen; aux_lnode++)
         {
            aux_gnode = el_pt->ien[aux_lnode]; 

/* put this aux node into list for ref node.  Insert in node order.*/

            this_link = ref_gnode-1;
            next_link = llist[2*this_link + 1];

            while(next_link != 0 && llist[2*next_link+0] < aux_gnode)
            {
               this_link = next_link;
               next_link = llist[next_link*2 + 1];
            }

                /* note case llist[..) == aux_gnode 
                    requires no action, since link already exists.
                */

            if(next_link == 0 || llist[next_link*2+0] > aux_gnode)
            {
               if(lfree <= max_nz)
               {
                  llist[lfree*2+0] = aux_gnode;
                  llist[lfree*2 + 1] = next_link;
                  llist[this_link*2 + 1] = lfree++;
               }
               else
               {
                 printf("Bad assumptions on max_nz: rewrite alloc strategy?\n");
                  exit(0);
               } /* end if lfree . . .*/
            }    /* end if next_link . . .*/
         }       /* for aux_lnode . . . */
      }          /* for ref_lnode . . . */
   }             /* for nel. . . */

/* now we have the linked lists: build "icol" adjacency list, "irow" index */
   next_icol = 1;
   num_col = 0;
   for(ref_gnode = 1; ref_gnode <= nnodes; ref_gnode++)
   {
      this_link = llist[(ref_gnode-1)*2+1];
      irow[ref_gnode] = next_icol;
        
      while(this_link != 0)
      {
         icol[next_icol++]= llist[this_link*2+0];
         this_link = llist[2*this_link + 1];
         num_col++; 
      }
   }
   irow[nnodes+1] = next_icol;

   total_nonzeros  = irow[nnodes+1]-1;

/* call rcm optimizer */

   genrcm(nnodes,irow,icol,invper,msk,xls);

   for(ref_gnode = 1; ref_gnode <= nnodes; ref_gnode++)
      global.permut[invper[ref_gnode]] = ref_gnode;
/*
   for(ref_gnode = 1; ref_gnode <= nnodes; ref_gnode++)
      printf("gn = %d, pn = %d\n",ref_gnode,global.permut[ref_gnode]);
*/
/* write to a file to plot nonzero distribution */
/*
   tfp = fopen("nonzeros","w");
   for(ref_gnode = 1; ref_gnode < nnodes; ref_gnode++)
   {
      for(index = irow[ref_gnode]; index <= irow[ref_gnode+1]-1; index++)
      {
          fprintf(tfp,"%d\t%d\t%d\t%d\n",
              ref_gnode,icol[index],
              global.permut[ref_gnode],global.permut[icol[index]]);
      }
   }
   fclose(tfp);
*/
 
/* overwrite id array with permuted information */

   id = global.id_pointer ;

  /* now assign equations ...    (local equations start at zero)  */

     ndof = fe_sys.ndof;

     next_eq = 0 ;

     for( pgnode = 1 ; pgnode <=nnodes ; pgnode++ )
         {
          for(idof=0 ; idof<ndof ; idof++){
              id[ndof*(pgnode-1) + idof] = 
     ( *(global.eqtype + ndof*(invper[pgnode] - 1) + idof) != 0) ? next_eq++ : -1;
 /*             printf("pg = %d, idof = %d, id = %d\n",
                pgnode,idof,id[ndof*(pgnode-1) + idof]);
*/
           }

         }   /* end loop over permuted global node number */
      
     
/* overwrite lm arrays with permuted implied equations */

   for( nel = 0 ;  nel < info->numel ; nel++ )
   {
      el_pt = grp_ptr->el_data + nel ;
      index = 0 ;

      for( ref_lnode = 0 ; ref_lnode < info->nen ; ref_lnode++)
      {
         pgnode = global.permut[ el_pt->ien[ref_lnode] ] ;

         for( idof = 0 ; idof < ndof ; idof++){
            el_pt->lm[index++] = id[ndof*(pgnode-1) + idof];
         }
/*         printf("el = %d, ln = %d, ien = %d, lm0 = %d, lm1 = %d, lm2 = %d\n",
                    nel,ref_lnode,el_pt->ien[ref_lnode],
                                    el_pt->lm[ndof*ref_lnode + 0],
                                    el_pt->lm[ndof*ref_lnode + 1],
                                    el_pt->lm[ndof*ref_lnode + 2]);
*/
      }
   }
/* note that now the id array yields internal equations by permuted node,
   element ien yields the external (nonpermuted) node by local node,
   element lm yields the internal equations.
   If the element-wise list of permuted nodes is desired, must construct
   global.permut[el_pt->ien[lnode]].
*/
}
/************************** end of reorder ***********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   genrcm   ***************
*/

/*  INTERFACE:   */

     void
     genrcm(
            int  neqns ,   /* NUMBER OF EQUATIONS */
            int  *xadj ,  /* ADJACENCY STRUCTURE OF THE GRAPH OF THE MATRIX */
            int  *adjncy ,/* ADJACENCY STRUCTURE OF THE GRAPH OF THE MATRIX */
            int  *perm ,   /* VECTOR THAT CONTAINS THE RCM ORDERING */
            int  *mask ,   /* MARK VARS NUMBERED IN ORDERING PROCESS */
            int  *xls   /* INDEX VECTOR FOR A LEVEL STRUCTURE */
           )

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine genrcm
** genrcm computes the reverse cuthill mckee ordering for a general
** adjacency graph such as a sparse finite element matrix. 
**
** Original FORTRAN comments: (transcribed into C by JWP from
** p_reorder.f of PHOEBUS3_T3D):
*****************************************************************
*****************************************************************
**********   GENRCM ..... GENERAL REVERSE CUTHILL MCKEE   *******
*****************************************************************
*****************************************************************
**
**     PURPOSE - GENRCM FINDS THE REVERSE CUTHILL-MCKEE                     
**        ORDERING FOR A GENERAL GRAPH. FOR EACH CONNECTED                 
**        COMPONENT IN THE GRAPH, GENRCM OBTAINS THE ORDERING             
**        BY CALLING THE SUBROUTINE RCM.                                 
**
**     INPUT PARAMETERS -                                               
**        NEQNS - NUMBER OF EQUATIONS                                  
**        (XADJ, ADJNCY) - ARRAY PAIR CONTAINING THE ADJACENCY        
**               STRUCTURE OF THE GRAPH OF THE MATRIX.               
**
**     OUTPUT PARAMETER -                                         
**        PERM - VECTOR THAT CONTAINS THE RCM ORDERING.          
**
**     WORKING PARAMETERS -                                             
**        MASK - IS USED TO MARK VARIABLES THAT HAVE BEEN              
**               NUMBERED DURING THE ORDERING PROCESS. IT IS          
**               INITIALIZED TO 1, AND SET TO ZERO AS EACH NODE      
**               IS NUMBERED.                                       
**        XLS - THE INDEX VECTOR FOR A LEVEL STRUCTURE.  THE       
**               LEVEL STRUCTURE IS STORED IN THE CURRENTLY       
**               UNUSED SPACES IN THE PERMUTATION VECTOR PERM.   
**
**     PROGRAM SUBROUTINES -                                    
**        FNROOT, RCM.                                         
**
*****************************************************************
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

{
   int ccsize, i, nlvl,
       num = 1,
       root;

   for(i=1; i<=neqns; i++)
      mask[i] = 1;

/* for each masked connected component . . . */

   for(i=1; i<=neqns; i++)
   {
      if (mask[i] != 0) 
      {
         root = i;

/* first find a pseudo-peripheral node root.
   Note that the level structure found by 
   fnroot is stored starting at perm[num].
   Then rcm is called to order the component
   using root as the starting node.
*/

         fnroot( &root,xadj,adjncy,mask,&nlvl,xls,perm,num );

         rcm( root,xadj,adjncy,mask,perm,num,&ccsize,xls );

         num += ccsize;
      
         if (num > neqns) 
            break;
      }
   }
}
/************************** end of genrcm ************************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   rcm   ***************
*/

/*  INTERFACE:   */

        void
        rcm(
            int  root ,   /* NODE THAT DEFINES THE CONNECTED COMPONENT */
            int  *xadj ,  /* ADJACENCY STRUCTURE OF THE GRAPH OF THE MATRIX */
            int  *adjncy ,/* ADJACENCY STRUCTURE OF THE GRAPH OF THE MATRIX */
            int  *mask ,   /* MARK VARS NUMBERED IN ORDERING PROCESS */
            int  *perm ,   /* VECTOR THAT CONTAINS THE RCM ORDERING */
            int  num ,   /*  */
            int  *ccsize ,   /* SIZE OF THE CONNECTED COMPONENT */
            int  *deg   /* TEMPORARY VECTOR USED TO HOLD THE DEGREE */
           )

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** rcm does connectivity analysis (jwp interpretation; see below).
** Original comments from Fortran code:
*****************************************************************
*****************************************************************
**********     RCM ..... REVERSE CUTHILL-MCKEE ORDERING   *******
*****************************************************************
*****************************************************************
**
**     PURPOSE - RCM NUMBERS A CONNECTED COMPONENT SPECIFIED BY             7.
**        MASK AND ROOT, USING THE RCM ALGORITHM.                           8.
**        THE NUMBERING IS TO BE STARTED AT THE NODE ROOT.                  9.
**
**     INPUT PARAMETERS -                                                  11.
**        ROOT - IS THE NODE THAT DEFINES THE CONNECTED                    12.
**               COMPONENT AND IT IS USED AS THE STARTING                  13.
**               NODE FOR THE RCM ORDERING.                                14.
**        (XADJ, ADJNCY) - ADJACENCY STRUCTURE PAIR FOR                    15.
**               THE GRAPH.                                                16.
**
**     UPDATED PARAMETERS -                                                18.
**        MASK - ONLY THOSE NODES WITH NONZERO INPUT MASK                  19.
**               VALUES ARE CONSIDERED BY THE ROUTINE.  THE                20.
**               NODES NUMBERED BY RCM WILL HAVE THEIR                     21.
**               MASK VALUES SET TO ZERO.                                  22.
**
**     OUTPUT PARAMETERS -                                                 24.
**        PERM - WILL CONTAIN THE RCM ORDERING.                            25.
**        CCSIZE - IS THE SIZE OF THE CONNECTED COMPONENT                  26.
**               THAT HAS BEEN NUMBERED BY RCM.                            27.
**
**     WORKING PARAMETER -                                                 29.
**        DEG - IS A TEMPORARY VECTOR USED TO HOLD THE DEGREE              30.
**               OF THE NODES IN THE SECTION GRAPH SPECIFIED               31.
**               BY MASK AND ROOT.                                         32.
**
**     PROGRAM SUBROUTINES -                                               34.
**        DEGREE.                                                          35.
**
*****************************************************************
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

{
   int fnbr,i,j,jstop,jstrt,k,l,lbegin,lnbr,lperm,lvlend,nbr,node;

/* find the degrees of the nodes in the
   component specified by mask and root.
*/

   degree( root,xadj,adjncy,mask,deg,ccsize,perm,num );

/* watch out for pass-by-value:  if anything gets changed here, need
to pass address up, not value .. . */

   mask[root] = 0;
   if (*ccsize <= 1) return ;
   lvlend = 0;
   lnbr = 1;

   do
   {

/* lbegin and lvlend point to the beginning and
   the end of the current level respectively.
*/

      lbegin = lvlend + 1;
      lvlend = lnbr;

/* for each node in the current level. . . 
*/

      for(i=lbegin; i<=lvlend; i++)
      {
         node = perm[i+num-1];
         jstrt = xadj[node];
         jstop = xadj[node+1] - 1;

/* Find the unnumbered neighbors of node.
   fnbr and lnbr point the first and last
   unnumbered neighbors respectively of the current
   node in perm.
*/
         fnbr = lnbr + 1;
         for(j=jstrt; j<=jstop; j++)
         {
            nbr = adjncy[j];
            if( mask[nbr] == 0)
               continue;
            lnbr++;
            mask[nbr] = 0;
            perm[lnbr+num-1] = nbr;
         }
         if(fnbr >= lnbr) 
            continue;

/* Sort the neighbors of node in increasing 
   order by degree.  Linear insertion is used.
*/

         k = fnbr;
         do
         {
            l = k++;
            nbr = perm[k+num-1];
            while( l >= fnbr )  
            {
               lperm = perm[l+num-1];
               if( deg[lperm] <= deg[nbr] )
                  break;
               else
               {
                  perm[l+num] = lperm;
                  l--;
               }
            }
            perm[l+num] = nbr;
         } while(k < lnbr);
      }
   } while(lnbr > lvlend);

/* We now have the Cuthill McKee ordering.
   Reverse it below. . . 
*/

   k = *ccsize/2;
   l = *ccsize;
   for(i=1; i<=k; i++)
   {
      lperm             = perm[l + num - 1];
      perm[l + num - 1] = perm[i + num - 1];
      perm[l + num - 1] = lperm;
      l--;
   }
}
/************************** end of rcm ***************************************/ 
         

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   degree   ***************
*/

/*  INTERFACE:   */

     void
     degree(
            int  root ,   /* NODE THAT DEFINES THE CONNECTED COMPONENT */
            int  *xadj ,  /* ADJACENCY STRUCTURE OF THE GRAPH OF THE MATRIX */
            int  *adjncy ,/* ADJACENCY STRUCTURE OF THE GRAPH OF THE MATRIX */
            int  *mask ,   /* SPECIFIES A SECTION SUBGRAPH */
            int  *deg  , /* ARRAY CONTAINING THE DEGREES OF THE NODES */
            int  *ccsize ,   /* SIZE OF THE COMPONENT SPECIFED */
            int  *ls ,   /* TEMPORARY VECTOR USED TO STORE THE NODES */
            int  num     /*  */
           )

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine degree
** degree computes the graph steps to each node in the subgraph. 
** Original Comments from Fortran code:
*****************************************************************
*****************************************************************
**********     DEGREE ..... DEGREE IN MASKED COMPONENT   ********
*****************************************************************
*****************************************************************
**
**     PURPOSE - THIS ROUTINE COMPUTES THE DEGREES OF THE NODES             7.
**        IN THE CONNECTED COMPONENT SPECIFIED BY MASK AND ROOT.            8.
**        NODES FOR WHICH MASK IS ZERO ARE IGNORED.                         9.
**
**     INPUT PARAMETER -                                                   11.
**        ROOT - IS THE INPUT NODE THAT DEFINES THE COMPONENT.             12.
**        (XADJ, ADJNCY) - ADJACENCY STRUCTURE PAIR.                       13.
**        MASK - SPECIFIES A SECTION SUBGRAPH.                             14.
**
**     OUTPUT PARAMETERS -                                                 16.
**        DEG - ARRAY CONTAINING THE DEGREES OF THE NODES IN               17.
**              THE COMPONENT.                                             18.
**        CCSIZE-SIZE OF THE COMPONENT SPECIFED BY MASK AND ROOT           19.
**
**     WORKING PARAMETER -                                                 21.
**        LS - A TEMPORARY VECTOR USED TO STORE THE NODES OF THE           22.
**               COMPONENT LEVEL BY LEVEL.                                 23.
**
*****************************************************************
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

{
   int i, ideg, j, jstop, jstrt,lbegin,lvlend,lvsize, nbr,node;

/* initialization . . .
   The array xadj is used as a temporary marker to 
   indicate which nodes have been considered so far.
*/

   ls[num] = root;
   xadj[root] *= -1;
   lvlend = 0;
   *ccsize = 1;

   do
   {

/* lbegin is the pointer to the beginning of the current
   level, and lvlend points to the end of this level.
*/

      lbegin = lvlend+1;
      lvlend = *ccsize;

/* Find the degrees of nodes in the current level,
   and at the same time, generate the next level.
*/

      for(i=lbegin; i<=lvlend; i++)
      {
         node = ls[i + num - 1];
         jstrt = -xadj[node];
         jstop = abs(xadj[node+1]) - 1;
         ideg = 0;
         if( jstop < jstrt) {
            deg[node] = ideg;
            continue;
          }
         for(j=jstrt; j<= jstop; j++)
         {
            nbr = adjncy[j];
            if( mask[nbr] == 0 ) continue;
            ideg++;
            if( xadj[nbr] < 0 ) continue;
            xadj[nbr] *= -1;
            (*ccsize)++;
            ls[*ccsize + num - 1] = nbr;
         }
         deg[node] = ideg;
      }

/* Compute the current level width.
   If it is nonzero, generate another level.
*/

      lvsize = *ccsize - lvlend;
   } while(lvsize > 0);

/* Reset xadj to its correct sign and return.
*/

   for(i=1; i<=*ccsize; i++)
   {
      node = ls[i + num - 1];
      xadj[node] *= -1;
   }
}
/************************** end of degree ************************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   fnroot   ***************
*/

/*  INTERFACE:   */

     void
     fnroot(
            int  *root ,   /* DEFINES THE COMP FOR PSEUDO-PERIPH NODE */
            int  *xadj ,  /* ADJACENCY STRUCTURE OF THE GRAPH OF THE MATRIX */
            int  *adjncy ,/* ADJACENCY STRUCTURE OF THE GRAPH OF THE MATRIX */
            int  *mask ,   /* SPECIFIES A SECTION SUBGRAPH */
            int  *nlvl ,  /* NUMBER OF LEVELS IN THE LEVEL STRUCTURE */
            int  *xls ,   /* LEVEL STRUCTURE ARRAY PAIR */
            int  *ls ,   /* LEVEL STRUCTURE ARRAY PAIR */
            int  num     /*  */
           )

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine fnroot
** fnroot finds the pseudo-peripheral node for a given subgraph.
** Original comments from Fortran:
*****************************************************************
*****************************************************************
*********     FNROOT ..... FIND PSEUDO-PERIPHERAL NODE    *******
*****************************************************************
*****************************************************************
**
**    PURPOSE - FNROOT IMPLEMENTS A MODIFIED VERSION OF THE                 7.
**       SCHEME BY GIBBS, POOLE, AND STOCKMEYER TO FIND PSEUDO-             8.
**       PERIPHERAL NODES.  IT DETERMINES SUCH A NODE FOR THE               9.
**       SECTION SUBGRAPH SPECIFIED BY MASK AND ROOT.                      10.
**
**    INPUT PARAMETERS -                                                   12.
**       (XADJ, ADJNCY) - ADJACENCY STRUCTURE PAIR FOR THE GRAPH.          13.
**       MASK - SPECIFIES A SECTION SUBGRAPH. NODES FOR WHICH              14.
**              MASK IS ZERO ARE IGNORED BY FNROOT.                        15.
**
**    UPDATED PARAMETER -                                                  17.
**       ROOT - ON INPUT, IT (ALONG WITH MASK) DEFINES THE                 18.
**              COMPONENT FOR WHICH A PSEUDO-PERIPHERAL NODE IS            19.
**              TO BE FOUND. ON OUTPUT, IT IS THE NODE OBTAINED.           20.
**
**    OUTPUT PARAMETERS -                                                  22.
**       NLVL - IS THE NUMBER OF LEVELS IN THE LEVEL STRUCTURE             23.
**              ROOTED AT THE NODE ROOT.                                   24.
**       (XLS,LS) - THE LEVEL STRUCTURE ARRAY PAIR CONTAINING              25.
**                  THE LEVEL STRUCTURE FOUND.                             26.
**
**    PROGRAM SUBROUTINES -                                                28.
**       ROOTLS.                                                           29.
**
*****************************************************************
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

{
   int ccsize, j, jstrt, k, kstop, kstrt, mindeg, nabor, ndeg, node, nunlvl;

/* determine the level structure rooted at root.
*/

   rootls(*root,xadj,adjncy,mask,nlvl,xls,ls,num);
   ccsize = xls[*nlvl+1] - 1;
   if(*nlvl == 1 || *nlvl == ccsize) return ;

/* Pick a node with minimum degree from the last level.
*/
   do
   {
      jstrt = xls[*nlvl];
      mindeg = ccsize;
      *root = ls[jstrt + num - 1];
      if( ccsize == jstrt )
      {
         rootls(*root,xadj,adjncy,mask,&nunlvl,xls,ls,num);
         if(nunlvl <= *nlvl) return ;
         *nlvl = nunlvl;
         if( *nlvl < ccsize) continue;
         return ;
      }
      for(j=jstrt; j<=ccsize; j++)
      {
         node = ls[j+num-1];
         ndeg = 0;
         kstrt = xadj[node];
         kstop = xadj[node+1] - 1;
         for(k=kstrt; k<= kstop; k++)
         {
             nabor = adjncy[k];
             if(mask[nabor] > 0) ndeg++;
         }
         if(ndeg >= mindeg) continue;
         *root = node;
         mindeg = ndeg;
      }

/* and generate its rooted level structure.
*/

      rootls(*root,xadj,adjncy,mask,&nunlvl,xls,ls,num);
      if(nunlvl <= *nlvl) return ;

      *nlvl = nunlvl;

   } while(*nlvl < ccsize);
}
/************************** end of fnroot ************************************/ 

      
/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   rootls   ***************
*/

/*  INTERFACE:   */

     void
     rootls(
            int  root ,   /* NODE AT WHICH LEVEL STRUCTURE IS ROOTED */
            int  *xadj ,  /* ADJACENCY STRUCTURE OF THE GRAPH OF THE MATRIX */
            int  *adjncy ,/* ADJACENCY STRUCTURE OF THE GRAPH OF THE MATRIX */
            int  *mask ,   /* SPECIFIES A SECTION SUBGRAPH */
            int  *nlvl ,  /* NUMBER OF LEVELS IN THE LEVEL STRUCTURE */
            int  *xls ,   /* LEVEL STRUCTURE ARRAY PAIR */
            int  *ls ,   /* LEVEL STRUCTURE ARRAY PAIR */
            int  num     /*  */
           )

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine rootls
** rootls generates the level structure corresponding to "root".
** Original Fortran comments:
*****************************************************************
*****************************************************************
**********     ROOTLS ..... ROOTED LEVEL STRUCTURE      *********
*****************************************************************
*****************************************************************
**
**     PURPOSE - ROOTLS GENERATES THE LEVEL STRUCTURE ROOTED
**        AT THE INPUT NODE CALLED ROOT. ONLY THOSE NODES FOR
**        WHICH MASK IS NONZERO WILL BE CONSIDERED.
**
**     INPUT PARAMETERS -
**        ROOT - THE NODE AT WHICH THE LEVEL STRUCTURE IS TO
**               BE ROOTED.
**        (XADJ, ADJNCY) - ADJACENCY STRUCTURE PAIR FOR THE
**               GIVEN GRAPH.
**        MASK - IS USED TO SPECIFY A SECTION SUBGRAPH. NODES
**               WITH MASK(I)=0 ARE IGNORED.
**
**     OUTPUT PARAMETERS -
**        NLVL - IS THE NUMBER OF LEVELS IN THE LEVEL STRUCTURE.
**        (XLS, LS) - ARRAY PAIR FOR THE ROOTED LEVEL STRUCTURE.
**
*****************************************************************
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

{
   int i,j,jstop,jstrt,lbegin,ccsize,lvlend,lvsize,nbr,node;

/* initialization . . .*/

   mask[root] = 0;
   ls[num] = root;
   *nlvl = 0;
   lvlend = 0;
   ccsize = 1;

   do
   {

/* lbegin is the pointer to the beginning of the current
   level, and lvlend points to the end of this level.
*/

      lbegin = lvlend + 1;
      lvlend = ccsize;
      (*nlvl)++;
      xls[*nlvl] = lbegin;

/* generate the next level by finding all the masked
   neighbors of nodes in the current level.
*/
      for(i=lbegin; i<= lvlend; i++)
      {
         node = ls[i+num-1];
         jstrt = xadj[node];
         jstop = xadj[node+1] - 1;
         if(jstop < jstrt) continue;
         for(j=jstrt; j<=jstop; j++)
         {
            nbr = adjncy[j];
            if(mask[nbr] == 0) continue;
            ccsize++;
            ls[ccsize+num-1] = nbr;
            mask[nbr] = 0;
         }
      }

/* Compute the current level width.
   If it is nonzero, generate the next level.
*/

      lvsize = ccsize - lvlend;
   } while (lvsize > 0);

   xls[*nlvl+1] = lvlend + 1;

/* Reset mask to one for the nodes in the level structure.
*/
   for(i=1; i<=ccsize; i++)
   {
      node = ls[i+num-1];
      mask[node] = 1;
   }
}
/************************** end of rootls ************************************/ 

      
/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   colht   ***************
*/

/*  INTERFACE:   */

      void
      colht(
            GROUP    *grp_ptr  ,  /* pointer to current element group */
            PROFILE  *a    /* ptr to assembled profile stiffness storage */
           )

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine colht
** colht computes column heights of global array.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
      ELEMENT_INFO  *info ;
      ELEMENT_DATA  *el_pt ;
      int    nel , min , i , num , m ;
 
/*    FILE *tfp;  */

      info = grp_ptr->el_info ;

      for( nel = 0 ;  nel < info->numel ; nel++ )   
          {
           min = 100000000 ;
           el_pt = grp_ptr->el_data + nel ;
           for ( i = 0 ; i < info->nee ; i++)
               {
                if( (num = (int) el_pt->lm[i] ) != -1 )
                   min = MIN(num,min) ;
               }

           for ( i = 0 ; i < info->nee ; i++ )
               {
                if( ( num = (int) el_pt->lm[i] ) != -1 )
                  if( ( m = num - min ) > (int) *(a->diag + num) )
                           *(a->diag + num) = (int) m ;
               }
            }

/* report colhts to file for debugging */
/*
        tfp = fopen("colhts","w");
        for(i=0;i<fe_sys.neq;i++)
          fprintf(tfp,"%d\t%d\n", i, a->diag[i]);
        fclose(tfp);
*/
      }

/************************** end of colht *************************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   profile_diag   ***************
*/

/*  INTERFACE:   */

 void
 profile_diag(
              PROFILE  *a  ,  /* ptr to assembled profile stiffness storage */
              int   profile_case    /* old flag for substructuring... */
             )

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine profile_diag
** profile diag computes the diagonal addresses.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

    {
      int i, neq ;

      *(a->diag) = 0 ;
      neq       = a->neq  ;
      for( i=1 ; i < neq ; i++ ) 
          {
           if( profile_case == INTERIOR )
               *(a->diag + i) += *(a->diag + i - 1) + 1 ;
           else
               *(a->diag + i) = *(a->diag + i - 1) + (int ) i + 1 ;
          }
      a->size = (int) *(a->diag + neq - 1) + 1 ;
    }

/************************** end of profile_diag ******************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   factor   ***************
*/

/*  INTERFACE:   */
       void
       factor(
              real  *stiff ,         /* assembled profile matrix */
              int   *diag ,           /* array of diagonal addresses */
              int   number_of_eqs     /* number of equations */
             )

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine factor
** factor performs profile-based factorization:
**                          t                       
**                     a=(u) * d * u (crout) 
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
      int j_row , j , j_diag ,j_ht ,i_start , i_end ,
          k , i_diag  , i , i_ht , i_row ;
      real d ;

      j_row = (-1) ;
      for (j = 0 ; j < number_of_eqs; j++ )
          {
           j_diag = (int) *(diag + j) ;
           j_ht   = j_diag - j_row ;
           i_start = j - j_ht + 2  ;
           if( j_ht - 2 > 0 )
               {
                i_end = j - 1 ;
                k = j_row + 2 ;
                i_diag = (int) *(diag + i_start - 1) ;
                for ( i = i_start ; i <= i_end ; i++ )
                    {
                     i_row = i_diag ;
                     i_diag = (int) *(diag + i) ;
                     i_ht = MIN(i_diag - i_row - 1,i - i_start + 1) ;
                     if( i_ht > 0) 
		     {
                       *(stiff + k) -= dot_real((stiff +  k - i_ht),
                                       (stiff + i_diag - i_ht),
                                        i_ht) ;
		     }
                     k++ ;
                    }
                }
           if( j_ht - 1 > 0 )
              {
               i_row = j_row + 1 ;
               i_end = j_diag - 1 ;
               k = j - j_diag ;
               for ( i = i_row ; i <= i_end ; i++)
                   {
                    i_diag = (int) *(diag + i + k)   ;
                    d =  *(stiff + i) ;
                    *(stiff + i) /= *(stiff + i_diag) ;
                    *(stiff + j_diag) -= ( d * *(stiff + i) ) ;          
                   }
              }

           j_row = j_diag ;
          }
     }

/************************** end of factor ************************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   full_back   ***************
*/

/*  INTERFACE:   */

    void
    full_back(
              real  *stiff ,         /* factored profile matrix */
              real  *rhs ,         /* assembled rhs vector */
              int   *diag ,           /* array of diagonal addresses */
              int   number_of_eqs     /* number of equations */
             )

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine full_back
** full_back performs all three steps of backsubstitution.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
      real d ;
      int j_row , j , j_diag , i_start , 
          j_ht  , i , k ;

 /*  forward reduction . . .  */
  
      j_row = (-1) ;
      for ( j = 0 ; j < number_of_eqs; j++ )
          {
           j_diag = (int) *(diag + j) ;
           j_ht   =  j_diag - j_row ;
           i_start = j - j_ht + 2  ;
           if( j_ht - 1 > 0 )
	   {
             *(rhs + j) 
                -= dot_real((stiff + j_row + 1) 
                ,(rhs + i_start - 1) , ( j_ht - 1 ) ) ;
	   }
           j_row = j_diag ;
          }

  /*  diagonal scaling . . .  */
  
      for( j = 0 ; j < number_of_eqs ; j++)
         {
           j_diag = (int) *(diag + j) ;
           *(rhs + j ) /= *( stiff + j_diag) ;
         }

  /*  backsubstitution . . .  */
  
      j = number_of_eqs - 1 ;
      j_diag = (int) *(diag + j) ;
      FOREVER     
            {
             d = *(rhs + j) ;
             j-- ;
             if( j < 0 ) break ;
             j_row = (int) *(diag + j) ;
             if ( j_diag - j_row  >  1 )
                {
                 i_start = j - j_diag + j_row + 2 ;
                 k = j_row - i_start + 1 ;
                 for( i = i_start ; i <= j ; i++ )
                    *(rhs + i) -= ( *(stiff + i + k) * d ) ;
                 }
             j_diag = j_row ;
            }
     } 
/************************** end of full_back *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   pcg_loop   ***************
*/

/*  INTERFACE:   */

     void
     pcg_loop( pamr_Mesh mesh[] )

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine pcg_loop
** pcg_loop performs the preconditioned conjugate gradient iteration
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

  {
   real  *M , *r , *d , *t , *x , *guess ;
   real  alpha , beta , rMr ;
   int   neq , k ;
   
   neq = loc_sys.neq ;
   /* set pointers */
   guess = force.last_result ;
   t = pcg.temp ;
   M = pcg.precond ;
   r = pcg.r ;
   d = pcg.d ;
   x = force.full_rhs ;
   
   /* intialize loop */
   k = -1 ;
   move_real(guess,d,neq) ; /* temporarily use "d" to hold x(0) */
   clear_real(t,neq) ;
   elgrp_loop( mesh, ESPROD ) ;   /* A x(0)  (result in "t") */
   globalize( mesh, t, 0 );    /* update matrix-vector product result */
   vadd(x,-1.0,t,r,neq) ;   /* r = b - Ax */
   vouter(M,r,d,neq) ;      /* d = M r */
   move_real(guess,x,neq) ; /* overwrite "rhs" array with current soln */
   
   /* enter CG loop */
   
   do
      {
       k++ ;
       vouter(M,r,t,neq) ;         /* M r */
       /* rMr = dot_real(r,t,neq) ; */   /* r M r */
       rMr = dot_par(mesh,r,t,loc_sys.neq_owned) ;   /* r M r */
       if(k==0) pcg.norm = rMr ;   /* save the starting norm */
       if(k==0 && time_data.time==ZERO) pcg.first_norm = rMr ;   /* save the starting norm */
       clear_real(t,neq) ;
       elgrp_loop( mesh, ESPROD ) ;      /* A d  (result in "t") */
       globalize( mesh, t, 0 );       /* update matrix-vector product result */
       /*       alpha = rMr / dot_real(d,t,neq) ; */ /* alpha = (r M r)/(d A d) */
       alpha = rMr / dot_par(mesh,d,t,loc_sys.neq_owned) ;  /* alpha = (r M r)/(d A d) */
       vadd(x,alpha,d,x,neq) ;    /* x <- x + alpha d */
       vadd(r,-alpha,t,r,neq) ;   /* r <- r - alpha A d */
       vouter(M,r,t,neq) ;        /* M r */
       /*       beta = dot_real(r,t,neq) / rMr ; */  /* beta = (r M r)new / (r M r)old */
       beta = dot_par(mesh,r,t,loc_sys.neq_owned) / rMr ;   /* beta = (r M r)new / (r M r)old */
       vadd(t,beta,d,d,neq) ;     /* d <- M r + beta d */
       *(pcg.hist+(k%CGMAX)) = rMr * beta ;
      } while(converged(k) != 1) ;
      
   /* iterations loop finished... 
      now preserve the result and return back to solver() */

 /* only save the result for next step if norm reduction is not excessively small */
  if(*(pcg.hist+(k%CGMAX))/pcg.norm > 1.0e-12 || time_data.time==ZERO)     
   move_real(x,guess,neq) ; /* save result as starting guess for next time */
   
 /* save the first result as a default guess to use as a last resort */
   if(time_data.time==ZERO)
      move_real(x,force.first_result,neq) ;
      
 /* Last resort - if starting norm is getting too tiny, use the default guess */
   if(pcg.norm/pcg.first_norm < 1.0e-12)
      move_real(force.first_result,guess,neq) ;
      
  }
  
/************************** end of pcg_loop **********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   converged   ***************
*/

/*  INTERFACE:   */

  int converged(
                int   k   /* iteration number */
               )

/*  RETURN VALUE:  flag indicating successful convergence (or not) */
/*  DESCRIPTION:   */
/*
** function converged()
** monitors convergence of CG solver from magnitude and history of 
** the residual norm
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

  {
   real  ratio , mean1 , mean2 , meanall , stddev ;
   int  i ;
  
   if( (ratio = *(pcg.hist+(k%CGMAX))/pcg.norm) <= CGTOL )
      {
       if(pamr_my_pid() == 0)
	{
	 fprintf(cgconv_file,"time=%g converged in %d iterations norm reduced %g\n",
	 time_data.time,k,ratio) ;
         fprintf(cgconv_file,"starting=%g , ending=%g\n",
                 pcg.norm,*(pcg.hist+(k%CGMAX)) ) ;
	 fflush(cgconv_file) ;
	}
       return(1) ;  /* converged within tolerance */
      }
   else if( k < CGMAX || k%CGMAX != 0 )
      {
       return(0) ;  /* not converged, but insufficient history to declare it stalled */
      }
   else
      {
       mean1 = ZERO ;
       mean2 = ZERO ;
       stddev = ZERO ;
       
       for(i=0;i<(CGMAX/2);i++)
          {
           mean1 += *(pcg.hist+i)/(((real)(CGMAX/2))*pcg.norm) ;
           mean2 += *(pcg.hist+i+(CGMAX/2))/(((real)(CGMAX/2))*pcg.norm) ;
          }
       meanall = (mean1+mean2)/2.0 ;
       for(i=0;i<CGMAX;i++)
          {
           stddev +=
            ((*(pcg.hist+i)/pcg.norm) - meanall)*((*(pcg.hist+i)/pcg.norm) - meanall)/((real)CGMAX) ;
          }
       stddev = sqrt(stddev) ;
       if( mean1 - mean2 < (stddev/((real)(CGMAX/2))) )   /* progress in CGMAX iterations is less than approx sdom */
          {
	   if(pamr_my_pid() == 0)
	     {
	      printf("*****WARNING***** conjugate gradient STALLED at t=%g\n",time_data.time) ;
	      fprintf(cgconv_file,"time=%g stalled in %d iterations norm reduced %g\n",
	      time_data.time,k,ratio) ;
         fprintf(cgconv_file,"starting=%g , ending=%g\n",
                 pcg.norm,*(pcg.hist+(k%CGMAX)) ) ;
	      fflush(cgconv_file) ;
	     }
           return(1) ;  /* declare it stalled and bail out */
          }
       else
          {
           return(0) ;  /* still making progress; keep going */
          }
      }
  }
  
/************************** end of converged *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   put_soln   ***************
*/

/*  INTERFACE:   */

       void
       put_soln(
                RHS_DATA  *solve , /* structure containing rhs arrays */
                int       *id      /* id array of eq numbers */
               )

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine put_soln
** put_soln transfers solution to nodal storage.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
      int      eq_number ;
      COUNTER  node , j , ndof , pgnode;

      ndof = fe_sys.ndof ;

/* march thru the external (unpermuted) nodes */
      for(node = 0 ; node < loc_sys.numnp ; node++)
	/*      for(node = 0 ; node < fe_sys.numnp ; node++) */
            {
             pgnode = global.permut[node+1];

             for( j = 0 ; j < ndof ; j++ )
                     {
                      eq_number = id[( pgnode - 1)*ndof + j] ;
/*
                      printf("nd = %d, pg = %d, idof = %d, eq = %d\n",
                              node,pgnode,j,eq_number);
*/

                      if(  eq_number != (-1) )
                            {
                             *(global.del_displ + node*ndof + j) =
                                  *(solve->full_rhs + eq_number)  ;
                            }

                      else if( time_data.elastic )
                            {
                             *(global.del_displ + node*ndof + j) =
                                  *(global.forv + node*ndof + j)  ;
                            }

                      else if( time_data.fail != 1 )
                            {
                             *(global.del_displ + node*ndof + j) =
                                  *(global.forv + node*ndof + j) *
                                  time_data.dt  ;
                            }
                     }
            }
     }
/************************** end of put_soln **********************************/ 

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   globalize   ***************
*/

/*  INTERFACE:   */

       void
       globalize(
	        const pamr_Mesh m[],     /* the input mesh */
                real  *in_vect,          /* item to be globalized */
                int reset                /* do buffer reset only */
               )

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine globalize
** globalize sums the partial contributions from data on partition boundaries
** by pair-wise communication
**
** Makes bndy buffers static to prevent repeated allocations
**
** Reset mode: if "reset" > 0, reset first_time and free buffers
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
     {
       static int send_nodes_size;
       static int recv_nodes_size;
       static int nproc;
       static int neibpe_cnt;        /* Num. of partition neighbors to this PE     */
       static int node_neibpe_cnt;   /* Total number of nodes sent to neighbor PEs */
       static int *bndy_nodes_ids = NULL;
       static int *neibpe_ids = NULL;       /* Processor IDs of partition neighbor PEs  */
       static int *neibpe_xportlst = NULL;  /* Num. of nodes sent to each PE (may be 0) */
       static int *neibpe_data = NULL;      /* Global IDs of nodes sent to neighbor PEs */

       static int *node_numbers = NULL;     /* List of local indices for global node IDs */
       static int *node_numbers_r = NULL;   /* List of local indices for global node IDs */

       static pamr_bndy_buf *send_buf = NULL;
       static pamr_bndy_buf *recv_buf = NULL;
       int i, j, ieq;
       int* node_number;
       
       
       if(pamr_num_procs() == 1)  return ;   /* do nothing if sequential */

       if (reset > 0)
	 {
	   /* By making first_time TRUE, we ensure next call to globalize
	      will make fresh allocation of buffers to the current size.
	      Free the old buffers so we don't leak memory. 
	   */

	   first_time = TRUE;

	   free(bndy_nodes_ids);
	   free(node_numbers);
	   free(node_numbers_r);
	   free(neibpe_ids);
	   free(neibpe_xportlst);
	   free(neibpe_data);
 	   for (i = 0; i < send_nodes_size; i++)
  	        pamr_bndy_buf_destroy(&send_buf[i]);
 	   for (i = 0; i < recv_nodes_size; i++)
  	        pamr_bndy_buf_destroy(&recv_buf[i]);
	   free(send_buf);
	   free(recv_buf);

	   pamr_move_neibpe_data_r( NULL, 0,
				  NULL, 0,
				  0,
				  NULL, 0,
				  NULL, 0,
				  0,
                                  NULL, 0, 1 );

           return;
	 }

       if(first_time == TRUE)
         {
          send_nodes_size = pamr_node_partition_count( (pamr_Mesh *)m );
          recv_nodes_size = pamr_get_node_neibpe_count( (pamr_Mesh *)m );
          nproc = pamr_num_procs();
          neibpe_cnt = pamr_get_neibpe_count( (pamr_Mesh *)m );
          node_neibpe_cnt = pamr_get_node_neibpe_count( (pamr_Mesh *)m );
          squawk("Initializing globalize comm buffers:\n") ;
          completion() ;
	  /*
          printf("IAM: %d  sent nodes=%d   recv nodes=%d\n"
             ,pamr_my_pid(),send_nodes_size,recv_nodes_size);
	  */
          fflush(stdout);

          first_time = FALSE ;
          
          bndy_nodes_ids = (int *) calloc(send_nodes_size, sizeof(int));
          node_numbers   = (int *) calloc(send_nodes_size, sizeof(int));
          node_numbers_r = (int *) calloc(recv_nodes_size, sizeof(int));
          neibpe_ids = (int *) calloc(neibpe_cnt, sizeof(int));
          neibpe_xportlst = (int *) calloc(nproc, sizeof(int));
          neibpe_data = (int *) calloc(node_neibpe_cnt, sizeof(int));

          send_buf = (pamr_bndy_buf *) calloc(send_nodes_size,
				   sizeof(pamr_bndy_buf));
          recv_buf = (pamr_bndy_buf *) calloc(recv_nodes_size,
				   sizeof(pamr_bndy_buf));

       /* List of nodes on partition boundary */
          pamr_get_partition_nodes( (pamr_Mesh *)m, bndy_nodes_ids );
       /* List of neighbors PE talks to by processor IDs */
          pamr_get_neibpe_data( (pamr_Mesh *)m, neibpe_ids );
       /* List of nodes on partition sent from each PE - maybe 0 */
          pamr_get_node_neibpe_xportlst( (pamr_Mesh *)m, neibpe_xportlst );
       /* List of nodes sent from each PE to neighbors */
          pamr_get_node_neibpe_data( (pamr_Mesh *)m, neibpe_data );

       /* List of boundary node ids using node index, not global ID  */
          for (i = 0; i < send_nodes_size; i++) {
               node_numbers[i] = pamr_my_node_indx( (pamr_Mesh *)m, bndy_nodes_ids[i] );
          }

	  for (i = 0; i < recv_nodes_size; i++) {
	    pamr_bndy_buf_init(&recv_buf[i], 0, 0, fe_sys.ndof);
	    }

	  for (i = 0; i < send_nodes_size; i++) {
	    pamr_bndy_buf_init(&send_buf[i], bndy_nodes_ids[i],
		pamr_my_pid(), fe_sys.ndof);
	    }
	}

       /* Assign in_vect data to send buffer */
       for (i = 0; i < send_nodes_size; i++) {
	    double *data;
            data = (double *)pamr_bndy_buf_data(&send_buf[i]);
            for (j = 0; j < fe_sys.ndof; j++ ) {
                 ieq = *(global.id_pointer + fe_sys.ndof*(node_numbers[i]-1) + j);
                 if ( ieq != -1 )
	              data[j] = in_vect[ieq];
                 else
		      data[j] = ZERO;
	   }
       }

       /* Move data among processors */
       pamr_move_neibpe_data_r( send_buf, send_nodes_size,
                              recv_buf, recv_nodes_size,
                              neibpe_cnt,
                              neibpe_ids, neibpe_cnt,
                              neibpe_xportlst, nproc,
                              node_neibpe_cnt,
                              neibpe_data, node_neibpe_cnt, 0 );

       /* Since the actual boundary node IDs (bndy_nodes_ids[]) used in sends and
          receives are the same (note more are received in general due to shared nodes
          on the boundary) AND since we already have a sorted list of these nodes and
          their local indices (node_numbers[]) find the node local index by binary
          search on the bndy_nodes_ids[] array. The index of that array must then be
          used in the node_numbers[] array to get the node local index */

       /* List of boundary node indices based on incoming global IDs  */
          for (i = 0; i < recv_nodes_size; i++) {
   	       node_number = bsearch( &(recv_buf[i].global_id), bndy_nodes_ids, send_nodes_size,
                                      sizeof(int), (void *)compare_integers );
	       node_numbers_r[i] = node_numbers[ (int) (node_number - bndy_nodes_ids) ];
          }

       /* Update in_vect data from recv buffer */
       for (i = 0; i < recv_nodes_size; i++) {
	    double *data;
            data = (double *)pamr_bndy_buf_data(&recv_buf[i]);
            for (j = 0; j < fe_sys.ndof; j++ ) {
                 ieq = *(global.id_pointer + fe_sys.ndof*(node_numbers_r[i]-1) + j);
                 if ( ieq != -1 )
		      in_vect[ieq] += data[j];
	    }
       }

     }

/************************** end of globalize *********************************/ 

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   dot_par   ***************
*/

/*  INTERFACE:   */

  real dot_par( pamr_Mesh mesh[], /* the mesh */
                real   *vect_1 , /* multiplicand array */
                real   *vect_2 , /* multiplicand array */
                int    n       /* number of entries */
               )

/*  RETURN VALUE:  value of the scalar product */
/*  DESCRIPTION:   */
/*
** Function dot_par
** dot_par forms dot product of two vectors. 
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

      {
       int   i, j   ;
       real  loc_sum, sum    ;

       loc_sum = ZERO ;
       for(j=0 ; j<n ; j++)
       {
		 i = global.owned_eqs[j];
		 loc_sum += (*(vect_1+i) * *(vect_2+i)) ;
       }

       if(pamr_num_procs() != 1)
         MPI_Allreduce((void *)&loc_sum, (void *)&sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
       else
         sum = loc_sum ;

       return(sum) ;
      }
/************************** end of dot_par **********************************/ 
