/*
***                         File solver.h                         ***
***                        GeoFEST version 4.5
*** Copyright (c) 2003, California Institute of Technology        ***
*** U.S.Sponsorship under NASA Contract NAS7-1260 is acknowledged ***
***
*** This software is designated for public release under JPL Task ***
*** Plan Number 83-6791 and may be publicly released after        ***
*** approval for release has been granted by GSFC Software        ***
*** Release  Authority
***
*** This file contains modules for the sparse matrix storage,
*** substructure reduction or other solvers of the matrix system:
***
***          == solver ==
***          == addstiff ==
***          == addfor ==
***          == estiffprod ==
***          == reorder ==
***          == genrcm ==
***          == rcm ==
***          == degree ==
***          == fnroot ==
***          == rootls ==
***          == colht ==
***          == profile_diag ==
***          == factor ==
***          == full_back ==
***          == pcg_loop ==
***          == converged ==
***          == put_soln ==
***/

/*
***GeoFEST version 4.5***
G. A. Lyzenga -- unified parallel/serial version and bug fixes 11/17/03
G. A. Lyzenga -- split node enhancements 8/16/02
G. A. Lyzenga -- modification to correctly relax hydrostatic stresses 6/15/00
J. W. Parker -- 3D support and deleted hypercube branches 9/16/98
G. A. Lyzenga  -- last modified on 3/03/95
Modification history:
2.1: fixes bug in boundary velocities and changes
     the way stresses are output.
2.2: implements split node faults
2.3: adds pass-thru comments
3.0: 3-D tetrahedra supported, hypercube branches deleted
3.1: hydrostatic relaxation changes
4.0: name change from visco to GeoFEST; correct implementation of 2-D out-of-
     plane stresses as well as hydrostatic relaxation; culling of remaining
     vestiges of hypercube stuff; one code without recompilation for both
     2-D and 3-D domains
4.1: Introduction of trilinear hexahedral bricks and PCG iterative solver option
4.2: Bug fixes; standard source prologues added (June 02)
4.3: Corrects inefficiency in split node algorithm and introduces new
     split node input format
4.4: Adds dummy calls to MPI and Pyramid to provide a single unified
     source for parallel and serial codes.  Also introduces a major
     algorithmic enhancement; stress equilibrium maintenance during visco
     elastic stepping for long-term stability.  Also more minor bug fixes.
4.5: Supports surface tractions and roller/spring duo options for tet elements.
     Provides enhanced conjugate gradient convergence monitoring.
 $Log: solver.h,v $
 Revision 1.1  2003/02/09 06:10:56  lyzenga
 *** empty log message ***

*/

#include <stdio.h>
#include <stdlib.h>
#include "mydefs.h"
#include "errcodes.h"
#include "finel.h"
#include "main.h"
#include <math.h>
#include <mpi.h>

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   solver   ***************
*/

/*  INTERFACE:   */

void
solver( pamr_Mesh m[], /* the mesh */
  int	code	/* which solver option to use				*/
  );

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine solver
** solver performs single right-hand-side matrix solution,
** currently using factorization and back substitution on
** single processor OR using PCG iteration.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of solver ************************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   addstiff   ***************
*/

/*  INTERFACE:   */

void
addstiff(
  PROFILE*	a,	/* ptr to assembled profile stiffness storage	*/
  real*		es,	/* ptr to element stiffness storage		*/
  ELEMENT_INFO*	info,	/* ptr to element info struct			*/
  ELEMENT_DATA*	el_pt	/* ptr to current element struct		*/
  );

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine addstiff
** addstiff adds element stiffness to global profile array,
** or to element storage if using PCG solver
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of addstiff **********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   addfor   ***************
*/

/*  INTERFACE:   */

void
addfor (
  real*		dest,	/* destination assembled rhs vector		*/
  real*		rhs,	/* source element rhs vector			*/
  ELEMENT_INFO*	info,	/* ptr to element info struct			*/
  ELEMENT_DATA*	el_pt	/* ptr to current element struct		*/
  );

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine addfor
** addfor assembles the global r.h.s. vector.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of addfor ************************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   estiffprod   ***************
*/

/*  INTERFACE:   */

void
estiffprod(
  GROUP*	grp_ptr	/* pointer to current element group		*/
  );

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine estiffprod
** calculates the product of a given vector with stiffness in element storage
**    t = A * d
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of estiffprod ********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   reorder   ***************
*/

/*  INTERFACE:   */

void
reorder(
  GROUP*	grp_ptr	/* pointer to current element group		*/
  );

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine reorder
** reorder uses ien information to build adjacency information and call
** permutation optimizing routines for minimizing matrix profile.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of reorder ***********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   genrcm   ***************
*/

/*  INTERFACE:   */

void
genrcm(
  int	neqns,	/* NUMBER OF EQUATIONS					*/
  int*	xadj,	/* ADJACENCY STRUCTURE OF THE GRAPH OF THE MATRIX	*/
  int*	adjncy,	/* ADJACENCY STRUCTURE OF THE GRAPH OF THE MATRIX	*/
  int*	perm,	/* VECTOR THAT CONTAINS THE RCM ORDERING		*/
  int*	mask,	/* MARK VARS NUMBERED IN ORDERING PROCESS		*/
  int*	xls	/* INDEX VECTOR FOR A LEVEL STRUCTURE			*/
  );

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine genrcm
** genrcm computes the reverse cuthill mckee ordering for a general
** adjacency graph such as a sparse finite element matrix.
**
** Original FORTRAN comments: (transcribed into C by JWP from
** p_reorder.f of PHOEBUS3_T3D):
*****************************************************************
*****************************************************************
**********   GENRCM ..... GENERAL REVERSE CUTHILL MCKEE   *******
*****************************************************************
*****************************************************************
**
**     PURPOSE - GENRCM FINDS THE REVERSE CUTHILL-MCKEE
**        ORDERING FOR A GENERAL GRAPH. FOR EACH CONNECTED
**        COMPONENT IN THE GRAPH, GENRCM OBTAINS THE ORDERING
**        BY CALLING THE SUBROUTINE RCM.
**
**     INPUT PARAMETERS -
**        NEQNS - NUMBER OF EQUATIONS
**        (XADJ, ADJNCY) - ARRAY PAIR CONTAINING THE ADJACENCY
**               STRUCTURE OF THE GRAPH OF THE MATRIX.
**
**     OUTPUT PARAMETER -
**        PERM - VECTOR THAT CONTAINS THE RCM ORDERING.
**
**     WORKING PARAMETERS -
**        MASK - IS USED TO MARK VARIABLES THAT HAVE BEEN
**               NUMBERED DURING THE ORDERING PROCESS. IT IS
**               INITIALIZED TO 1, AND SET TO ZERO AS EACH NODE
**               IS NUMBERED.
**        XLS - THE INDEX VECTOR FOR A LEVEL STRUCTURE.  THE
**               LEVEL STRUCTURE IS STORED IN THE CURRENTLY
**               UNUSED SPACES IN THE PERMUTATION VECTOR PERM.
**
**     PROGRAM SUBROUTINES -
**        FNROOT, RCM.
**
*****************************************************************
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of genrcm ************************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   rcm   ***************
*/

/*  INTERFACE:   */

void
rcm(
  int	root,	/* NODE THAT DEFINES THE CONNECTED COMPONENT		*/
  int*	xadj,	/* ADJACENCY STRUCTURE OF THE GRAPH OF THE MATRIX	*/
  int*	adjncy,	/* ADJACENCY STRUCTURE OF THE GRAPH OF THE MATRIX	*/
  int*	mask,	/* MARK VARS NUMBERED IN ORDERING PROCESS		*/
  int*	perm,	/* VECTOR THAT CONTAINS THE RCM ORDERING		*/
  int	num,	/* 							*/
  int*	ccsize,	/* SIZE OF THE CONNECTED COMPONENT			*/
  int*	deg	/* TEMPORARY VECTOR USED TO HOLD THE DEGREE		*/
  );

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** rcm does connectivity analysis (jwp interpretation; see below).
** Original comments from Fortran code:
*****************************************************************
*****************************************************************
**********     RCM ..... REVERSE CUTHILL-MCKEE ORDERING   *******
*****************************************************************
*****************************************************************
**
**     PURPOSE - RCM NUMBERS A CONNECTED COMPONENT SPECIFIED BY             7.
**        MASK AND ROOT, USING THE RCM ALGORITHM.                           8.
**        THE NUMBERING IS TO BE STARTED AT THE NODE ROOT.                  9.
**
**     INPUT PARAMETERS -                                                  11.
**        ROOT - IS THE NODE THAT DEFINES THE CONNECTED                    12.
**               COMPONENT AND IT IS USED AS THE STARTING                  13.
**               NODE FOR THE RCM ORDERING.                                14.
**        (XADJ, ADJNCY) - ADJACENCY STRUCTURE PAIR FOR                    15.
**               THE GRAPH.                                                16.
**
**     UPDATED PARAMETERS -                                                18.
**        MASK - ONLY THOSE NODES WITH NONZERO INPUT MASK                  19.
**               VALUES ARE CONSIDERED BY THE ROUTINE.  THE                20.
**               NODES NUMBERED BY RCM WILL HAVE THEIR                     21.
**               MASK VALUES SET TO ZERO.                                  22.
**
**     OUTPUT PARAMETERS -                                                 24.
**        PERM - WILL CONTAIN THE RCM ORDERING.                            25.
**        CCSIZE - IS THE SIZE OF THE CONNECTED COMPONENT                  26.
**               THAT HAS BEEN NUMBERED BY RCM.                            27.
**
**     WORKING PARAMETER -                                                 29.
**        DEG - IS A TEMPORARY VECTOR USED TO HOLD THE DEGREE              30.
**               OF THE NODES IN THE SECTION GRAPH SPECIFIED               31.
**               BY MASK AND ROOT.                                         32.
**
**     PROGRAM SUBROUTINES -                                               34.
**        DEGREE.                                                          35.
**
*****************************************************************
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of rcm ***************************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   degree   ***************
*/

/*  INTERFACE:   */

void
degree(
  int	root,	/* NODE THAT DEFINES THE CONNECTED COMPONENT		*/
  int*	xadj,	/* ADJACENCY STRUCTURE OF THE GRAPH OF THE MATRIX	*/
  int*	adjncy,	/* ADJACENCY STRUCTURE OF THE GRAPH OF THE MATRIX	*/
  int*	mask,	/* SPECIFIES A SECTION SUBGRAPH				*/
  int*	deg,	/* ARRAY CONTAINING THE DEGREES OF THE NODES		*/
  int*	ccsize,	/* SIZE OF THE COMPONENT SPECIFED			*/
  int*	ls,	/* TEMPORARY VECTOR USED TO STORE THE NODES		*/
  int	num	/* 							*/
  );

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine degree
** degree computes the graph steps to each node in the subgraph.
** Original Comments from Fortran code:
*****************************************************************
*****************************************************************
**********     DEGREE ..... DEGREE IN MASKED COMPONENT   ********
*****************************************************************
*****************************************************************
**
**     PURPOSE - THIS ROUTINE COMPUTES THE DEGREES OF THE NODES             7.
**        IN THE CONNECTED COMPONENT SPECIFIED BY MASK AND ROOT.            8.
**        NODES FOR WHICH MASK IS ZERO ARE IGNORED.                         9.
**
**     INPUT PARAMETER -                                                   11.
**        ROOT - IS THE INPUT NODE THAT DEFINES THE COMPONENT.             12.
**        (XADJ, ADJNCY) - ADJACENCY STRUCTURE PAIR.                       13.
**        MASK - SPECIFIES A SECTION SUBGRAPH.                             14.
**
**     OUTPUT PARAMETERS -                                                 16.
**        DEG - ARRAY CONTAINING THE DEGREES OF THE NODES IN               17.
**              THE COMPONENT.                                             18.
**        CCSIZE-SIZE OF THE COMPONENT SPECIFED BY MASK AND ROOT           19.
**
**     WORKING PARAMETER -                                                 21.
**        LS - A TEMPORARY VECTOR USED TO STORE THE NODES OF THE           22.
**               COMPONENT LEVEL BY LEVEL.                                 23.
**
*****************************************************************
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of degree ************************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   fnroot   ***************
*/

/*  INTERFACE:   */

void
fnroot(
  int*	root,	/* DEFINES THE COMP FOR PSEUDO-PERIPH NODE		*/
  int*	xadj,	/* ADJACENCY STRUCTURE OF THE GRAPH OF THE MATRIX	*/
  int*	adjncy,	/* ADJACENCY STRUCTURE OF THE GRAPH OF THE MATRIX	*/
  int*	mask,	/* SPECIFIES A SECTION SUBGRAPH				*/
  int*	nlvl,	/* NUMBER OF LEVELS IN THE LEVEL STRUCTURE		*/
  int*	xls,	/* LEVEL STRUCTURE ARRAY PAIR				*/
  int*	ls,	/* LEVEL STRUCTURE ARRAY PAIR				*/
  int	num	/* 							*/
  );

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine fnroot
** fnroot finds the pseudo-peripheral node for a given subgraph.
** Original comments from Fortran:
*****************************************************************
*****************************************************************
*********     FNROOT ..... FIND PSEUDO-PERIPHERAL NODE    *******
*****************************************************************
*****************************************************************
**
**    PURPOSE - FNROOT IMPLEMENTS A MODIFIED VERSION OF THE                 7.
**       SCHEME BY GIBBS, POOLE, AND STOCKMEYER TO FIND PSEUDO-             8.
**       PERIPHERAL NODES.  IT DETERMINES SUCH A NODE FOR THE               9.
**       SECTION SUBGRAPH SPECIFIED BY MASK AND ROOT.                      10.
**
**    INPUT PARAMETERS -                                                   12.
**       (XADJ, ADJNCY) - ADJACENCY STRUCTURE PAIR FOR THE GRAPH.          13.
**       MASK - SPECIFIES A SECTION SUBGRAPH. NODES FOR WHICH              14.
**              MASK IS ZERO ARE IGNORED BY FNROOT.                        15.
**
**    UPDATED PARAMETER -                                                  17.
**       ROOT - ON INPUT, IT (ALONG WITH MASK) DEFINES THE                 18.
**              COMPONENT FOR WHICH A PSEUDO-PERIPHERAL NODE IS            19.
**              TO BE FOUND. ON OUTPUT, IT IS THE NODE OBTAINED.           20.
**
**    OUTPUT PARAMETERS -                                                  22.
**       NLVL - IS THE NUMBER OF LEVELS IN THE LEVEL STRUCTURE             23.
**              ROOTED AT THE NODE ROOT.                                   24.
**       (XLS, LS) - THE LEVEL STRUCTURE ARRAY PAIR CONTAINING              25.
**                  THE LEVEL STRUCTURE FOUND.                             26.
**
**    PROGRAM SUBROUTINES -                                                28.
**       ROOTLS.                                                           29.
**
*****************************************************************
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of fnroot ************************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   rootls   ***************
*/

/*  INTERFACE:   */

void
rootls(
  int	root,	/* NODE AT WHICH LEVEL STRUCTURE IS ROOTED		*/
  int*	xadj,	/* ADJACENCY STRUCTURE OF THE GRAPH OF THE MATRIX	*/
  int* adjncy,	/* ADJACENCY STRUCTURE OF THE GRAPH OF THE MATRIX	*/
  int* mask,	/* SPECIFIES A SECTION SUBGRAPH				*/
  int*	nlvl,	/* NUMBER OF LEVELS IN THE LEVEL STRUCTURE		*/
  int*	xls,	/* LEVEL STRUCTURE ARRAY PAIR				*/
  int*	ls,	/* LEVEL STRUCTURE ARRAY PAIR				*/
  int	num	/*							*/
  );

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine rootls
** rootls generates the level structure corresponding to "root".
** Original Fortran comments:
*****************************************************************
*****************************************************************
**********     ROOTLS ..... ROOTED LEVEL STRUCTURE      *********
*****************************************************************
*****************************************************************
**
**     PURPOSE - ROOTLS GENERATES THE LEVEL STRUCTURE ROOTED
**        AT THE INPUT NODE CALLED ROOT. ONLY THOSE NODES FOR
**        WHICH MASK IS NONZERO WILL BE CONSIDERED.
**
**     INPUT PARAMETERS -
**        ROOT - THE NODE AT WHICH THE LEVEL STRUCTURE IS TO
**               BE ROOTED.
**        (XADJ, ADJNCY) - ADJACENCY STRUCTURE PAIR FOR THE
**               GIVEN GRAPH.
**        MASK - IS USED TO SPECIFY A SECTION SUBGRAPH. NODES
**               WITH MASK(I) = 0 ARE IGNORED.
**
**     OUTPUT PARAMETERS -
**        NLVL - IS THE NUMBER OF LEVELS IN THE LEVEL STRUCTURE.
**        (XLS, LS) - ARRAY PAIR FOR THE ROOTED LEVEL STRUCTURE.
**
*****************************************************************
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of rootls ************************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   colht   ***************
*/

/*  INTERFACE:   */

void
colht(
  GROUP*	grp_ptr,/* pointer to current element group		*/
  PROFILE*	a	/* ptr to assembled profile stiffness storage	*/
  );

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine colht
** colht computes column heights of global array.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of colht *************************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   profile_diag   ***************
*/

/*  INTERFACE:   */

void
profile_diag(
  PROFILE*	a,	/* ptr to assembled profile stiffness storage	*/
  int		profile_case	/* old flag for substructuring...	*/
  );

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine profile_diag
** profile diag computes the diagonal addresses.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of profile_diag ******************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   factor   ***************
*/

/*  INTERFACE:   */

void
factor(
  real*	stiff,		/* assembled profile matrix			*/
  int*	diag,		/* array of diagonal addresses			*/
  int	number_of_eqs	/* number of equations				*/
  );

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine factor
** factor performs profile-based factorization:
**                          t
**                     a=(u) * d * u (crout)
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of factor ************************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   full_back   ***************
*/

/*  INTERFACE:   */

void
full_back(
  real*	stiff,		/* factored profile matrix			*/
  real*	rhs,		/* assembled rhs vector				*/
  int*	diag,		/* array of diagonal addresses			*/
  int	number_of_eqs	/* number of equations				*/
  );

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine full_back
** full_back performs all three steps of backsubstitution.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of full_back *********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   pcg_loop   ***************
*/

/*  INTERFACE:   */

void
pcg_loop(pamr_Mesh mesh[]);

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine pcg_loop
** pcg_loop performs the preconditioned conjugate gradient iteration
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of pcg_loop **********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   converged   ***************
*/

/*  INTERFACE:   */

int converged(
  int	k	/* iteration number					*/
  );

/*  RETURN VALUE:  flag indicating successful convergence (or not) */
/*  DESCRIPTION:   */
/*
** function converged()
** monitors convergence of CG solver from magnitude and history of
** the residual norm
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of converged *********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   put_soln   ***************
*/

/*  INTERFACE:   */

void
put_soln(
  RHS_DATA*	solve, 	/* structure containing rhs arrays		*/
  int*		id	/* id array of eq numbers			*/
  );

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine put_soln
** put_soln transfers solution to nodal storage.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of put_soln **********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   globalize   ***************
*/

/*  INTERFACE:   */

       void
       globalize(
	        const pamr_Mesh m[],     /* the input mesh */
                real  *in_vect,          /* item to be globalized */
                int reset                /* do buffer reset only */
		);

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine globalize
** globalize sums the partial contributions from data on partition boundaries
** by pair-wise communication
**
** NOTE: Consider making bndy buffers static to prevent repeated allocations
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   dot_par   ***************
*/

/*  INTERFACE:   */

  real dot_par( pamr_Mesh mesh[], /* the mesh */
                real   *vect_1 , /* multiplicand array */
                real   *vect_2 , /* multiplicand array */
                int    n       /* number of entries */
		);

/*  RETURN VALUE:  value of the scalar product */
/*  DESCRIPTION:   */
/*
** Function dot_par
** dot_par forms dot product of two vectors. 
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
