/*
***		                       File stiff.h                       ***
***                        GeoFEST version 4.5
*** Copyright (c) 2003, California Institute of Technology        ***
*** U.S.Sponsorship under NASA Contract NAS7-1260 is acknowledged ***
***
*** This software is designated for public release under JPL Task ***
*** Plan Number 83-6791 and may be publicly released after        ***
*** approval for release has been granted by GSFC Software        ***
*** Release  Authority
***
*** This file contains modules which construct the finite
*** element stiffness matrix and the right-hand-side "force" vector.
***
***          == form_stiff ==
***          == form_rhs ==
***          == form_bc ==
***          == form_slip ==
***          == adjust_bc ==
***          == lame_form ==
***          == duo_form ==
***          == force_form ==
***          == surf_form ==
***          == buoy_form ==
***          == shape ==
***          == dotsh ==
***          == adfldp ==
***          == p_shape ==
***          == form_equil ==
***/

/*
***GeoFEST version 4.5***
G. A. Lyzenga -- unified parallel/serial version and bug fixes 11/17/03
G. A. Lyzenga -- split node enhancements 8/16/02
G. A. Lyzenga -- modification to correctly relax hydrostatic stresses 6/15/00
J. W. Parker -- 3D support and deleted hypercube branches 9/16/98
G. A. Lyzenga  -- last modified on 3/03/95
Modification history:
2.1: fixes bug in boundary velocities and changes
     the way stresses are output.
2.2: implements split node faults
2.3: adds pass-thru comments
3.0: 3-D tetrahedra supported, hypercube branches deleted
3.1: hydrostatic relaxation changes
4.0: name change from visco to GeoFEST; correct implementation of 2-D out-of-
     plane stresses as well as hydrostatic relaxation; culling of remaining
     vestiges of hypercube stuff; one code without recompilation for both
     2-D and 3-D domains
4.1: Introduction of trilinear hexahedral bricks and PCG iterative solver option
4.2: Bug fixes; standard source prologues added (June 02)
4.3: Corrects inefficiency in split node algorithm and introduces new
     split node input format
4.4: Adds dummy calls to MPI and Pyramid to provide a single unified
     source for parallel and serial codes.  Also introduces a major
     algorithmic enhancement; stress equilibrium maintenance during visco
     elastic stepping for long-term stability.  Also more minor bug fixes.
4.5: Supports surface tractions and roller/spring duo options for tet elements.
     Provides enhanced conjugate gradient convergence monitoring.
 $Log: stiff.h,v $
 Revision 1.1  2003/02/09 06:10:57  lyzenga
 *** empty log message ***

*/

#include <stdio.h>
#include <stdlib.h>
#include "mydefs.h"
#include "errcodes.h"
#include "finel.h"
#include <math.h>

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   form_stiff   ***************
*/

/*  INTERFACE:   */

void
form_stiff(
  GROUP*	grp_ptr,/* pointer to current element group */
  int		code	/* code specifying elastic or visco */
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine form_stiff
** form_stiff computes the element-wise stiffness array for the elt group.
** "code" indicates if this is for an elastic problem or a VE step.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of form_stiff ********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   form_rhs   ***************
*/

/*  INTERFACE:   */

void
form_rhs(
  GROUP*	grp_ptr,/* pointer to current element group */
  int		code	/* code specifying elastic or visco */
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine form_rhs
** form_rhs computes the right-hand-side vector for the FE problem.
** "code" indicates if this is for the elastic problem or a VE step.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of form_rhs **********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   form_bc   ***************
*/

/*  INTERFACE:   */

void
form_bc(
  GROUP*	grp_ptr,/* pointer to current element group */
  int		code	/* code specifying elastic or visco */
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine form_bc
** form_bc computes the boundary condition terms for stiffness and rhs
** due to imposed conditions
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of form_bc ***********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   form_slip   ***************
*/

/*  INTERFACE:   */

void
form_slip(
  int		code,	/* code specifying elastic or visco		*/
  real*		rhs,	/* temp array to hold element rhs contrib	*/
  real*		stiff,	/* temp array to hold el stiffness contrib	*/
  ELEMENT_INFO*	info,	/* ptr to element info struct			*/
  int		nel,	/* current element number			*/
  int		node,	/* current node number				*/
  real*		slip,	/* array containing fault slip amplitudes	*/
  ELEMENT_DATA*	el_pt,	/* ptr to current element struct		*/
  ELEMENT_MAT*	mat_pt	/* ptr to el material prop struct		*/
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine form_slip
** form_slip computes the split-nodes fault offsets and their influence
** on the finite element stiffness and RHS.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of form_slip *********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   adjust_bc   ***************
*/

/*  INTERFACE:   */

void
adjust_bc(
  real*	disp,	/* array containing fault slip amplitudes		*/
  real*	rhs,	/* temp array to hold element rhs contrib		*/
  real*	es,	/* temp array to hold el stiffness contrib		*/
  int	nee	/* number of element equations				*/
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine adjust_bc
** adjust_bc computes the necessary terms that modify the right-hand-side
** due to imposed displacements.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of adjust_bc *********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   lame_form   ***************
*/

/*  INTERFACE:   */

void
lame_form(
  int		code,	/* code specifying elastic or visco		*/
  real*		es,	/* temp array to hold el stiffness contrib	*/
  ELEMENT_INFO*	info,	/* ptr to element info struct			*/
  ELEMENT_MAT*	mat_pt,	/* ptr to el material prop struct		*/
  ELEMENT_DATA*	el_pt	/* ptr to current element struct		*/
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine lame_form
** lame_form computes the volumetric stiffness term for one element,
** that is based on the constitutive volume constants.
** Results are stored in the element stiffness array.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of lame_form *********************************/
/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   duo_form   ***************
*/

/*  INTERFACE:   */

     void
     duo_form(
               int  code ,     /* code specifying elastic or visco */
               real  *es ,    /* temp array to hold el stiffness contrib */
               ELEMENT_INFO  *info ,   /* ptr to element info struct */
               ELEMENT_MAT   *mat_pt ,  /* ptr to el material prop struct */
               ELEMENT_DATA  *el_pt    /* ptr to current element struct */
              );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine duo_form
** duo_form computes the stiffness term for a two-node element,
** with options for either a truss "roller" or an isotropic spring. 
** Results are stored in the element stiffness array.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of duo_form *********************************/
/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   force_form   ***************
*/

/*  INTERFACE:   */

void
force_form(
  int		code,	/* code specifying elastic or visco		*/
  real*		rhs,	/* temp array to hold element rhs contrib	*/
  ELEMENT_INFO*	info,	/* ptr to element info struct			*/
  ELEMENT_MAT*	mat_pt,	/* ptr to el material prop struct		*/
  ELEMENT_DATA*	el_pt	/* ptr to current element struct		*/
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine force_form
** force_form computes the "dumb" gravity contribution to the element
** right-hand side term for the elastic case, and the strain-rate
** term for the VE case.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of force_form ********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   surf_form   ***************
*/

/*  INTERFACE:   */

void
surf_form(
  int		code,	/* code specifying elastic or visco		*/
  real*		rhs,	/* temp array to hold element rhs contrib	*/
  ELEMENT_INFO*	info,	/* ptr to element info struct			*/
  int		side,	/* index number of the element side		*/
  real*		trac,	/* array containg traction components		*/
  ELEMENT_DATA*	el_pt	/* ptr to current element struct		*/
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine surf_form
** surf_form computes the surface traction forcing term for the
** right-hand-side in the finite element problem.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of surf_form *********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   buoy_form   ***************
*/

/*  INTERFACE:   */

     void
     buoy_form(
               int  code ,     /* code specifying elastic or visco */
               real  *rhs ,   /* temp array to hold element rhs contrib */
               ELEMENT_INFO  *info ,   /* ptr to element info struct */
               int  side ,    /* index number of the element side */
               real  *up_vec ,   /* array containing local "up" */
               real  rho_g ,   /* array containing density*g contrast */
               ELEMENT_DATA  *el_pt    /* ptr to current element struct */
              );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine buoy_form
** buoy_form computes the buoyancy traction term due to displacement
** of a density contrast in the finite element grid.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of buoy_form *********************************/
/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   shape   ***************
*/

/*  INTERFACE:   */

void
shape(
  ELEMENT_DATA*	el_pt,	/* ptr to current element struct		*/
  ELEMENT_INFO*	info,	/* ptr to element info struct			*/
  real*		sh_pt,	/* el storage to hold shape functions		*/
  real*		det_pt,	/* el storage to hold determinant		*/
  int		type,	/* element type number				*/
  int		nen,	/* number of element nodes			*/
  int		nint,	/* number of Gauss integration points		*/
  int		all	/* flag indicating we need derivatives also	*/
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine shape
** shape computes the shape functions and gradients for a single element.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of shape *************************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   dotsh   ***************
*/

/*  INTERFACE:   */

real dotsh(
  real*		array,	/* array of nodal quantities to interpolate	*/
  int		i,	/* derivative component index number		*/
  int		j,	/* dof index number of quantity component	*/
  real*		sh,	/* el storage containing shape functions	*/
  ELEMENT_DATA*	el_pt,	/* ptr to current element struct		*/
  ELEMENT_INFO*	info,	/* ptr to element info struct			*/
  int		nen,	/* number of element nodes			*/
  int		ndim,	/* number of dimensions in array		*/
  int		st_flag	/* flag to invoke split node adjustment		*/
  );

/*  RETURN VALUE:  the element-interpolated sum value */
/*  DESCRIPTION:   */
/*
** Function dotsh
** dotsh computes the dot product of a set of shape-functions (or gradients)
** with a vector of values, hence computing an interpolated coordinate
** or function. If st_flag, x is assumed to represent displacement, and
** a correction for split-nodes is applied prior to the dot product.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of dotsh *************************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   adfldp   ***************
*/

/*  INTERFACE:   */

real adfldp(
            int  j ,     /* dof index number of slip component */
            ELEMENT_DATA  *el_pt  ,  /* ptr to current element struct */
            int  node ,    /* number of current node */
            ELEMENT_INFO  *info    /* ptr to element info struct */
           );

/*  RETURN VALUE:  the required additional nodal displacement */
/*  DESCRIPTION:   */
/*
** Function adfldp
** adfldp returns to dotsh() the added nodal displacements
** needed to account for split node fault slip in calculating strain
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of adfldp ************************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   p_shape   ***************
*/

/*  INTERFACE:   */

void
p_shape(
  GROUP*	grp_ptr	/* pointer to current element group		*/
  );

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine p_shape
** p_shape computes the parent-space shape function for common element types.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of p_shape ***********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   form_equil   ***************
*/

/*  INTERFACE:   */
     void
     form_equil(
                GROUP   *grp_ptr    /* pointer to current element group */
               );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine form_equil
** computes corrections necessary to preserve equilibrium 
** in stress increments calculated during viscous stepping
**
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of form_equil *********************************/ 

