/*
***                           File strain.h                       ***
***                        GeoFEST version 4.5
*** Copyright (c) 2003, California Institute of Technology        ***
*** U.S.Sponsorship under NASA Contract NAS7-1260 is acknowledged ***
***
*** This software is designated for public release under JPL Task ***
*** Plan Number 83-6791 and may be publicly released after        ***
*** approval for release has been granted by GSFC Software        ***
*** Release  Authority
***
*** This file contains modules which perform calculations 
*** related to element stresses and strains:
***
***          == form_stress ==
***          == form_beta ==
***          == form_dbar ==
***/

/*
***GeoFEST version 4.5***
G. A. Lyzenga -- unified parallel/serial version and bug fixes 11/17/03
G. A. Lyzenga -- split node enhancements 8/16/02
G. A. Lyzenga -- modification to correctly relax hydrostatic stresses 6/15/00
J. W. Parker -- 3D support and deleted hypercube branches 9/16/98
G. A. Lyzenga  -- last modified on 3/03/95
Modification history:
2.1: fixes bug in boundary velocities and changes
     the way stresses are output.
2.2: implements split node faults
2.3: adds pass-thru comments
3.0: 3-D tetrahedra supported, hypercube branches deleted
3.1: hydrostatic relaxation changes
4.0: name change from visco to GeoFEST; correct implementation of 2-D out-of-
     plane stresses as well as hydrostatic relaxation; culling of remaining
     vestiges of hypercube stuff; one code without recompilation for both
     2-D and 3-D domains
4.1: Introduction of trilinear hexahedral bricks and PCG iterative solver option
4.2: Bug fixes; standard source prologues added (June 02)
4.3: Corrects inefficiency in split node algorithm and introduces new
     split node input format
4.4: Adds dummy calls to MPI and Pyramid to provide a single unified
     source for parallel and serial codes.  Also introduces a major
     algorithmic enhancement; stress equilibrium maintenance during visco
     elastic stepping for long-term stability.  Also more minor bug fixes.
4.5: Supports surface tractions and roller/spring duo options for tet elements.
     Provides enhanced conjugate gradient convergence monitoring.
 $Log: strain.h,v $
 Revision 1.1  2003/02/09 06:10:57  lyzenga
 *** empty log message ***

*/

#include <stdio.h>
#include "mydefs.h"
#include "errcodes.h"
#include "finel.h"
#include <math.h>

/* real	dotsh(); */

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   form_stress   ***************
*/

/*  INTERFACE:   */

void form_stress(
  GROUP*	grp_ptr,/* pointer to current element group	*/
  int		code	/* code specifying elastic or visco	*/
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine form_stress
** form_stress computes the stress or time-step stress increment
** based on the FE solution.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of form_stress *******************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   form_beta   ***************
*/

/*  INTERFACE:   */

void
form_beta(
  ELEMENT_DATA*	el_pt,	/* ptr to current element struct	*/
  ELEMENT_MAT*	mat_pt,	/* ptr to el material prop struct	*/
  ELEMENT_INFO*	info	/* ptr to element info struct		*/
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine form_beta
** form_beta computes beta, the viscoplastic strain rate for
** this element and time step.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of form_beta *********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   form_dbar   ***************
*/

/*  INTERFACE:   */

void
form_dbar(
  ELEMENT_DATA*	el_pt,	/* ptr to current element struct	*/
  ELEMENT_MAT*	mat_pt,	/* ptr to el material prop struct	*/
  ELEMENT_INFO*	info	/* ptr to element info struct		*/	
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine form_dbar
** form_dbar computes dbar, the VE single-step
** constitutive matrix:
**
** dbar === (S + alpha delta_t beta')^-1
**
** but we compute the factored form, not the explicit inverse.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of form_dbar *********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   duostress   ***************
*/

/*  INTERFACE:   */
     void
     duostress(
               ELEMENT_DATA  *el_pt ,  /* ptr to current element struct */
               ELEMENT_MAT   *mat_pt , /* ptr to el material prop struct */
               ELEMENT_INFO  *info     /* ptr to element info struct */
              );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine duostress
** calculates stresses associated with 2-node truss or spring type elements
** (special case of 4-node tetrahedral elements)
**
** 1 component for trusses; 3 components for springs
**
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of duostress *********************************/
