/*
***                           File utility.h                      ***
***                        GeoFEST version 4.5
*** Copyright (c) 2003, California Institute of Technology        ***
*** U.S.Sponsorship under NASA Contract NAS7-1260 is acknowledged ***
***
*** This software is designated for public release under JPL Task ***
*** Plan Number 83-6791 and may be publicly released after        ***
*** approval for release has been granted by GSFC Software        ***
*** Release  Authority
***
*** This file contains miscellaneous utility routines used throughout
*** the finite element program: 
***
***          == move_real ==
***          == clear_real ==
***          == dot_real ==
***          == incr_real ==
***          == vadd ==
***          == vouter ==
***          == squawk ==
***/

/*
***GeoFEST version 4.5***
G. A. Lyzenga -- unified parallel/serial version and bug fixes 11/17/03
G. A. Lyzenga -- split node enhancements 8/16/02
G. A. Lyzenga -- modification to correctly relax hydrostatic stresses 6/15/00
J. W. Parker -- 3D support and deleted hypercube branches 9/16/98
G. A. Lyzenga  -- last modified on 3/03/95
Modification history:
2.1: fixes bug in boundary velocities and changes
     the way stresses are output.
2.2: implements split node faults
2.3: adds pass-thru comments
3.0: 3-D tetrahedra supported, hypercube branches deleted
3.1: hydrostatic relaxation changes
4.0: name change from visco to GeoFEST; correct implementation of 2-D out-of-
     plane stresses as well as hydrostatic relaxation; culling of remaining
     vestiges of hypercube stuff; one code without recompilation for both
     2-D and 3-D domains
4.1: Introduction of trilinear hexahedral bricks and PCG iterative solver option
4.2: Bug fixes; standard source prologues added (June 02)
4.3: Corrects inefficiency in split node algorithm and introduces new
     split node input format
4.4: Adds dummy calls to MPI and Pyramid to provide a single unified
     source for parallel and serial codes.  Also introduces a major
     algorithmic enhancement; stress equilibrium maintenance during visco
     elastic stepping for long-term stability.  Also more minor bug fixes.
4.5: Supports surface tractions and roller/spring duo options for tet elements.
     Provides enhanced conjugate gradient convergence monitoring.
 $Log: utility.h,v $
 Revision 1.1  2003/02/09 06:10:58  lyzenga
 *** empty log message ***

*/

#include <stdio.h>
#include "mydefs.h"
#include "errcodes.h"
#include "finel.h"

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   move_real   ***************
*/

/*  INTERFACE:   */

void
move_real(
  real*	from,	/* source array		*/
  real* to,	/* destination array	*/
  int	n	/* number of entries	*/
  );

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine move_real
** move_real copies data to a new location.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of move_real *********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   clear_real   ***************
*/

/*  INTERFACE:   */

void
clear_real(
  real*	array,	/* array to be zapped	*/
  int	n	/* number of entries	*/
  );

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine clear_real
** clear_real nulls a data area.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of clear_real ********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   dot_real   ***************
*/

/*  INTERFACE:   */

real dot_real(
  real*	vect_1,	/* multiplicand array	*/
  real*	vect_2,	/* multiplicand array	*/
  int	n	/* number of entries	*/
  );

/*  RETURN VALUE:  value of the scalar product */
/*  DESCRIPTION:   */
/*
** Function dot_real
** dot_real forms dot product of two vectors. 
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of dot_real **********************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   incr_real   ***************
*/

/*  INTERFACE:   */

      void
      incr_real(
                real   *from , /* source array */
                real   *to ,   /* destination array */
                int    n       /* number of entries */
               );

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine incr_real
** incr_real increments the contents of array "to" with the contents of "from".
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of incr_real *********************************/ 

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   vadd   ***************
*/

/*  INTERFACE:   */

void vadd(
  real*	v1,	/* first array				*/
  real	mult,	/* scalar multiplier for second array	*/
  real*	v2,	/* second array				*/
  real*	dest,	/* destination array			*/
  int	neq	/* number of entries			*/
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine vadd
** calculates the linear combination of vectors
**    dest = v1 + mult*v2
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of vadd **************************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   vouter   ***************
*/

/*  INTERFACE:   */

void vouter(
  real*	v1,	/* first array		*/
  real*	v2,	/* second array		*/
  real*	dest,	/* destination array	*/
  int	neq	/* number of entries	*/
  );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine vouter
** calculates the outer product of two vectors
** actually a misnomer -- it's really the element-by-element product
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of vouter ************************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   squawk   ***************
*/

/*  INTERFACE:   */

         void
         squawk(
                char   *message  /* string to be sent to stdout */
               );

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine squawk
** Prints a message to stdout that is common to all processors only on
** processor zero
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/************************** end of squawk ************************************/ 
